#ifndef AESANODE_H
#define AESANODE_H

#include <vector>
#include <string>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <fstream>
#include <filesize.h>
#include <myagrep.h>
#include <myers.hpp>

#define INVALID_SIZE (-1)

class CAESANodeBase;


using namespace std;


typedef vector <string >          StringList;
typedef vector <CAESANodeBase*>   FQNodeList;

class CAESANodeBase {
public:
    virtual size_t DumpToFile (off_t offset, int fd) = 0;
    virtual void Insert(const StringList & lst, bool hasTransp, int BucketSize, bool silent, int level) = 0;
    virtual void SearchAtLevel (const std::string & s, int nerr, 
                        const myagrep & a, 
#ifdef MYERS_DIST
                        CMyersEdist& MyersEdist,
#endif
                        int SeqCutoff,
                        int *numComp, StringList * resList, bool hasTransp, 
                        bool silent) = 0;
    virtual void Print (int level = 0) = 0;
    virtual void Search(const std::string & s, unsigned int nerr, int SeqCutoff, int *numComp, 
                 StringList * resList, bool hasTransp, bool silent) {
        myagrep a(s.c_str(), nerr, hasTransp);
#ifdef MYERS_DIST
        CMyersEdist MyersEdist(s.c_str(), hasTransp);
#endif

        SearchAtLevel (s, nerr, a, 
#ifdef MYERS_DIST
                             MyersEdist, 
#endif
        SeqCutoff, numComp, resList, hasTransp, silent);
    }
    virtual ~CAESANodeBase() = 0; 

    const static char   termChar;
};


class CAESANode : public CAESANodeBase {
protected:
    std::string         mRoot;
    FQNodeList          mvpNodes;
public:
   CAESANode () {}
   virtual ~CAESANode();

    virtual size_t DumpToFile (off_t offset, int fd);
    virtual void Insert(const StringList & lst, bool hasTransp, int BucketSize, bool silent, int level);
    virtual void SearchAtLevel (const std::string & s, int nerr,
                        const myagrep & a, 
#ifdef MYERS_DIST
                             CMyersEdist& MyersEdist, 
#endif
                        int SeqCutoff, 
                        int *numComp, StringList * resList, bool hasTransp, 
                        bool silent);
    virtual void Print (int level = 0);
};


class CAESABucket : public CAESANodeBase {
protected:
    StringList          mvList;
    u_int8_t*           mpDistMatr; // Half of the matrix mpDistMatr[i, j] = EditDistance(ith string, jth string)
public:
    CAESABucket() : mpDistMatr(NULL) {}
    virtual ~CAESABucket();

    virtual size_t DumpToFile (off_t offset, int fd);
    virtual void Insert(const StringList & lst, bool hasTransp, int BucketSize, bool silent, int level);
    virtual void SearchAtLevel (const std::string & s, int nerr, const myagrep & a, 
#ifdef MYERS_DIST
                             CMyersEdist& MyersEdist, 
#endif
                        int SeqCutoff, 
                        int *numComp, StringList * resList, bool hasTransp, 
                        bool silent);
    virtual void Print (int level = 0);
    inline size_t GetMatrSize(unsigned N) { return N * (N + 1) / 2; }
    inline u_int8_t& GetDistCell(int i, int j) {
        if (j > i) { swap(i, j); }
        return mpDistMatr[i * (i + 1) / 2 + j];
    }
};

#endif
