#include "index.h"
#include "myagrep.h"

CAESAIndex::CAESAIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false)
{
}

CAESAIndex::~CAESAIndex ()
{
    if (opened)
        CloseDB ();
}

bool CAESAIndex::Insert(const char *str)
{
    cerr << "Does not support insertion of single strings" << endl;

    return false;
}

bool CAESAIndex::InsertList(const StringList & lst)
{
    cerr << "InsertList is not supported, use CreateInMemIndex instead" << endl;
    return true;
}

bool CAESAIndex::Search(const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    string s (str);
    int SeqCutoff = SEQ_CUTOFF;

    const char *pBuck = getenv("SEQ_CUTOFF");
    if (pBuck && atol(pBuck)) {
        SeqCutoff = atol(pBuck);
    }

    node.Search (s, kerr, SeqCutoff, numComp, resList, mHasTransp, silent);
    return true;
}

bool CAESAIndex::SearchInDB(const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
/*
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return node.SearchInFile (fd, string (str), kerr, numComp, resList, mHasTransp,
                              silent);
*/
    cerr << "SearchInDB is not supported" << endl;

    return false;
}

bool CAESAIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();

    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }

    return (opened = true);
}

bool CAESAIndex::CloseDB ()
{
    if (!opened)
        return true;

    if (close (fd) != 0)
        return false;

    opened = false;
    return true;
}


bool CAESAIndex::DumpDB (const char *file_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    bool stat = (node.DumpToFile (0, fd) != INVALID_SIZE);
    if (stat == -1) {
        Perror("Cannot dump");
        return false;
    }
    return stat && CloseDB ();
}

bool CAESAIndex::CreateInMemIndex (const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }
    StringList lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);
        lst.push_back (s);

        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }


    int BucketSize = BUCKET_SIZE;
    const char *pBuck = getenv("BUCKET_SIZE");
    if (pBuck && atol(pBuck)) {
        BucketSize = atol(pBuck);
    }
    cout << "Bucket size: "<< BucketSize << endl;

    node.Insert(lst, mHasTransp, BucketSize, silent, 0);

#if 0
    Print();
#endif
    return true;
}


void CAESAIndex::Print ()
{
    node.Print ();
}
