#ifndef AHC_INDEX_H
#define AHC_INDEX_H

#include <vector>
#include <string>

using namespace std;

typedef vector<int> VI;
typedef vector<string> VS;

class CAhcTree {
private:
    const VS &                mDict;
    const CAhcTree *          mFirst, * mSecond;
    VI                        mElems;
    VI                        mElemRad;
    int                       mCenterIndex;
    // this constructor is for private usage only so as one cannot construct an object without elements!
    CAhcTree(const VS & dict):mDict(dict), mFirst(NULL), mSecond(NULL) {}
public:
    CAhcTree(const VS & dict, int indx):mDict(dict), mFirst(NULL), mSecond(NULL) {
        mCenterIndex = indx;
        mElems.push_back(indx);
        mElemRad.push_back(0);
    };
    virtual ~CAhcTree() {
        delete mFirst;
        delete mSecond;
    }
    void search(const string & pat, int k, VS & res, int & cnt, bool first = true) const;
    int getRad(){return mElemRad[mCenterIndex];}
    int getCenterIndex(){return mCenterIndex;}
    friend CAhcTree * joinClusters(const CAhcTree & lhs, const CAhcTree &rhs);
    friend int getJointRad(const CAhcTree & lhs, const CAhcTree &rhs);
    unsigned int cnt(int addVal) const;
    unsigned int depth(int addVal) const;
    void fill(VS & str, int row, int col) const;
};


typedef vector<CAhcTree *>   AhcForest;
void generateTrees(AhcForest & res, const VS & dict, unsigned int maxTreeCnt);
CAhcTree * joinClusters(const CAhcTree & lhs, const CAhcTree &rhs);
// new object must be destroyed using delete function 
int getJointRad(const CAhcTree & lhs, const CAhcTree &rhs);
void print(CAhcTree & tree);

#endif
