#include "index.h"


CSpellIndex::CSpellIndex (bool hasTransp, EAlphabet AlphaType)
              : CInterface(hasTransp, AlphaType), opened (false)
{
    const char * _ABC = GetAlphabet(AlphaType);

    int size = strlen (_ABC);
    ABC = new char[size + 1];
    memcpy (ABC, _ABC, size + 1);
}


CSpellIndex::~CSpellIndex ()
{
    delete[]ABC;
}

bool CSpellIndex::Insert (const char *str)
{
    Dict.insert (string (str));
    return true;
}

bool CSpellIndex::Search (const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    StringList       lst;

    FillExpansionList (string (str), mHasTransp, kerr, &lst, ABC, 0, false, NULL);

    for (StringList::iterator i = lst.begin (); i != lst.end (); i++) {
        if (numComp)
            (*numComp)++;
        if (Dict.find (*i) != Dict.end ()) {
            if (resList != NULL)
                resList->push_back (string (*i));
            if (!silent) {
                cout << *i << endl;
            }
        }
    }
    return true;
}

bool CSpellIndex::CloseDB ()
{
    if (!opened)
        return true;
    if ((errno = dbp->close (dbp, 0)) != 0) {
        cerr << db_strerror (errno) << endl;
        return false;
    }
    free (dbenv);
    dbenv = NULL;
    opened = false;
    return true;
}

bool CSpellIndex::OpenDB (const char *base_name, bool create)
{
    if (opened) {
        CloseDB ();
    }
    dbenv = db_init (NULL, CSpellIndex::cachesize);
    if (dbenv == NULL) {
        return false;
    }
    /* Initialize the database. */
    if ((errno = db_create (&dbp, dbenv, 0)) != 0) {
        cerr << db_strerror (errno) << endl;
        return false;
    }

    if ((errno = dbp->set_pagesize (dbp, CSpellIndex::pagesize)) != 0) {
        cerr << db_strerror (errno) << endl;
        return false;
    }

    /* Opening the database. */
    if ((errno = dbp->open (dbp, NULL, base_name, NULL, DB_HASH,
                          (!create) ? DB_RDONLY : DB_CREATE | DB_TRUNCATE,
                          defaultAccess))) {
        cerr << (db_strerror (errno))<<endl;
        return false;
    }
    return opened = true;
}

bool CSpellIndex::SearchInDB (const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened" << endl;
        return false;
    }
    /* Searching the database */
    StringList lst;
    DBC *dbcp;
    DBT key, data;
    memset (&key, 0, sizeof (key));
    memset (&data, 0, sizeof (data));
    FillExpansionList (string (str), mHasTransp, kerr, &lst, ABC, 0, false, NULL);
    for (StringList::iterator i = lst.begin (); i != lst.end (); i++) {
        if (numComp)
            (*numComp)++;
        //if (!silent) cout<<*i<<endl;
        key.data = (void *) i->c_str ();
        key.size = i->size () + 1;
        if ((errno = dbp->cursor (dbp, NULL, &dbcp, 0))) {
            cerr << strerror (errno) << "\n";
            return false;
        }
        unsigned int flags = DB_SET;
        if ((errno = dbcp->c_get (dbcp, &key, &data, flags))
            && errno != DB_NOTFOUND) {
            cerr << strerror (errno) << "\n";
            return false;
        }
        if (!errno) {
            if (!silent)
                cout << *i << endl;
            if (resList != NULL)
                resList->push_back (string (*i));
        }
        if ((errno = dbcp->c_close (dbcp)) != 0) {
            cerr << strerror (errno) << "\n";
            return false;
        }

    }

    return true;
}

bool CSpellIndex::DumpDB (const char *base_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database


    DBT key, data;
    memset (&key, 0, sizeof (key));
    memset (&data, 0, sizeof (data));
    data.data = NULL;
    data.size = 0;
    /* Dumping words */
    for (StringSet::iterator it = Dict.begin (); it != Dict.end (); it++) {
        key.data = (void *) it->c_str ();
        key.size = it->size () + 1;
        errno = dbp->put (dbp, NULL, &key, &data, DB_NOOVERWRITE);
        if (errno) {
            std::cerr << "Error storing entry in Berkley DB: " << (db_strerror (errno)) << std::endl;
            return false;
        }
    }
    return CloseDB ();
}
