#include "index.h"
#include "hash.h"

static long long seekVal;

CSpellIndex::CSpellIndex (bool hasTransp, EAlphabet AlphaType)
              : CInterface(hasTransp, AlphaType), opened (false)
{
    const char * _ABC = GetAlphabet(AlphaType);

    int size = strlen (_ABC);
    ABC = new char[size + 1];
    memcpy (ABC, _ABC, size + 1);
}


CSpellIndex::~CSpellIndex ()
{
    delete [] ABC;
}

bool CSpellIndex::Insert(const char *p)
{
    mSpellIndex.insert(p);

    return true;
}

void CSpellIndex::ExpandAndSearch(StringHashSet& hGenerated,
                        int start,
                        int &numComp, 
                        string & str, 
                        int kerr, 
                        StringList *resList,
                        bool    bFullExpansion)
{
    //if (hGenerated.find(str) == hGenerated.end()) {
        //hGenerated.insert(str);
    if (1) {
#if 0
        if (GetEnvSwitch("SHOW_VARIANTS")) {
            std::cout << "Variant: " << str << std::endl;
        }
#endif
        // If bFullExpansion == true  then mSpellIndex looks for exact locations only
        //  HasTransp flag is false, because transpositions will be accounted for by the code below
        bool bExOnly = bFullExpansion;
        mSpellIndex.search(str, mHasTransp, &numComp, resList, bExOnly);
    }

    int size = str.size ();
    if (!kerr || start > size) return;

    int i = 0;
    // Substitution
    for (i = start; i < size; i++) {
        char origChar = str[i];
        for (const char *ptr = ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                str[i] = *ptr;
                ExpandAndSearch(hGenerated, i + 1, numComp,  str, kerr - 1, resList, bFullExpansion);
            }
        }
        str[i] = origChar;
    }
    // We well try to make as few insertions and erases as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    str.insert(str.begin() + start, ' ');
    for (i = start; i <= size; i++) {
        for (const char *ptr = ABC; *ptr; ptr++) {
            str[i] = *ptr; 
            ExpandAndSearch(hGenerated, i + 1, numComp,  str, kerr - 1, resList, bFullExpansion);
        }
        if (i < size) {
            swap(str[i], str[i + 1]);
        }
    }
    str.erase(str.end() - 1);
    if (mHasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(str[i], str[i+1]);
            ExpandAndSearch(hGenerated, i + 2, numComp,  str, kerr - 1, resList, bFullExpansion);
            swap(str[i], str[i+1]);
        }
    }
    // Deletion
    if (start >= size) return;
    char origChar = str[start];
    str.erase(str.begin() + start);
    for (i = start; i < size; i++) {
        ExpandAndSearch(hGenerated, i, numComp,  str, kerr - 1, resList, bFullExpansion);
        if (i < size - 1) {
           swap(str[i], origChar);
        }
    }
    str.insert (str.end(), origChar);

}


bool CSpellIndex::Search(const char *p, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string str(p);
    if (!kerr) {
        mSpellIndex.search(str, mHasTransp, numComp, resList, true);
    } if (kerr < 2) {
        mSpellIndex.search(str, mHasTransp, numComp, resList);
    } else {
        int numComp1 = 0;
        StringHashSet hGenerated;

        if (str.size() >= kerr + 2) {
            ExpandAndSearch(hGenerated, 0, numComp1, str, kerr - 1, resList, false);
        } else {
            // If a pattern is too short, just full neighborhood generation is much better
            ExpandAndSearch(hGenerated, 0, numComp1, str, kerr, resList, true);
        }
        if (numComp) *numComp = numComp1;
    }
    return true;
}

bool CSpellIndex::OpenDB (const char *base_name, bool create)
{
    if (opened)
        CloseDB ();

    string    headerName = string(base_name) + ".hdr";
    string    hashName = string(base_name) + ".hash";
    string    dictName = string(base_name) + ".dict";

    int flags = O_RDWR;
    int access = 0;

    pDict = NULL;
    pHash = NULL;
    pParams = NULL;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }
    fdParams = open(headerName.c_str(), flags, access);
    if (fdParams < 0) {
        Perror("Cannot open header file");
        return false;
    }
    fdHash = open(hashName.c_str(), flags, access);
    if (fdHash < 0) {
        Perror("Cannot open hash file");
        close(fdParams);
        return false;
    }
    fdDict = open(dictName.c_str(), flags, access);
    if (fdDict < 0) {
        Perror("Cannot dict file");
        close(fdHash);
        close(fdParams);
        return false;
    }

    if (!create) {
        opened = true;
        if (!mapData()) {
            CloseDB();
            return false;
        }
    }
    
    return opened = true;
}

bool CSpellIndex::DumpDB(const char *base_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database

    size_t numWords = mSpellIndex.GetDictSize();
    vector<pair<size_t, COneDelIndex::ConstIndexIterator> >   pos2it;
    
    size_t dictSize = 0;
    vuc arrPos, arrSym;

    int cntIt = 0;
    int cntRec = 0, cntRecMax = 0;
    // The first set iteration - write 0 and calc size
    for (COneDelIndex::ConstIndexIterator it = mSpellIndex.begin(); it != mSpellIndex.end(); ++it, ++cntIt) {
        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

#if 0
        if (cntIt % 100000 == 0) {
            cout << cntIt <<" of " << numWords << " " << 100 * cntIt / numWords << "%" << endl;
        }
#endif

        int cnt;
        it.GetPosSym(arrPos, arrSym, cnt);

        cntRec += cnt;
        cntRecMax = max(cnt, cntRecMax);

        // if the word length is less than 255, then 
        // no more than 255 words can be converted into a single
        // word by the deletion of a single character
        if (cnt > 255) {
           CloseDB();
           cerr << Format("Count %d for word '%s' is too large\n", cnt, str.c_str());
           return false;
        }
        size_t addSize = strSize + 1 + 2 * cnt; // 1 counter 
        
        char           zeroBuf[addSize];

        memset(zeroBuf, 0, addSize);
        if (write(fdDict, zeroBuf, addSize) != addSize) {
            Perror("Cannot write to dict file");
            CloseDB();
            return false;
        }
        dictSize += addSize;
        pos2it.push_back(make_pair(0U, it));
    }

    if (GetEnvSwitch("PRINT_ADD_STAT")) {
        cout << "Average list len (k = 1): " << double(cntRec) / cntIt << " max: " << cntRecMax << endl;
    }

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 
    if (!hashSize) hashSize = 1;


    size_t tmp(0);
    if (lseek(fdHash, getPageAlignedValue((hashSize - 1) * sizeof(size_t)), SEEK_SET) < 0 ||
         write(fdHash, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write");
        CloseDB();
        return false;
    }
    CHashParams prm(fillRatio, hashSize, dictSize);

    if (write(fdParams, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params");
        CloseDB();
        return false;
    }

    if (!mapData()) {
        CloseDB();
        return false;
    }

    memset(pHash, 0, pParams->hashSize * sizeof(size_t));

    size_t n = 0;
    cntIt = 0;
    seekVal = 0;
    // Find out which hash table elements to fill
    for (COneDelIndex::ConstIndexIterator it = mSpellIndex.begin(); it != mSpellIndex.end(); ++it, ++cntIt) {
        const string & str = it.GetWord();
        size_t         buckNum;
        if (!insertHashDB(str, buckNum, *pParams, pHash, &seekVal)) {
            cerr << "Failed to insert string internally" << endl;
            CloseDB();
            return false;
        }
#if 0
        if (cntIt % 100000 == 0) {
            cout << cntIt <<" of " << numWords << " " << 100 * cntIt / numWords << "% (2) seekval" << double(seekVal) / max(1, cntIt) << endl;
        }
#endif
        pos2it[n++].first = buckNum;
    }
    // Finally store strings in the order they are in hash table
    // and fill hash positions
    sort(pos2it.begin(), pos2it.end()); // will sort by position

    size_t pos = 0;
    for (n = 0; n < pos2it.size(); ++n) {
        const COneDelIndex::ConstIndexIterator &it = pos2it[n].second;

        pHash[pos2it[n].first] = pos + 1; // do not store zero value

        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

        memcpy(&pDict[pos], str.data(), strSize);
        pos += strSize;

        int cnt;
        it.GetPosSym(arrPos, arrSym, cnt);

        // if the word length is less than 255, then 
        // no more than 255 words can be converted into a single
        // word by the deletion of a single character
        
        if (cnt > 255) {
           CloseDB();
           cerr << Format("Count %d for word '%s' is too large\n", cnt, str.c_str());
           return false;
        }

        unsigned char uCnt = cnt;
        memcpy(&pDict[pos], &uCnt, sizeof uCnt);
        pos += 1;

        memcpy(&pDict[pos], &arrPos[0], cnt);
        pos += cnt;

        memcpy(&pDict[pos], &arrSym[0], cnt);
        pos += cnt;

    }

    return CloseDB();
}

bool CSpellIndex::mapData()
{
    if (!opened) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }
    pParams = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof *pParams), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdParams, 0));
    if (!pParams || pParams == MAP_FAILED) {
        Perror("Cannot map params");
        pParams = NULL;
        return false;
    }
    pHash = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(pParams->hashSize * sizeof(size_t)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdHash, 0));
    if (!pHash || pHash == MAP_FAILED) {
        Perror("Cannot map hash");
        pHash = NULL;
        unmapData();
        return false;
    }

    pDict = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(pParams->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdDict, 0));
    if (!pDict || pDict == MAP_FAILED) {
        Perror("Cannot map dict");
        pDict = NULL;
        unmapData();
        return false;
    }
    return true;
}

void CSpellIndex::unmapData(){
    if (pParams) {
        if (pHash && munmap(pHash, getPageAlignedValue(pParams->hashSize * sizeof(size_t)))) {
            Perror("Cannot unmap hash");
        }
        pHash = NULL;
        if (pDict && munmap(pDict, getPageAlignedValue(pParams->dictSize))) {
            Perror("Cannot unmap dict");
        }
        pDict = NULL;
        if (munmap(pParams, getPageAlignedValue(sizeof *pParams))) {
            Perror("Cannot unmap params");
        }
        pParams = NULL;
    }
}

bool CSpellIndex::CloseDB()
{
    if (!opened)
        return true;
    unmapData();
    if (close(fdDict)) {
        Perror("Cannot close dict file");
        return false;
    }
    if (close(fdHash)) {
        Perror("Cannot close hash file");
        return false;
    }
    if (close(fdParams)) {
        Perror("Cannot close param file");
        return false;
    }
    opened = false;
    return true;
}
