#ifndef _INDEX
#define _INDEX


#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <string>
#include <iostream>
#include <fcntl.h>
#include <errno.h>
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <set>
#include <list>
#include "interface.h"
#include "onedel.h"
#include "twodel.h"
#include "../Common/hash.h"

using namespace std;

class CSpellIndex:public CInterface {

  public:
    CSpellIndex (bool hasTransp, EAlphabet AlphaType);
    virtual ~CSpellIndex ();
    virtual bool Insert (const char *str);
    virtual bool Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent = true);
    virtual bool SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent = true) {
        cerr << "SearchInDB is not supported";
        return true;
    }
    virtual bool DumpDB (const char *base_name);
    virtual bool OpenDB (const char *base_name, bool create = false);
    virtual bool CloseDB ();

    typedef StringHashSet StringSet;
  protected:
    COneDelIndex  mSpellIndex1;
    CTwoDelIndex  mSpellIndex2;
    char          *ABC;
    bool          openedAll;
    bool          opened1;
    bool          opened2;
    /* 
     * If bFullExpansion == true, then the method uses plain vanilla full neighborhood generation 
     * It the most practical method for short patterns.
     *
     */
    void ExpandAndSearch(int start, int &numComp, string & str, int kerr, StringList *resList, bool bFullExpansion);
    int  fdDict1;
    int  fdHash1;
    int  fdParams1;
    int  fdDict2;
    int  fdHash2;
    int  fdParams2;

    char    *pDict1;
    size_t  *pHash1;
    CHashParams  *pParams1;
    char    *pDict2;
    size_t  *pHash2;
    CHashParams  *pParams2;

    bool mapData();
    bool mapData1();
    bool mapData2();
    void unmapData1();
    void unmapData2();
    bool OpenDB1(const char *base_name, bool create = false);
    bool OpenDB2(const char *base_name, bool create = false);
    bool CloseDB1();
    bool CloseDB2();
    bool DumpDB1(const char *base_name);
    bool DumpDB2(const char *base_name);
};

template <>
inline bool SearchDbNotSupported<CSpellIndex>() {
    return true;
}

#endif
