#ifndef ONEDEL_H
#define ONEDEL_H

#include <vector>
#include <set>
#include <list>
#include "interface.h"

using namespace std;

typedef vector<unsigned char> vuc;

const unsigned int DefaultSize = 1000000U;

#define INTERN_CNT 4 // Should not be >= 2!!
#define REALLOC_QTY1 100000

class COneDelIndex {
public:
    struct Record {
        unsigned char mPos[INTERN_CNT];
        unsigned char mSym[INTERN_CNT];
        unsigned char mIsExt ;
        unsigned char mCnt   ;
        Record() { memset(this, 0, sizeof *this); }
        unsigned int GetPos() const { return *((u_int32_t *)&mPos[0]);}
        unsigned int SetPos(u_int32_t val) { 
            *((u_int32_t *)&mPos[0]) = val;
            mIsExt = 1;
        }
    };
    typedef hash_map<const char *, Record, MyHash, Equal >  hmapci;
protected:
    struct MemBlock {
        char *mpBeg;
        int  mSize;
        int  mFreePos;
        MemBlock(int size) : mSize(size), mpBeg(new char[size]), mFreePos(0) {}
        ~MemBlock(){delete [] mpBeg;}
    };
public:
    class ConstIndexIterator {
        friend class COneDelIndex;
        hmapci::const_iterator   mIt;
        const COneDelIndex       *mpIndex;
        ConstIndexIterator() :mpIndex(NULL) {}
        ConstIndexIterator(const COneDelIndex *pIndex, const hmapci::const_iterator &it) : mIt(it), mpIndex(pIndex) {}
    public:
        ConstIndexIterator(const ConstIndexIterator &oth) {
            mpIndex = oth.mpIndex;
            mIt = oth.mIt;
        }
        const ConstIndexIterator &operator=(const ConstIndexIterator &oth) {
            mpIndex = oth.mpIndex;
            mIt = oth.mIt;
        }
        ConstIndexIterator &operator++() {++mIt;return *this;} // prefix
        ConstIndexIterator operator++(int k) { // postfix
            hmapci::const_iterator oldIt(mIt);
            ++mIt;
            return ConstIndexIterator(mpIndex, oldIt);
        }
        bool operator<(const ConstIndexIterator & oth) const {
            return mpIndex == oth.mpIndex && (*mIt).first < (*oth.mIt).first;
        }
        bool operator==(const ConstIndexIterator & oth) const {
            return mpIndex == oth.mpIndex && mIt == oth.mIt;
        }
        bool operator!=(const ConstIndexIterator & oth) const {return !this->operator==(oth);}
        const string GetWord() const {
            return mIt->first;
        }
        void GetPosSym(vuc &arrPos, vuc &arrSym, int &cnt) const {
            const Record &rec = mIt->second;
            if (rec.mIsExt) {
                unsigned int pos = rec.GetPos();
                assert(pos < mpIndex->mExtPos.size());
                arrPos =  mpIndex->mExtPos[pos];
                arrSym = mpIndex->mExtSym[pos];
                cnt = arrPos.size();
            } else {
                cnt  = rec.mCnt;
                arrPos.resize(cnt);
                arrSym.resize(cnt);
                for (int i = 0; i < cnt; ++i) {
                   arrPos[i] = rec.mPos[i]; 
                   arrSym[i] = rec.mSym[i];
                }
            }
        }
    };
    COneDelIndex() {allocMemBlock(DefaultSize);RecalcIterators();}
    ~COneDelIndex() {
        for(list<MemBlock *>::iterator it = mMemBlocks.begin(); it != mMemBlocks.end(); ++it) {
            delete *it;
        }
    }
    void insert(const char *p);
    void search(string &str, bool hasTransp, int *numCmp, StringList *resList, bool exOnly = false);
    int  GetDictSize() {return mIndex.size();}
protected:
    ConstIndexIterator mBegin, mEnd;
    friend class ConstIterIndex;
    hmapci             mIndex;
    vector<vuc>        mExtPos;
    vector<vuc>        mExtSym;
    list<MemBlock *>   mMemBlocks;
    MemBlock           *mpLastMemBlock;
    void allocMemBlock(int sz) {
        mpLastMemBlock = new MemBlock(sz);
        mMemBlocks.insert(mMemBlocks.begin(), mpLastMemBlock);
    }
    char *strDup(const string &str) {
        assert(mpLastMemBlock); 
        unsigned int sz = str.size() + 1;
        int freeSz = mpLastMemBlock->mSize - mpLastMemBlock->mFreePos;
        if (freeSz < sz) {
            allocMemBlock(max(DefaultSize, sz));
        }
        char *p = mpLastMemBlock->mpBeg + mpLastMemBlock->mFreePos;
        strcpy(p, str.c_str());
        mpLastMemBlock->mFreePos += sz;
        return p;
    }
    void insertInternal(const char *p, unsigned char pos, unsigned char sym);
    void searchInternal(const char* p, bool hasTransp, unsigned char pos, unsigned char sym, int *numCmp, StringList *resList, bool exOnly = false);
    void RecalcIterators() {
        mBegin = ConstIndexIterator(this, mIndex.begin());
        mEnd = ConstIndexIterator(this, mIndex.end());
    }
public:
    const ConstIndexIterator & begin() const {return mBegin;}
    const ConstIndexIterator & end()   const {return mEnd;}
};


#endif
