#ifndef ALTERN_REDALPHA_INDEX_H
#define ALTERN_REDALPHA_INDEX_H
#include "interface.h"
#include <fcntl.h>
#include <iostream>

#include "../SignHash/calchash.h"
#include "../SignHash/rstat.h"
#include "../Common/hash.h"
#include "langcase.h"
#include "alphabet.h"
#include "myexcept.h"

#define REDALPHA_NOT_CHAR       '@' // should be < REDALPHA_FIRST_CHAR
#define REDALPHA_FIRST_CHAR     'A'
// Don't make it more than 256 - REDALPHA_FIRST_CHAR
#define MAX_REDALPHA_SIZE       64

using namespace std;

// Neighborhood generation with reduced alphabet

struct CHashElem {
    unsigned    mQty;
    string      mStrings; // concatenated strings of the same length, no separators are used
    CHashElem() : mQty(0) {}
    bool operator<(const CHashElem& That) const {
        if (mStrings != That.mStrings) {
            return mStrings < That.mStrings;
        }
        return mQty < That.mQty;
    }
};

typedef hash_map<string, CHashElem>     THashData;

struct THashDataIteratorWrap {
    THashData::iterator     mIt;
    THashDataIteratorWrap(const THashData::iterator& it) : mIt(it) {}
    bool operator<(const THashDataIteratorWrap& That) const {
        if (mIt->first != That.mIt->first) {
            mIt->first < That.mIt->first;
        }
        
        return mIt->second < That.mIt->second;
    }
};

class CAlternGenRedAlpha : public CInterface {
  public:
     CAlternGenRedAlpha (bool hasTransp, EAlphabet AlphaType);
    ~CAlternGenRedAlpha () {}

    bool CreateInMemIndex(const char *inputFile,
                             int insertCount, int *dcnt, 
                             bool silent = true);
    //=========================================================================
    bool Insert(const char *str);
    bool Search(const char *str, int kerr, int *numComp,
                 StringList * resList, bool silent = true);
    //=========================================================================
    bool SearchInDB(const char *str, int kerr, int *numComp,
                    StringList * resList, bool silent = true) {
        cerr << "SearchInDB is not supported" << endl;
        return false;
    }
    //=========================================================================
    bool DumpDB (const char *base_name);
    bool OpenDB (const char *base_name, bool create = false);
    bool CloseDB ();
#ifdef _DEBUG
    void Print ();
#endif
    static string HeaderFileName(const string& BaseName) { return BaseName + ".hdr"; }
    static string HashFileName(const string& BaseName) { return BaseName + ".hash";}
    static string DictFileName(const string& BaseName) { return BaseName + ".dict";}
  protected:
    bool            mbOpened;
    int             mFdHash;
    int             mFdParams;
    int             mFdDict;
    HMap            mAlphaMap;
    const char*     mpABC;
    unsigned        mAlphaSize;
    unsigned char   maRedABC[MAX_REDALPHA_SIZE + 1];
    unsigned        mRedAlphaSize;
    THashData       mRedHash;

    char*           mpDict;
    size_t*         mpHash; 
    CHashParams*    mpParams;

    void                    InsertInternal(const string& s);
    int                     GetRedAlphaSize();

    inline unsigned char    ToRedChar(char c) {
        int indx = mAlphaMap[c];
        if (indx < 1 || indx > mRedAlphaSize) {
            return REDALPHA_NOT_CHAR;
        }
        return maRedABC[indx - 1];
    }
    inline void             ToRedStr(string& s) {
        for (unsigned i = 0; i < s.size(); ++i) { s[i] = ToRedChar(s[i]); }
    }
    void ExpandAndSearch (unsigned start, int& NumComp, 
                          string& sRedPat, 
                          string& sOrigPat, 
                          unsigned kErr, 
                          unsigned kErrOrig, 
                          StringList* pResList);

    bool mapData();
    void unmapData();
};

template <>
inline bool SearchDbNotSupported<CAlternGenRedAlpha>() {
    return true;
}

#endif
