#include <iostream>
#include <string>

#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <fcntl.h>
#include "../Interface/interface.h"

#include "utils.h"
#include "filesize.h"
#include <set>

using namespace std;

typedef set<off_t> seto;

int clear(const char *pName)
{
    FILE *f0 = fopen (DEV_CLDEV, "w");
    if (!f0) {
        Perror ("Cannot open cldev");
        return 0;
    }
    // Write to the device to clear caches
    if (fputs(pName, f0) < 0) {
        Perror ("Cannot write to cldev");
        return 0;
    }

    if (fclose (f0)) {
        Perror ("Cannot close cldev");
        return 0;
    }
    return 1;
}

int usage(const char *prog) 
{
    cerr << "Usage: \n" 
            " Sequential search benchmark: " << prog << " -s <filename>\n" <<
            " Random access benchmark: " << prog << " -r -n num <filename>\n";
    return 1;
}

int main(int argc, char *argv[]) 
{
    const char *pFileName = NULL;
    bool       seq = false;
    int        num = 0;
    int        MinSeqNum = 8 * 1024;

    if (!strcmp(argv[1], "-s")) {
        if (argc != 3) return usage(argv[0]);
        seq = true;
        pFileName = argv[2];
    } else if (argc == 5 && !strcmp(argv[1], "-r") && !strcmp(argv[2], "-n") &&
               (num=atoi(argv[3]))) {
        pFileName = argv[4];
    } else {return usage(argv[0]);}

    off_t fsize = FileSize(pFileName);
    int   bsize = 4096;
    off_t bnum  = fsize / bsize;

    if (fsize < 0) {
        cerr << "Cannot find file '" << pFileName << "'" << endl;
        return 2;
    }

    if (fsize < 0) {
        cerr << "Cannot find file '" << pFileName << "'" << endl;
        return 3;
    }

    cout << "Test type: " << ((seq) ? "sequential read":"random access") << endl;
    cout << "File name: " << pFileName << endl;
    cout << "File size: " << fsize << endl;

    if (seq) {
        num = bnum;
    }
    cout << "Number of block reads: " << num << endl;
    
    if ((seq && num < MinSeqNum) || (!seq && 10 * bnum < num)) {
        cerr << "File '" << pFileName << "' is too small, the number of blocks is only: " << bnum << endl;
        return 4;
    }

    if (!clear(pFileName)) {
        cerr << "Cannot purge caches of '" << pFileName << "'" << endl;
        return 5;
    }

    int fd = open(pFileName, O_RDONLY);

    if (fd < 0) {
        Perror("Cannot open '" +string(pFileName) + "'");
        return 5;
    }

    srand(time(NULL));

    seto  bset;
    off_t pos = 0;
    char  buf[bsize];

    DECLARE_STATIC_COUNTER(t1)

    START_COUNTER(t1)

    for (int i = 0; i < num; ++i) {
        if (!seq) {
            do {
                pos = (rand() + RAND_MAX * rand()) % bnum;
                if (pos < 0) pos = -pos;
                pos *= bsize;
            } while (bset.find(pos) != bset.end());
            bset.insert(pos);

            if (lseek(fd, pos, SEEK_SET) != pos) {
                Perror("Cannot lseek");
                return 6;
            }
        }
        if (read(fd, buf, bsize) != bsize) {
            Perror("Cannot read");
            return 7;
        }
    }

    STOP_COUNTER(t1)

    off_t  total = num * bsize;
    off_t  diff = GET_COUNTER(t1);

    cout << "Total read: " << total << " bytes\n";
    cout << "Total running time (ms): " << diff / 1000.0 << endl;
    cout << "Speed: " << (1e6 * total / diff / (1024 * 1024)) << " Mbyte/sec\n";

    return 0;
};
