#include <stdexcept>
#include <iostream>

#include "trieex.h"
#include "langcase.h"

static unsigned memSize = 0;

unsigned CTrieExact::GetUsedMem() { return memSize; }

void 
CTrieExact::allocateNodesFull(int N)
{
    unsigned int sz = N * m_nodeSizeFull;
    char *p = (char *)malloc(sz);

    memset(p, 0, sz);
    if (!p) {
        throw bad_alloc();
    }
    m_memBlocks.push_back(p);
    for (int i = 0; i < N; ++i) {
        m_freeNodesFull.insert(m_freeNodesFull.end(), (CTrieNodeFull *)p);
        p += m_nodeSizeFull;
    }
    memSize += N * m_nodeSizeFull;
    //cout << "Number of bytes allocated: " << memSize << endl;
}

void 
CTrieExact::allocateNodesShort(int N)
{ 
    unsigned int sz = N * m_nodeSizeShort;
    char *p = (char *)malloc(sz);
    memset(p, 0, sz);
    if (!p) {
        throw bad_alloc();
    }
    m_memBlocks.push_back(p);
    for (int i = 0; i < N; ++i) {
        m_freeNodesShort.insert(m_freeNodesShort.end(), (CTrieNodeShort *)p);
        p += m_nodeSizeShort;
    }
    memSize += N * m_nodeSizeShort;
    //cout << "Number of bytes allocated: " << memSize << endl;
}

CTrieExact::CTrieExact(const char *pABC)
{
    m_abcSize = 0; 
    // Prepare the map first
    memset(m_abcMap, 0, sizeof m_abcMap);
    strncpy(m_ABC, pABC, sizeof m_ABC);
    for (const char *p = pABC; *p; ++p) {
        unsigned char c1 = tolower_eng_rus(*p);
        unsigned char c2 = toupper_eng_rus(*p);
        m_abcMap[c1] = m_abcMap[c2] = ++m_abcSize;
    }
    m_abcMap[EOS] = ++m_abcSize;
    m_nodeSizeFull = sizeof(CTrieNodeFull) + sizeof(CTrieNodeBase *) * (m_abcSize + 1);
    m_nodeSizeShort = sizeof(CTrieNodeShort);

    m_pRoot = getNewNodeShort(0);
}

CTrieExact::CTrieNodeFull*
CTrieExact::getNewNodeFull(unsigned char c)
{
    list<CTrieNodeFull *>::iterator it = m_freeNodesFull.begin();
    if (it == m_freeNodesFull.end()) {
        allocateNodesFull(100);
        it = m_freeNodesFull.begin();
        assert(it != m_freeNodesFull.end());
    }

    CTrieNodeFull *p = *it;
    p->mIsFull = true;
    p->mSym = c;
    m_freeNodesFull.erase(it);
    return p;
}

CTrieExact::CTrieNodeShort*
CTrieExact::getNewNodeShort(unsigned char c)
{
    list<CTrieNodeShort *>::iterator it = m_freeNodesShort.begin();
    if (it == m_freeNodesShort.end()) {
        allocateNodesShort(100);
        it = m_freeNodesShort.begin();
        assert(it != m_freeNodesShort.end());
    }

    CTrieNodeShort *p = *it;
    p->mSym = c;
    m_freeNodesShort.erase(it);
    return p;
}

CTrieExact::~CTrieExact()
{
    list<void *>::iterator endIt = m_memBlocks.end();
    for (list<void *>::iterator it = m_memBlocks.begin();it != endIt; ++it) {
        free(*it);
    }
}

