#ifndef _INDEX
#define _INDEX


#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <string>
#include <iostream>
#include <fcntl.h>
#include <errno.h>
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <set>
#include <list>

#include "interface.h"
#include "erratatrie.h"

using namespace std;

class CColeEtAlIndex:public CInterface {

  public:
    CColeEtAlIndex (bool hasTransp, EAlphabet AlphaType);
    virtual ~CColeEtAlIndex ();
    virtual bool Insert (const char *str);
    virtual bool Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent = true);
    virtual bool SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent = true) {
        cerr << "SearchInDB is not supported";
        return true;
    }
    virtual bool DumpDB (const char *base_name);
    virtual bool OpenDB (const char *base_name, bool create = false);
    virtual bool CloseDB ();

    virtual bool CreateInMemIndex(const char *inputFile, int insertCount, int *dcnt, bool silent = true);

    static std::string TrieOffsetsName(const std::string& base_name) {
        return base_name + "_troff.img";
    }
    static std::string TrieName(const std::string& base_name) {
        return base_name + "_tr.img";
    }
    static std::string AddNodeInfoName(const std::string& base_name) {
        return base_name + "_addndnfo.img";
    }
    static std::string CombKetOffsetsName(const std::string& base_name) {
        return base_name + "_cmbktoff.img";
    }
    static std::string CombKetName(const std::string& base_name) {
        return base_name + "_cmbkt.img";
    }
  protected:
    bool            mbOpened;
    CErrataTree     mTrie;

    int             mFDTrieOffsets;
    int             mFDTrie;
    int             mFDAddNodeInfo;
    int             mFDCombKetOffsets;
    int             mFDCombKets;
};

template <>
inline int MaxTestKerr<CColeEtAlIndex>() {
    //return 2;
    return 1;
}

#endif
