#ifndef NODE_ALLOC_H
#define NODE_ALLOC_H

#include "nodehelper.h"
#include <vector>

#define RESERVE_THRESHOLD 1000

struct CNodeAllocator {

    // (Alloc|Free)MemSimple functions, just allocates memory
    // and update mMemUsed.
    static CNode* AllocMemSimple(size_t NodeSize);
    static void FreeMemSimple(CNode *p, size_t NodeSize);

    // (Alloc|Free)Mem additionaly handle a list of free node indices
    static CNode* AllocMem(size_t NodeSize, NodeIndex &Index);
    static void FreeMem(CNode *p, size_t NodeSize, NodeIndex Index);

    // Higher level node allocation, reallocation, and disposal functions
    static NodeIndex AllocNode(SmallCounter LabelLen, const char* pLabel, SmallCounter ChildQty, NodeIndex *pChildren = NULL);
    static void      FreeNode(NodeIndex Index);
    static void      ReallocNode(NodeIndex Index, SmallCounter LabelLen, SmallCounter ChildQty);

    static size_t GetTotalMemUsed() {
        return mMemUsed + GetNodeQty() * sizeof(mpNodePtr[0]);
    }
    static size_t GetMemUsedByNodes() {
        return mMemUsed;
    }
    static size_t GetNodeQty() {
        return mpNodePtr.size() - mFreeIndices.size();
    }
    static NodeIndex GetMaxNodeId() {
        return static_cast<NodeIndex>(mpNodePtr.size() - 1);
    }

    static std::vector<CNode*>     mpNodePtr;     // This vector allows to obtain node ptr  by node index.
    static void PurgeAllMem() {
        for (size_t i = 0; i < mpNodePtr.size(); ++i) {
            FreeNode(i); // If mpNodePtr[i] is empty, nothing happens.
        }
        mFreeIndices.clear();
    }
private:
    static size_t                  mMemUsed;
    static std::vector<NodeIndex>  mFreeIndices;  // This vector recycles indices of deleted nodes
};

#endif
