#include "db_init.h"

void db_close(DB_ENV *& dbenv)
{
   if (dbenv) {
       (void)dbenv->close(dbenv, 0);
       //free(dbenv); // it is incorrect to call free here it is already done by dbenv->close
       dbenv = NULL;
   }
}

DB_ENV *db_init (char *home, int cacheSize, FILE * errfp)
{
    DB_ENV *dbenv;
    int ret;

    /* Rely on calloc to initialize the structure. */
    if ((ret = db_env_create (&dbenv, 0)) != 0) {
        cerr << db_strerror (ret) << endl;
        return NULL;
    }
    dbenv->set_errfile (dbenv, errfp);
    dbenv->set_errpfx (dbenv, "");

    if ((ret = dbenv->set_cachesize (dbenv, 0, cacheSize, 0)) != 0) {
        dbenv->close (dbenv, 0);
        cerr << "Error setting cache size :" << db_strerror (ret) << endl;
        return NULL;
    }

    /* Databases are in a subdirectory. */
    (void) dbenv->set_data_dir (dbenv, (home != NULL) ? home : "");


    if ((ret =
         dbenv->open (dbenv, home, DB_CREATE | DB_INIT_MPOOL, 0)) != 0) {
        cerr << db_strerror (errno);
        return NULL;
    }

    return (dbenv);
}
