#ifndef MEM_ALLOC_H
#define MEM_ALLOC_H

#include <stack>
#include <list>
#include <vector>

// Allocator is desigined for
// 1. Avoid memory fragmenation
// 2. Make allocation of the fixed-size chunks really fast. Only one new call for several allocation (this 
// number is controllable by the second constructor parameter (1000 by default)
class CFastAllocator {
private:
    std::vector<char *>       mAllocatedBlocks;
#ifdef USE_LIST
    std::list<char *>         mFreeChunks;
#else
    std::stack<char *>        mFreeChunks;
#endif
    unsigned int              mChunkSize;
    unsigned int              mReserveCnt;
    unsigned int              mBlockSize;
public:
    CFastAllocator(unsigned int chunkSize, unsigned int reserveCnt = 10000) 
                   : mChunkSize(chunkSize), mReserveCnt(reserveCnt), mBlockSize(reserveCnt * chunkSize){}
    void *  allocChunk(); // throws std::bad_alloc
    void    freeChunk(void *);
    virtual ~CFastAllocator();
    unsigned int getAllocatedMem(){return mBlockSize * mAllocatedBlocks.size();}
};

#endif
