#ifndef _BIT_MACRO
#define _BIT_MACRO

#include <sys/types.h>

typedef int32_t int32;
typedef u_int32_t uint32;

/* !!! Note StartInput must be written always after StartOutput!    */

/* This is a universal declaration for the input and output buffers */
/* Although for each buffer output or input its own DeclareNamedBuf */
/* must be specified with unique Name parameter                     */

#define DeclareOutNamedBuf(_BufSize,Name)	 		\
	int32	 	Name##BufSize = (_BufSize);  		\
	char 		Name##buf[(_BufSize)]; 			\
	int		Name##fd;
	
#define DeclareInNamedBuf(_BufSize,Name) 			\
	int32	 	Name##BufSize = (_BufSize);  		\
	char 		Name##buf[(_BufSize)]; 			\
 	int32		Name##rb;				\
	int		Name##fd;

/* Here go macros for outputing a bit                               */
#define StartOutput(_fd,Name)					\
	register char	Name##bit	=1;			\
	register int    Name##BytePos   =0;			\
	register char	Name##cchar	=0;			\
	Name##fd	=_fd;					\


/* An exit label Name##Error must exists!!! */
#define PutBit(v,Name)						\
 if (!Name##bit) /* After 128 was shifted by << */		\
 {								\
  Name##buf[Name##BytePos++]=Name##cchar;			\
  Name##bit   = 1;						\
  Name##cchar = 0;						\
  if (Name##BytePos >= Name##BufSize )				\
  {								\
   if (write(Name##fd,Name##buf,Name##BufSize) < Name##BufSize) \
	goto Name##Error; 					\
   Name##BytePos = 0;						\
  }								\
 }                                                              \
 if ((v)) Name##cchar |= Name##bit;				\
 Name##bit <<= 1;    						

/* An exit label Name##Error must exists!!! */
#define PutByte(v,Name)						\
{								\
  Name##buf[Name##BytePos++]=v;					\
  if (Name##BytePos >= Name##BufSize )				\
  {								\
   if (write(Name##fd,Name##buf,Name##BufSize) < Name##BufSize) \
	goto Name##Error; 					\
   Name##BytePos = 0;						\
  }								\
}	

#define FinishOutput(Name)					\
 if (Name##BytePos || Name##bit!=1)				\
 {								\
  Name##buf[Name##BytePos++]=Name##cchar; 			\
  if (write(Name##fd,Name##buf,Name##BytePos) < Name##BytePos) 	\
	goto Name##Error;					\
 }								\
 Name##BytePos=0;                                               \
 Name##bit=1;

/* Here go macros for inputing a bit                               */
#define StartInput(_fd,Name) 					\
 char	Name##bit	=1;				        \
 int32  Name##BytePos	=0;				        \
 Name##fd = _fd;						\
 Name##rb = read (Name##fd,Name##buf,Name##BufSize);		\
 int32	Name##ReadLen = Name##rb;			        \
 char	Name##cchar=Name##buf[Name##BytePos];		        \
 if (Name##rb < 0)						\
	{goto Name##Error;} 	\
 if (!Name##rb)                                                	\
	{goto Name##Error;} 

/* For this macro the label named NameFinish must exist!        */
/* When the stream is exausted, the control is passed there     */
#define SkipBitsInput(Name)					\
{int v;								\
 for(;Name##bit<8;) {GetBit(v,Name);}                           \
}


/* For this macro the label named NameFinish must exist!        */
/* When the stream is exausted, the control is passed there     */
#define GetBit(v,Name)						\
 if (Name##bit> 8)						\
 {								\
  Name##BytePos++;						\
  if (Name##BytePos < Name##ReadLen )				\
  {								\
   Name##bit  = 1;						\
   Name##cchar = Name##buf[Name##BytePos];		   	\
  }								\
  else								\
  {								\
   Name##rb = read (Name##fd,Name##buf,Name##BufSize);		\
   if (Name##rb < 0)						\
	{goto Name##Error;}  \
   if (!Name##rb)    goto Name##Finish;				\
   Name##ReadLen = Name##rb;					\
   Name##bit     = 1;                                           \
   Name##BytePos = 0;                                           \
   Name##cchar   = Name##buf[Name##BytePos];                    \
  }                                                             \
 }                                                              \
 (v) = Name##cchar & 0x1;                                         \
 Name##cchar >>= 1;                                             \
 Name##bit++;

#define GetByte(v,Name) \
  Name##BytePos++;						\
  if (Name##BytePos < Name##ReadLen )				\
  {								\
   Name##bit  = 1;						\
   Name##cchar = Name##buf[Name##BytePos];		   	\
  }								\
  else								\
  {								\
   Name##rb = read (Name##fd,Name##buf,Name##BufSize);		\
   if (Name##rb < 0)						\
	{Error.SetError(goto Name##Error;}  \
   if (!Name##rb)    goto Name##Finish;				\
   Name##ReadLen = Name##rb;					\
   Name##bit     = 1;                                           \
   Name##BytePos = 0;                                           \
   Name##cchar   = Name##buf[Name##BytePos];                    \
  }                                                             \
 }                                                              \
 (v) = Name##cchar

#endif
