library(graphics)
library(stats)
library(Hmisc)

Usage <- function(err) {
    print(paste("Error:", err))
    stop("Usage: <OutputFile> <InputFile1> <transp1> <InputFile2> <transp2> ...")
}

MaxIgnoreNA <- function(x) {
    res <- NA
    Len   <- length(x)
    Index <- 1
    while (Index <= Len) {
        if (!is.na(x[Index])) {
            if (is.na(res)) {
                res <- x[Index]
            } else if (res < x[Index]) {
                res <- x[Index]
            }
        }
        Index <- Index + 1;
    }
    return (res);
}

args <- commandArgs(TRUE)

if ((length(args) %% 2) == 0) {
   Usage("Even number of arguments!")
}

OutputFile      <- args[1]

InputFiles <- c()
HasTransp  <- c()
iArg <- 2
NGraph <- (length(args) - 1)/2;

while (iArg < length(args)) {
    InputFiles <- c(InputFiles, args[iArg])
    HasTransp  <- c(HasTransp , args[iArg + 1])
    iArg <- iArg + 2
}

#print(InputFiles)
#print(HasTransp)

YMax <- 0

for (i in 1:2) {

    source(InputFiles[i])

    YMax <- MaxIgnoreNA(
                c(YMax,
                MaxIgnoreNA(NRgrep),
                MaxIgnoreNA(agrep),
                MaxIgnoreNA(MyersHyyro),
                MaxIgnoreNA(DFA),
                MaxIgnoreNA(LazyDFA),
                MaxIgnoreNA(magrep1),
                MaxIgnoreNA(magrep2))
            )
}

#YLim1      <- YMax  + 100
#YLimHeight <- YMax  + 200
YLim1      <- YMax  + 5
YLimHeight <- YMax  + 10

postscript(OutputFile, family = "Times", encoding = "TeXtext.enc", height= 6, width= 6 * NGraph, paper = "special")
par(mfrow = c(1, NGraph) )
par(cex = 0.9)

par(font.main = 1) # Main title font is plan instead of bold
par(font.sub = 1) # Sub title font is plan instead of bold

XLim  <- c(0.75, 3.25)
YLabel <- "Time (ms)"

k       <- c(1,2,3)
DY      <- 100
NY      <- round(YLim1  / DY, 0)
yax     <- DY * 0:NY
leg.txt <- c("NR-grep","magrep1","magrep2","DFA-estimated", "Myers-Hyyro", "Lazy DFA", "agrep")

Color_agrep     <- "darkmagenta"
Color_NRgrep    <- "darkviolet"
Color_magrep1   <- "green4"
Color_magrep2   <- "darkorange4"
Color_DFA       <- "red"
Color_LazyDFA   <- "black"
Color_Myers     <- "blue"

PlotCommonMeth <- function(HasTransp) {
    par(new=FALSE)
    plot(k,NRgrep,"b",col=Color_NRgrep,pch=1,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
    par(new=T)
    plot(k,magrep1,"b",col=Color_magrep1,pch=2,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
    par(new=T)
    plot(k,magrep2,"b",col=Color_magrep2,pch=3,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
    par(new=T)
    plot(k,DFA,"b",col=Color_DFA,pch=4,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
    par(new=T)
    plot(k,MyersHyyro,"b",col=Color_Myers,pch=5,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)

    if (HasTransp == 0) {
        par(new=T)
        plot(k,LazyDFA,"b",col=Color_LazyDFA,pch=6,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
        par(new=T)
        plot(k,agrep,"b",col=Color_agrep,pch=7,axes=FALSE, xlim = XLim,ylim=c(0,YLimHeight),xlab="k",ylab=YLabel)
    }
}

Colors <- c(Color_NRgrep, Color_magrep1, Color_magrep2, Color_DFA, Color_Myers, Color_LazyDFA, Color_agrep)

for (iGraph in 1:NGraph) {
    source(InputFiles[iGraph])

    par(mai=c(1.5,1.1,0.25,0.1))
    if (iGraph > 1) {
        par(new=F)
    }

    PlotCommonMeth(HasTransp[iGraph])

    axis(1, 1:3, 1:3, 1:3)
    axis(2, at=yax, label=yax, line=yax)
    box()

    MethQty = 7

    if (HasTransp[iGraph] != 0) {
        MethQty = 5
    }
    par(new=T)
    legend(0.75,YLimHeight,leg.txt[1:MethQty],col=Colors[1:MethQty],pch=1:MethQty,lty=1, ncol =2)

}
dev.off()

