#include "freqvector.h"

// Frequency vector map implementation

CFreqVectorMap::CFreqVectorMap ():m_MapSize (0)
{
    memset (m_MapArray, -1, 256 * sizeof (int16_t));
}

CFreqVectorMap::
CFreqVectorMap (const CFreqVectorMap & smap):m_MapSize (smap.m_MapSize)
{
    memcpy (m_MapArray, smap.m_MapArray, 256 * sizeof (int16_t));
}

const CFreqVectorMap &
    CFreqVectorMap::operator= (const CFreqVectorMap & smap)
{
    m_MapSize = smap.m_MapSize;
    memcpy (m_MapArray, smap.m_MapArray, 256 * sizeof (int16_t));
    return (*this);
}

int
 CFreqVectorMap::AddChar (unsigned char c)
// 0 - Successfull
// 1 - char already exists
{
    if (m_MapArray[c] > 0)
        return 1;
    else
        m_MapArray[c] = m_MapSize++;
}

int CFreqVectorMap::AddCharAlias (unsigned char c, unsigned char calias)
{
    if (m_MapArray[c] < 0)
        return 0;
    m_MapArray[calias] = m_MapArray[c];
    return 1;
}

u_int8_t CFreqVectorMap::Size () const
{
    return m_MapSize;
}

// Freq vector implementation

u_int8_t CFreqVector::Size () const
{
    return m_EndIndex - m_StartIndex + 1;
}


void CFreqVector::InitMem ()
{
    m_Vector = NULL;
    if (Size () > 0)
        m_Vector = new u_int8_t[Size ()];
}

void CFreqVector::FreeMem ()
{
     delete [] m_Vector;
    m_Vector = NULL;
}

CFreqVector::CFreqVector (const CFreqVector & v) : m_StartIndex (v.m_StartIndex), m_EndIndex (v.m_EndIndex), m_map (v.m_map)
{
    InitMem ();
    if (Size () > 0)
        memcpy (m_Vector, v.m_Vector, Size ());
}

CFreqVector::~CFreqVector ()
{
    FreeMem ();
}

CFreqVector CFreqVector::SubVector (int si, int ei) const
{
    if (si < m_StartIndex)
        si = m_StartIndex;
    if (ei > m_EndIndex)
        ei = m_EndIndex;
    if (si > ei)
        si = ei;
    // Now m_StartIndex<=si<=ei<=m_EndIndex
    //CFreqVector *pv =
        //new CFreqVector (m_map, si, ei, &m_Vector[si - m_StartIndex]);
    //return *pv;
    return CFreqVector (m_map, si, ei, &m_Vector[si - m_StartIndex]);
}

CFreqVector::CFreqVector (int StartIndex, int EndIndex):m_map (NULL), m_StartIndex (StartIndex), m_EndIndex (EndIndex)
{
    ;
}


CFreqVector::CFreqVector
    (const CFreqVectorMap * map, int StartIndex,
     int EndIndex): m_StartIndex (StartIndex), m_EndIndex (EndIndex), m_map (map)
{
    InitMem ();
    if (Size () > 0)
        memset (m_Vector, 0, Size ());
}


CFreqVector::CFreqVector
    (const CFreqVectorMap * map, int StartIndex, int EndIndex,
     u_int8_t * pExtMem): m_StartIndex (StartIndex),
m_EndIndex (EndIndex), m_map (map)
{
    InitMem ();
    if (Size () > 0) {
        memcpy (m_Vector, pExtMem, Size ());
    }
}

const CFreqVector & CFreqVector::operator= (const CFreqVector & v)
{
    if (this != &v) {
        m_EndIndex = v.m_EndIndex;
        m_StartIndex = v.m_StartIndex;
        m_map = v.m_map;
        FreeMem ();
        InitMem ();
        if (Size () > 0)
            memcpy (m_Vector, v.m_Vector, Size ());
    }
    return (*this);
}

u_int8_t CFreqVector::StartIndex () const
{
    return m_StartIndex;
}

u_int8_t CFreqVector::EndIndex () const
{
    return m_EndIndex;
}


int CFreqVector::CopyToMem (void *mem) const
{
    int s = Size ();
    if (!s)
        return 0;
    memcpy (mem, (void *) m_Vector, s);
    return s;
}

void CFreqVector::InitFromString (const char *str)
{
    if (!m_map) {
        std::cerr << "Initilization with empty map not allowed!\n";
        assert (0);
    }
    for (const char *ps = str; *ps; ps++) {
        unsigned char c = (unsigned char) *ps;
        int num = m_map->CharNum (c) - m_StartIndex;
        if (num >= 0 && num < Size ())
            m_Vector[num]++;
    }
}

#ifdef _DEBUG
void CFreqVector::Print () const
{
    for (int i = 0; i < Size (); i++)
        cout << int (m_Vector[i]) << " ";
    cout << "\n";
}
#endif


int
Diff (const CFreqVector & v1, const CFreqVector & v2, int *posdist,
      int *negdist)
{
    int ToIndex = ((v1.Size () < v2.Size ())? v1.Size () : v2.Size ()) - 1;
    *posdist = *negdist = 0;
    int iFirstDiff = -1;
    for (int i = 0; i <= ToIndex; i++) {
        if (v1.m_Vector[i] != v2.m_Vector[i]) {
            if (iFirstDiff < 0)
                iFirstDiff = i;
            int diff = v1.m_Vector[i] - v2.m_Vector[i];
            if (diff > 0)
                *posdist += diff;
            else
                *negdist -= diff;
        }
    }
    if (iFirstDiff < 0)
        if (v1.Size () > v2.Size ())
            iFirstDiff = v2.Size ();
        else if (v2.Size () > v1.Size ())
            iFirstDiff = v1.Size ();

    return iFirstDiff;

}

int CompFunc (const CFreqVector & v1, const CFreqVector & v2)
{
    int i;
    int ToIndex = ((v1.Size () < v2.Size ())? v1.Size () : v2.Size ()) - 1;
    for (i = 0; i <= ToIndex && v1.m_Vector[i] == v2.m_Vector[i]; i++);
    if (i <= ToIndex)
        return (int (v2.m_Vector[i]) - int (v1.m_Vector[i]));
    return v2.Size () - v1.Size ();

}
