#ifndef _FREQVECTOR_H
#define _FREQVECTOR_H

#include <sys/types.h>
#include <string.h>
#include <iostream>
#include <assert.h>

// This class contains information on mapping letters
// to vector indexes
class CFreqVectorMap {
  protected:
    u_int8_t m_MapSize;
    int16_t m_MapArray[256];
  public:
     CFreqVectorMap ();
     CFreqVectorMap (const CFreqVectorMap &);
    const CFreqVectorMap & operator= (const CFreqVectorMap & smap);
    int AddChar (unsigned char c);  // 0 - Successfull
    // 1 - char already exists
    int AddCharAlias (unsigned char c, unsigned char calias);
    // This function returns the position of the character in the frequency vector
    // It returns -1 if the character is not "mapped"
    inline int CharNum (unsigned char) const;
    //
    u_int8_t Size () const;
};

int CFreqVectorMap::CharNum (unsigned char c) const
{
    return m_MapArray[c];
}

// The frequency sub-vector class
class CFreqVector {
  protected:
    // The sub-vector itself
    u_int8_t * m_Vector;
    // The interval of indexes of the frequency vector 
    // which sub-vector contains
    u_int8_t m_StartIndex, m_EndIndex;
    const CFreqVectorMap * m_map;

    void InitMem ();
    void FreeMem ();
  public:
    //=== Constructors, destructor, equation operator
     CFreqVector (int StartIndex = 0, int EndIndex = -1);   // Creates an empty vector
     CFreqVector (const CFreqVector &);
     CFreqVector (const CFreqVectorMap * map, int StartIndex,
                  int EndIndex);
     CFreqVector (const CFreqVectorMap * map, int StartIndex, int EndIndex,
                  u_int8_t *);
    ~CFreqVector ();
    const CFreqVector & operator= (const CFreqVector &);
    //==== SubVector
    // for efficiency returned subvector is created with m_ExternalMem flag
    // freq. array will be copied only when operator= or copy constructor
    // is used
    CFreqVector SubVector (int si, int ei) const;

    //=== 
    u_int8_t StartIndex () const;
    u_int8_t EndIndex () const;
    u_int8_t Size () const;
    //===
    // Copies only the vector content, returns the number of bytes written
    int CopyToMem (void *mem) const;
    //==== Initialization of the sub-vector according to the contents 
    // of the string 
    void InitFromString (const char *str);
#ifdef _DEBUG
    void Print () const;
#endif
    //==== Comparisons
    friend int CompFunc (const CFreqVector & v1, const CFreqVector & v2);
    /* 
       Diff returns the position of the first unequal digits
       it returns -1 in case of equal vectors
       Positive distance is the sum of digit differences
       where the second argument is less than the first
       Negative distance is the sum of the digit differences
       where the first argument is less than the second
     */
    friend int Diff (const CFreqVector & v1, const CFreqVector & v2,
                     int *posdist, int *negdist);
};

int CompFunc (const CFreqVector & v1, const CFreqVector & v2);
int Diff (const CFreqVector & v1, const CFreqVector & v2, int *posdist,
          int *negdist);

#endif
