#include "index.h"
#include "myagrep.h"
#include "langcase.h"

CFreqTrieIndex::CFreqTrieIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false)
{
    const char* pABC = GetAlphabet(mAlphaType);
    int charNum = strlen(pABC);

    char letSmall[charNum];
    char letCaps[charNum];

    bool printAddStat = GetEnvSwitch("PRINT_ADD_STAT");

    int VectorLength = _VECTOR_LENGTH;
    GetEnv("VECTOR_LENGTH", VectorLength);

    if (printAddStat) {
        cout << "Using vector length: " << VectorLength << endl;
    }

    for (unsigned int i = 0; i < charNum; ++i) {
        letSmall[i] = tolower_eng_rus(pABC[i]);
        letCaps[i] = toupper_eng_rus(pABC[i]);
    }
    
    for (unsigned int i = 0; i < charNum; ++i) {
        int pos = i % VectorLength;

        if (i < VectorLength) {
            vMap.AddChar(letSmall[i]);
            vMap.AddCharAlias(letSmall[i], letCaps[i]);
        } else {
            vMap.AddCharAlias(letSmall[pos], letSmall[i]);
            vMap.AddCharAlias(letSmall[pos], letCaps[i]);
        }
    }
}

CFreqTrieIndex::~CFreqTrieIndex ()
{
    if (opened)
        CloseDB ();
}

bool CFreqTrieIndex::Insert (const char *str)
{
    ind.Insert (vMap, str);
    return true;
}

bool CFreqTrieIndex::Search (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent)
{
    myagrep a(str, kerr, mHasTransp);
    (kerr) ? ind.Search (vMap, str, kerr, a, numComp, resList,
                              mHasTransp, silent) : ind.SearchExact (vMap, str,
                                                         a, numComp, resList,
                                                         silent);
    return true;
}

bool CFreqTrieIndex::SearchInDB (const char *str, int kerr, int *numComp,
                                 StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    myagrep a(str, kerr, mHasTransp);
    return ind.SearchInFile (fd, vMap, str, kerr, a, numComp, resList,
                             mHasTransp, silent);
}

bool CFreqTrieIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CFreqTrieIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CFreqTrieIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = ind.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

void CFreqTrieIndex::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
     if (!kerr && type == SearchTypeMem) { // Print once
        ostr << "    The number of vectors: " <<  ind.GetVectorCount() << endl;
     }
}

#ifdef _DEBUG
void CFreqTrieIndex::Print ()
{
    node.Print ();
}
#endif
