#include "index.h"
#include "../SignHash/calchash.h"
#include "../SignHash/rstat.h"
#include "langcase.h"
#include "alphabet.h"

CFreqTrieIndexOptimized::CFreqTrieIndexOptimized (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false)
{
}

CFreqTrieIndexOptimized::~CFreqTrieIndexOptimized ()
{
    if (opened)
        CloseDB ();
}

bool CFreqTrieIndexOptimized::Insert (const char *str)
{
    ind.Insert (vMap, str);
    return true;
}

bool CFreqTrieIndexOptimized::Search (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent)
{
    myagrep a(str, kerr, mHasTransp);
    (kerr != 0) ? ind.Search (vMap, str, kerr, a, numComp, resList,
                              mHasTransp, silent) : ind.SearchExact (vMap, str,
                                                         a, numComp, resList,
                                                         silent);
    return true;
}

bool CFreqTrieIndexOptimized::SearchInDB (const char *str, int kerr, int *numComp,
                                 StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    myagrep a(str, kerr, mHasTransp);
    return ind.SearchInFile (fd, vMap, str, kerr, a, numComp, resList,
                             mHasTransp, silent);
}

bool CFreqTrieIndexOptimized::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CFreqTrieIndexOptimized::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CFreqTrieIndexOptimized::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = ind.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

bool CFreqTrieIndexOptimized::CreateInMemIndex(const char *inputFile,
                                     int insertCount, int *dcnt, bool silent)
{
    HMap      hsign;
    string    s;
    const int n = 3;

    // Gather statistics
    const char *p;
    bool printAddStat = GetEnvSwitch("PRINT_ADD_STAT");

    int MaxDigit = _VECTOR_LENGTH;
    GetEnv("VECTOR_LENGTH", MaxDigit);

    if (printAddStat) {
        cout << "Using vector length: " << MaxDigit << endl;
    }
    vector<char>   used(MaxDigit);

    *dcnt = 0;

    CollectStat st;

    st.CollectFileStat (inputFile);
    if (!CalcHMap(MaxDigit, st.stat (), hsign, printAddStat)) {
        std::cerr << "Hash generation failed!" << std::endl;
        return false;
    }

    // Fill the map
    const char* pABC = GetAlphabet(mAlphaType);
    
    for (const char *pSym = pABC; unsigned char c = *pSym; ++pSym) {
        int pos = hsign[c] - 1;
        if (pos == -1) {
            continue;
        }
        if (pos < -1 || pos >= MaxDigit) {
            cerr << "Internal error: pos >= MaxDigit || pos < 0 (pos=" << pos << ", but VectLen = " << MaxDigit << ", c = '" << c << "')" << endl;
            abort();
        }

        if (!used[pos]) {
            vMap.AddChar (toupper_eng_rus(c));
            vMap.AddCharAlias (c, tolower_eng_rus(c));
            used[pos] = c;
        } else {
            char c1 = used[pos];
            vMap.AddCharAlias (c1, toupper_eng_rus(c));
            vMap.AddCharAlias (c1, tolower_eng_rus(c));
        }
    }
   


    ifstream f (inputFile);
    StringList lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);

        if (insertCount == 1) {
            Insert (s.c_str ());
        } else {
            lst.push_back (s);
            lstCnt++;
            if (lstCnt >= insertCount) {
                this->InsertList (lst);
                lstCnt = 0;
                lst.clear ();
            }
        }
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }
    if (lstCnt)
        this->InsertList (lst);

#if 0
    Print();
#endif
    return true;
}

#ifdef _DEBUG
void CFreqTrieIndexOptimized::Print ()
{
    node.Print ();
}
#endif
