#include "cgram.h"
#include <string.h>
#include <stdio.h>

#define MULT 5

void
StringToNgramVec (const char *str, unsigned int ngramSize,
                  unsigned int vecSize, VectorType * vec)
{
    unsigned int mask = (unsigned int) (1LL << (8 * ngramSize)) - 1;
    int len = strlen (str);

    memset (vec, 0, vecSize * sizeof (*vec));

    if (vecSize == 0)
        return;

    for (int i = 0; i < len; i++) {

        
        unsigned int val = 0;
        for (int j = i; j < i + ngramSize; ++j) {
            unsigned char c = (j >= len) ? ' ' : str[j];

            val = val * MULT + c;
        }

        int ind = val % vecSize;
        if (vec[ind] < VectorTypeMax) {
            vec[ind]++;
        }
    }
}

void PrintNgramVec (unsigned int vecSize, const VectorType * vec)
{
    for (unsigned int i = 0; i < vecSize; i++) {
        printf ("%u%c", vec[i], (i == vecSize - 1) ? '\n' : '\t');
    }
}

unsigned int
NgramMaxErrNum (unsigned int kerr, unsigned int ngramSize,
                unsigned int vecSize, bool hasTransp)
{
    {
        unsigned int res = ngramSize * kerr;
        if (hasTransp) {
            res += kerr;       /* 
                                   this is because of the transpostions 
                                   one letter changes <= ngramSize ngrams, additional 
                                   modified letter (this is the case with transpositions)
                                   changes <= 1 additional ngram. Coefficient 2 occurs
                                   because after modification of one common n-gram strings
                                   may have 2 distinct new n-grams.
                                   e.g. ABCD -> ACBD
                                   AB, BC, CD -> CA, CB, BD,
                                   hence 3 n-grams are modified and difference is 6.
                                   In case of n-gram (n>2), take n = 3 for example
                                   ABCDER -> ABDCER.
                                   ABC, BCD, CDE, DER -> ABD, BDC, DCE, CER 
                                   4 ngram are modified and distance is 8 = 2*( 3 * kerr + 2 * kerr), kerr = 1
                                 */
        }

        return res;


    }
}
