#include <sys/timeb.h>
#include <utility>
#include "len1index.h"
#include "format.h"
#include "../Compress/int_coder_inl.h"

//#define SHOW_LISTS

const int CLen1Index::MaxIntersectCount;
const int CLen1Index::pagesize;
const int CLen1Index::cachesize;
int CLen1Index::recFetched = 0;
int CLen1Index::lstFetched = 0;
unsigned int CLen1Index::fetchTime = 0;

DEFINE_CLASS_STATIC_COUNTER(CLen1Index, check_string)

CLen1Index::~CLen1Index() {
    for ( mPtrList::iterator it= invgram.begin (); it != invgram.end (); it++) 
        delete it->second;

    for ( mPtrListComp::iterator it= invgramComp.begin (); it != invgramComp.end (); it++) 
        delete it->second.mpStart;
}

int CLen1Index::Insert (const char *str0)
{
    int len = strlen (str0);
    char str[len + NgramLen];

    int i = 0;
    for (i = 0; i < len + NgramLen - 1; i++)
        str[i] = (i < len) ? str0[i] : ' ';
    str[i] = 0;

    //if (len<NgramLen) len=NgramLen;
    // str is already stored in dict (CNgramIndex)
    ngram g;
    // Filling all of the ngram entries
    // Shortened (trailing) ngrams are also stored
    for (i = 0; i < len; i++) {
        g.pack (&str[i], NgramLen, i);
        PtrList *p = invgram[g];
        if (!p) {
            p = new PtrList;
        }
        ListInsert (p, str0);
        invgram[g] = p;
    }

    return 1;
}

int CLen1Index::Search(const char *pStrOrig, int kerr, myagrep & a,
                        int *numComp, StringList * resList, bool silent)
{
    size_t StrLen = strlen(pStrOrig);
    char   str[StrLen + 1];

    strcpy(str, pStrOrig);

    // Filling all the ngram entries
    int dlen = StrLen / (1 + kerr);

    if (dlen >= 1) {
        PtrList pl[2];
        int     nc = 0;

        int     H = dlen * (kerr + 1);
        
        for (int h = 0; h < H; h += dlen) {
            PtrList p;
            //printf("h=%d i=%d dlen=%d kerr = %d\n",h,i,dlen,kerr);
            ReadList(&str[h], dlen, h - kerr, h + kerr, &p);
            ListUnion (pl[nc], p, &pl[1 - nc]);
            nc = 1 - nc;
        }

        if (mHasTransp) {
        
            for (int h = 0, hSwap = dlen - 1; h < H - dlen; h += dlen, hSwap += dlen) {
                PtrList p;
                //printf("h=%d i=%d dlen=%d kerr = %d\n",h,i,dlen,kerr);

                assert(hSwap < H);

                swap(str[hSwap], str[hSwap + 1]);

                ReadList(&str[h], 2 * dlen - int(dlen > 1), h - kerr, h + kerr, &p);

                swap(str[hSwap], str[hSwap + 1]);

                ListUnion (pl[nc], p, &pl[1 - nc]);
                nc = 1 - nc;
            }
        }
#ifdef SHOW_LISTS
        cout << "In-memory search of " << str << " kerr " << kerr << " list size: " << pl[nc].size() << " :\n";
#endif
        for (PtrListIterator it = pl[nc].begin (); it != pl[nc].end (); it++) {
#ifdef SHOW_LISTS
            cout << *it << ", ";
#endif
            if (numComp)
                (*numComp)++;
            if (a.checkRTRFilter(*it)) {
                if (resList != NULL)
                    resList->push_back (string (*it));
                if (!silent)
                    cout << *it << "\n";
            }
        }
#ifdef SHOW_LISTS
        cout << endl;
#endif
    }
    return true;
}

void CLen1Index::ReadList(const char *str, int StrLen, 
                           int minpos, int maxpos,
                           PtrList * ptr)
{
    if (minpos < 0) minpos = 0;
    
    assert (str);
    ptr->clear ();
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        int i = 0;
        // filling commong part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;
        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);

        mPtrList::iterator mapit = invgram.lower_bound (gmin);
        u_int64_t upper_bound = gmax, lower_bound = gmin;
        for (; mapit != invgram.end (); mapit++) {
            u_int64_t cgram = u_int64_t ((*mapit).first);
            if (cgram < lower_bound)
                continue;
            if (cgram > upper_bound)
                break;
            // Let's check if position is ok
            ngram currNgram(cgram);
            if (currNgram.Pos() >= minpos && currNgram.Pos() <= maxpos) {
                PtrList *p = (*mapit).second;
                if (p) {
                    size_t old_sz = ptr->size();
                    ptr->resize(old_sz + p->size());

                    memcpy(&(*ptr)[old_sz], &((*p)[0]), p->size() * sizeof((*p)[0]));
                }
            }
        }
        sort(ptr->begin(), ptr->end());
        ptr->erase(unique(ptr->begin(), ptr->end()), ptr->end()); // Remove possibly repeating strings

        return;
    } else {
        // Create an intersection of all lists
        PtrList     CurrList, OtherList;
        PtrList     *pTmpList[2] = {ptr, &OtherList};
        int         nc = 0;
        bool        first = true;

        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            ngram gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            mPtrList::iterator mapit = invgram.lower_bound (gmin);
            u_int64_t upper_bound = gmax, lower_bound = gmin;

            PtrList *pCurr = (first) ? ptr:&CurrList;
            pCurr->clear();

            for (; mapit != invgram.end (); mapit++) {
                u_int64_t cgram = u_int64_t ((*mapit).first);
                if (cgram < lower_bound)
                   continue;
               if (cgram > upper_bound)
                   break;
               // Let's check if position is ok
               ngram currNgram(cgram);
               if (currNgram.Pos() >= minpos + i && currNgram.Pos() <= maxpos + i) {
                    PtrList * p = mapit->second;
                    if (p) {
                        size_t old_sz = pCurr->size();
                        pCurr->resize(old_sz + p->size());

                        memcpy(&(*pCurr)[old_sz], &((*p)[0]), p->size() * sizeof((*p)[0]));
                    }
               }
            }
            sort(pCurr->begin(), pCurr->end());
            pCurr->erase(unique(pCurr->begin(), pCurr->end()), pCurr->end()); // Remove possibly repeating strings
            // First time, we just fill ptr, next we do intersections
            if (first) {
                first = false;
            } else {
                ListIntersection(*pCurr, *pTmpList[nc], pTmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (pTmpList[nc]->size() <= 1) {
                break;
            }

        }
        if (pTmpList[nc] != ptr) {ptr->swap(*pTmpList[nc]);}

        return;

    }
}

#ifdef _DEBUG
void CLen1Index::Print ()
{
    for (mPtrList::iterator mapit = invgram.begin ();
         mapit != invgram.end (); ++mapit) {
        u_int64_t u = (*mapit).first;
        cout << "gram:" << u << "\n";
        PrintList (*(*mapit).second);
    }
}
#endif

bool CLen1Index::SearchInDB (const char *dict,
                             const char *pStrOrig,
                             const myagrep & a,
                             int kerr,
                             int *numComp,
                             StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened";
        return false;
    }
    size_t StrLen = strlen(pStrOrig);
    char   str[StrLen + 1];

    strcpy(str, pStrOrig);

    // Filling all the ngram entries
    int dlen = StrLen / (1 + kerr);
    /* Searching the database */

    // Filling all the ngram entries
    if (dlen >= 1) {
        OffsetList pl[2];
        int nc = 0;
        int H = dlen * (kerr + 1);
        for (int h = 0; h < H; h += dlen) {
            OffsetList p;
            ReadListDB(&str[h], dlen, h - kerr, h + kerr, &p);
            ListUnion (pl[nc], p, &pl[1 - nc]);
            nc = 1 - nc;
        }

        if (mHasTransp) {
        
            for (int h = 0, hSwap = dlen - 1; h < H - dlen; h += dlen, hSwap += dlen) {
                OffsetList p;
                //printf("h=%d i=%d dlen=%d kerr = %d\n",h,i,dlen,kerr);

                assert(hSwap < H);

                swap(str[hSwap], str[hSwap + 1]);

                ReadListDB(&str[h], 2 * dlen - int(dlen > 1), h - kerr, h + kerr, &p);

                swap(str[hSwap], str[hSwap + 1]);

                ListUnion (pl[nc], p, &pl[1 - nc]);
                nc = 1 - nc;
            }
        }
        //cout<<"Total after union list size "<<pl[nc].size()<<"\n";
#ifdef SHOW_LISTS
        cout << "Disc search of " << str << " kerr " << kerr << " list size: " << pl[nc].size() << " :\n";
#endif
        vector<off_t>      sortList(pl[nc].begin(), pl[nc].end());
        // for disk search it should be faster to access data in increasing offset order
        sort(sortList.begin(), sortList.end());
        START_COUNTER(check_string)

        for (unsigned int i = 0; i < sortList.size (); i++) {
#ifdef SHOW_LISTS
            cout << &dict[sortList[i]] << ", ";
#endif
            if (numComp)
                (*numComp)++;
            if (dict != NULL && a.checkRTRFilter(&dict[sortList[i]])) {
                if (resList != NULL)
                    resList->push_back (string (&dict[sortList[i]]));
                if (!silent)
                    cout << &dict[sortList[i]] << "\n";
            }
        }
        STOP_COUNTER(check_string)
#ifdef SHOW_LISTS
        cout << endl;
#endif
    }
    return true;
}

bool CLen1Index::SearchComp (const char *dict,
                             const char *pStrOrig,
                             const myagrep & a,
                             int kerr,
                             int *numComp,
                             StringList * resList, bool silent)
{
    size_t StrLen = strlen(pStrOrig);
    char   str[StrLen + 1];

    strcpy(str, pStrOrig);

    int dlen = StrLen / (1 + kerr);
    // Filling all the ngram entries
    if (dlen >= 1) {
        OffsetList pl[2];
        int nc = 0;
        int H = dlen * (kerr + 1);
        for (int h = 0; h < H; h += dlen) {
            OffsetList p;
            if (!ReadListComp(&str[h], dlen, h - kerr, h + kerr, &p)) {
                cerr << "ReadListComp failed\n";
                return false;
            }
            ListUnion (pl[nc], p, &pl[1 - nc]);
            nc = 1 - nc;
        }

        if (mHasTransp) {
        
            for (int h = 0, hSwap = dlen - 1; h < H - dlen; h += dlen, hSwap += dlen) {
                OffsetList p;
                //printf("h=%d i=%d dlen=%d kerr = %d\n",h,i,dlen,kerr);

                assert(hSwap < H);

                swap(str[hSwap], str[hSwap + 1]);

                ReadListComp(&str[h], 2 * dlen - int(dlen > 1), h - kerr, h + kerr, &p);

                swap(str[hSwap], str[hSwap + 1]);

                ListUnion (pl[nc], p, &pl[1 - nc]);
                nc = 1 - nc;
            }
        }

        //cout<<"Total after union list size "<<pl[nc].size()<<"\n";
#ifdef SHOW_LISTS
        cout << "In-memory (compressed) search of " << str << " kerr " << kerr << " list size: " << pl[nc].size() << " :\n";
#endif
        vector<off_t>      sortList(pl[nc].begin(), pl[nc].end());
        // for disk search it should be faster to access data in increasing offset order
        sort(sortList.begin(), sortList.end());
        START_COUNTER(check_string)

        for (unsigned int i = 0; i < sortList.size (); i++) {
#ifdef SHOW_LISTS
            cout << &dict[sortList[i]] << ", ";
#endif
            if (numComp)
                (*numComp)++;
            if (dict != NULL && a.checkRTRFilter(&dict[sortList[i]])) {
                if (resList != NULL)
                    resList->push_back (string (&dict[sortList[i]]));
                if (!silent)
                    cout << &dict[sortList[i]] << "\n";
            }
        }
        STOP_COUNTER(check_string)
#ifdef SHOW_LISTS
        cout << endl;
#endif
    }
    return true;
}

bool CLen1Index::CompressInMem(mOffset &offset_map)
{
    off_t       offset = 0;
    char        keyStr[NgramLen + 2];
    CBitWriter  bufferedWriter(1024 * 1024);

    vector<off_t> buffer(1024);

    for ( mPtrList::iterator it= invgram.begin (); it != invgram.end (); it++) {
        ngram    g(it->first);
        PtrList  plist = *it->second;
        size_t   lsize = plist.size ();

        // Let's write a compressed list
        // Ensure it has enough space. 4x size should be more than enough
        if (buffer.size() < lsize * 4) { buffer.resize(lsize * 4); }

        CMemBitWriter bufferedWriter((char *)&buffer[0], buffer.size() * sizeof buffer[0]);

        try {
            if (lsize > 0) {
                vector<off_t> offsets(lsize);
                
                for (int i = 0; i < lsize; i++) {
                    offsets[i] = offset_map[plist[i]];
                }
                sort(offsets.begin(), offsets.end());

                off_t last = 0;
                for (int i = 0; i < lsize; ++i) {
                    if (!mUseDelta) {
                        VarByteEncode(1 + offsets[i] - last, bufferedWriter);
                    } else {
                        DeltaEncode(1 + offsets[i] - last, bufferedWriter);
                    }
                    last = offsets[i];
                }

                bufferedWriter.FinishOutput();

            }
        } catch (const std::runtime_error e) {
            cerr << "Error while writing lists: " << e.what() << endl;
            return false;
        }

        size_t    CompSize = bufferedWriter.getPos();
        char *pNewBuf = new char[CompSize];

        memcpy(pNewBuf, (char *)&buffer[0],CompSize);

        EntryDataCompressInMem data(pNewBuf, lsize, CompSize);
        invgramComp[g] = data; 
    }

    mIsCompInMem = true;

    return true;
}

bool CLen1Index::ReadListComp(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr) 
{
    if (minpos < 0) minpos = 0;
    
    assert (str);
    ptr->clear ();
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        int i = 0;
        // filling commong part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;
        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);

        mPtrListComp::iterator mapit = invgramComp.lower_bound (gmin);
        u_int64_t upper_bound = gmax, lower_bound = gmin;
        for (; mapit != invgramComp.end (); mapit++) {
            u_int64_t cgram = u_int64_t ((*mapit).first);
            if (cgram < lower_bound)
                continue;
            if (cgram > upper_bound)
                break;
            // Let's check if position is ok
            ngram currNgram(cgram);
            if (currNgram.Pos() >= minpos && currNgram.Pos() <= maxpos) {
                const EntryDataCompressInMem &data = mapit->second;
               
                size_t old_sz = ptr->size();
                ptr->resize(old_sz + data.cnt);

                CMemBitReader      reader(data.mpStart, data.size);

                // Getting compressed in-list offsets
                try {
                    off_t last = 0;
                    off_t t;
                    if (!mUseDelta) {
                        for (int j = 0; j < data.cnt; j++) {
                            VarByteDecode(t, reader);
                            t--;
                            last += t;
		            (*ptr)[old_sz + j] = last;
                        }
                    } else {
                        for (int j = 0; j < data.cnt; j++) {
                            DeltaDecode(t, reader);
                            t--;
                            last += t;
		            (*ptr)[old_sz + j] = last;
                        }
                    }
                } catch (const CReachedEof &) {
                    cerr << "unexpected EOF" << endl;
                    return false;
                }
              
            }
        }
        sort(ptr->begin(), ptr->end());
        ptr->erase(unique(ptr->begin(), ptr->end()), ptr->end()); // Remove possibly repeating strings
    } else {
        // Create an intersection of all lists
        OffsetList  CurrList, OtherList;
        OffsetList  *pTmpList[2] = {ptr, &OtherList};
        int         nc = 0;
        bool        first = true;

        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            ngram gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            mPtrListComp::iterator mapit = invgramComp.lower_bound (gmin);
            u_int64_t upper_bound = gmax, lower_bound = gmin;

            OffsetList *pCurr = (first) ? ptr:&CurrList;
            pCurr->clear();

            for (; mapit != invgramComp.end (); mapit++) {
                u_int64_t cgram = u_int64_t ((*mapit).first);
                if (cgram < lower_bound)
                   continue;
               if (cgram > upper_bound)
                   break;
               // Let's check if position is ok
               ngram currNgram(cgram);
               if (currNgram.Pos() >= minpos + i && currNgram.Pos() <= maxpos + i) {
                    const EntryDataCompressInMem &data = mapit->second;
                    size_t old_sz = pCurr->size();
                    pCurr->resize(old_sz + data.cnt);

                    CMemBitReader      reader(data.mpStart, data.size);

                    // Getting compressed in-list offsets
                    try {

                        off_t last = 0;
                        off_t t;
                        if (!mUseDelta) {
                        for (int j = 0; j < data.cnt; j++) {
                            VarByteDecode(t, reader);
                            t--;
                            last += t;
		            (*pCurr)[old_sz + j] = last;
                        }
                    } else {
                        for (int j = 0; j < data.cnt; j++) {
                            DeltaDecode(t, reader);
                            t--;
                            last += t;
		            (*pCurr)[old_sz + j] = last;
                        }
                    }
                } catch (const CReachedEof &) {
                    cerr << "unexpected EOF" << endl;
                    return false;
                }

               }
            }
            sort(pCurr->begin(), pCurr->end());
            pCurr->erase(unique(pCurr->begin(), pCurr->end()), pCurr->end()); // Remove possibly repeating strings
            // First time, we just fill ptr, next we do intersections
            if (first) {
                first = false;
            } else {
                ListIntersection(*pCurr, *pTmpList[nc], pTmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (pTmpList[nc]->size() <= 1) {
                break;
            }

        }
        if (pTmpList[nc] != ptr) {ptr->swap(*pTmpList[nc]);}
    }
    return true;
}
