#ifndef ALPHABET_H
#define ALPHABET_H

#include <memory.h>
#include <ctype.h>

#include <string>

enum EAlphabet {
    eEnglish,
    eRussian,
    eDNA,
    eEnglishNum,
    eUnknown
};

extern const char* EnglishABC;
extern const char* RussianABC;
extern const char* EnglishNumABC;
extern const char* DNA_ABC;

inline const char *GetAlphabet(EAlphabet type) {
    switch (type) {
        case eEnglish: return EnglishABC;
        case eRussian: return RussianABC;
        case eDNA:     return DNA_ABC;
        case eEnglishNum: return EnglishNumABC;
    }
    return NULL;
}

inline EAlphabet GetAlphabetType(const std::string& Name) {
    if (!strcasecmp(Name.c_str(), "eng")) return eEnglish;
    if (!strcasecmp(Name.c_str(), "engnum")) return eEnglishNum;
    if (!strcasecmp(Name.c_str(), "rus")) return eRussian;
    if (!strcasecmp(Name.c_str(), "dna")) return eDNA;

    return eUnknown;
}

inline int DnaCharNum(char c) {
    c = toupper(c);
    switch (c) {
        case 'A': return 0;
        case 'C': return 1;
        case 'G': return 2;
        case 'T': return 3;
        case 'N': return 4;
        default:break;
    }
    return -1;
}

#endif
