#include <assert.h>
#include <iostream>
#include <algorithm>
#include "fqnode.h"
#include "myagrep.h"


#ifdef MYERS_DIST
CMyersEdistFastPair EdCalc;
#endif


// Do not change this
const char CFQnode::termChar = 0;

void CFQnode::Insert(const StringList & lst, bool hasTransp, bool silent, int level)
{
    if (!silent) {
        std:: cout << " Insert of " << lst.size () << " strings, level = "<<level<<" this = "<<this << std::endl;
    }
    // First of all we choose the root randomly if not exists
    unsigned int ListSize = lst.size ();
    if (!ListSize)
        return;
    if (!nodes.size ()) {
        int num = 0;
        
        num = rand () % ListSize;

        root = lst[num];

        nodes.insert (nodes.end (), (CFQnode *) NULL);
    }

    VStringList SplitList(max_dist + 1);
    StringList EmptyList;
    for (StringList::const_iterator it = lst.begin (); it != lst.end (); it++) {
#ifdef MYERS_DIST
        int dst = EdCalc.CalcDist(root, *it, hasTransp);
#else
        int dst = EditDistance(root, *it, hasTransp);
#endif
        if (dst > 0) {
            if (dst > max_dist) {
                for (int i = max_dist + 1; i <= dst; i++) {
                    nodes.insert (nodes.end (), (CFQnode *) NULL);  //< First insert NULL-node
                    SplitList.insert(SplitList.end (), EmptyList);
                }
                max_dist = dst;
            }
            SplitList[dst].insert(SplitList[dst].end (), *it);
        }
    }
    // Now the lists are split and must be inserted into the nodes
    // NULL-nodes must be created
    for (int i = 1; i <= max_dist; i++) {
        if (SplitList[i].size ()) {
            if (nodes[i] == NULL) {
                nodes[i] = new CFQnode ();
            }
            nodes[i]->Insert(SplitList[i], hasTransp, silent, level + 1);
        }
    }
}

int CFQnode::DumpToFile(off_t offset, int fd)
{
    // We writing data in the following order
    // 0. buflen
    // 1. K - number of nodes u_int16_t 
    // 2. root (which might be a sequence of words)
    // 3. bulkCnt
    // 4. K times subtree records in the form
    //   subtree distance from the root (u_int16_t)
    //   offset of the subtree
    // 5. K - subtrees
    u_int16_t K = 0;
    int buflen = sizeof (K) + sizeof (buflen) + sizeof(bulkCnt) /* number of nodes + buffer size */ 
                            + root.size() + (bulkCnt == 1); /* total length of all words in root */

    for (int i = 1; i < nodes.size (); i++) {
        if (nodes[i]) {
            K++;
            buflen += sizeof (offset) + sizeof (K);
        }
    }

    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 0.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    // Step 1.
    *(u_int16_t *) (void *) sbuf = K;
    sbuf += sizeof (K);
    // Step 2.
    *((u_int16_t *)sbuf) = bulkCnt;
    sbuf += sizeof bulkCnt;
    // Step 3.
    if (bulkCnt > 1) {
        memcpy (sbuf, root.data(), root.size());
        sbuf += root.size();
    } else {
        memcpy (sbuf, root.c_str(), root.size() + 1);
        sbuf += root.size() + 1;
    }
    // Step 4.
    off_t *ptrs[K];
    int cnt = 0;
    for (u_int16_t i = 1; i < nodes.size (); i++)
        if (nodes[i]) {
            *(u_int16_t *) (void *) sbuf = i;
            sbuf += sizeof (u_int16_t);
            ptrs[cnt++] = (off_t *) (void *) sbuf;
            assert (cnt <= K);
            sbuf += sizeof (off_t);
        }
    int sub_size = 0;
    // Step 5.
    cnt = 0;
    for (u_int16_t i = 1; i < nodes.size (); i++)
        if (nodes[i]) {
            int s_sz =
                nodes[i]->DumpToFile (offset + buflen + sub_size, fd);
            if (s_sz < 0) {
                munmap (buf, buflen + delta);
                return -1;
            }
            *ptrs[cnt++] = offset + buflen + sub_size;
            assert (cnt <= K);
            sub_size += s_sz;
        }


    assert (sbuf - (char *) buf == delta + buflen);
    if (munmap (buf, buflen + delta) < 0)
        return -1;
    return buflen + sub_size;
}

void CFQnode::SearchAtLevel (const std::string & s, int nerr, const myagrep & a,
#ifdef MYERS_DIST
                            CMyersEdist &MyersEdist,
#endif
                             int *numComp, StringList * resList, bool hasTransp,
                             bool silent)
{
    if (numComp)
        (*numComp)++;
    if (!max_dist) {
        if (numComp)
            (*numComp) += bulkCnt;
#ifdef MYERS_DIST
        if (nerr > 2) { // For nerr == 1,2 Myers algorithm is slower than myagrep
            const char* pBeg    = root.data();
            const char* pEnd    = pBeg + root.size();
            const char* pCurr   = pBeg;

            for (const char *p = pBeg; pBeg < pEnd; pBeg = p + 1) {
                p = strchr(pBeg, termChar);

                if (!p) {
                    cerr << "Internal error: expect TermChar" << endl;
                    abort();
                }


                int dist = MyersEdist.CalcDist(pBeg);
        
                if (dist <= nerr) {
                    if (!silent)
                        cout << pBeg << "\n";
                    if (resList != NULL)
                        resList->push_back(pBeg);
                }
            }
            return;
        }
#endif
        a.findPattern(CInterface::filterType, termChar,
                          root.data(), root.data() + root.size(), silent, resList);
        return;
    }
    // We must check the root
#ifdef MYERS_DIST
    int d = MyersEdist.CalcDist(root.c_str());
#else
    int d = EditDistance(s, root, hasTransp);
#endif
    if (d <= nerr) {
        if (!silent)
            cout << root << "\n";
        if (resList != NULL)
            resList->push_back(root);
    }
    int imin = d - nerr, imax = d + nerr;
    if (imin < 1)
        imin = 1;
    if (imax > max_dist)
        imax = max_dist;
    for (int i = imin; i <= imax; i++)
        if (nodes[i])
            nodes[i]->SearchAtLevel (s, nerr, a, 
#ifdef MYERS_DIST
            MyersEdist,
#endif
            numComp, resList, hasTransp, silent);
}

bool CFQnode::SearchInFileAtLevel (int fd, off_t offset,
                                   const std::string & s, const myagrep & a,
                                   int nerr, int *numComp,
                                   StringList * resList, bool hasTransp,
                                   bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0 || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen)) {
        return -1;
    }
    int delta = 0;
    if (offset) {
        delta = (offset % getpagesize ());
    }
    void *buf = mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;

    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int K = *((u_int16_t *) (void *) sbuf);
    sbuf += sizeof (u_int16_t);
    u_int16_t bulkCnt = *((u_int16_t *)sbuf);
    sbuf += sizeof (u_int16_t);

    size_t strLen = buflen - sizeof(buflen) - sizeof(u_int16_t) - sizeof(bulkCnt) - K * (sizeof(u_int16_t) + sizeof(off_t)); ;

    if (!K) {
        if (numComp)
            (*numComp) += bulkCnt;
        //std::string root(sbuf, strLen - 1); //< do not use full length here, it includes zero terminating char
        //cout << "@@@ '" << root << "' '" << (sbuf) << "' "  << root.size() << " -> '" << s << "' " << s.size() << endl;
        // If no subtree just check, unmap and return 
        a.findPattern(CInterface::filterType, termChar, 
                      sbuf, sbuf + strLen, silent, resList);

        return (!munmap (buf, buflen + delta));
    }
    std::string root;
    if (bulkCnt > 1) {
       root = string(sbuf, strLen);
    } else {
       root = string(sbuf, strLen - 1);
    }
    sbuf += strLen;

    if (numComp)
        (*numComp)++;
#ifdef MYERS_DIST
    int edist = EdCalc.CalcDist (root, s, hasTransp);
#else
    int edist = EditDistance(root, s, hasTransp);
#endif
    //cout << "### '" << root << "' '" << (sbuf - strLen) << "' "  << root.size() << " -> '" << s << "' " << s.size() << " edist = " << edist << endl;
    //cout<<" edist "<<edist<<endl;
    // We must check the root
    if (edist <= nerr) {
        if (!silent)
            cout << sbuf << "\n";
        if (resList != NULL)
            resList->push_back (root);
    }
    for (int i = 1; i <= K; i++) {
        u_int16_t dist = *(u_int16_t *) (void *) sbuf;
        sbuf += sizeof (u_int16_t);
        off_t sub_off = *(off_t *) (void *) sbuf;
        sbuf += sizeof (off_t);
        //cout<<" Offset "<<sub_off<<" dist "<<dist<<endl;
        if (dist >= edist - nerr && dist <= edist + nerr)
            if (!SearchInFileAtLevel
                (fd, sub_off, s, a, nerr, numComp, resList, hasTransp, silent)) {
                munmap (buf, buflen + delta);
                return false;
            }
    }
    assert (sbuf - (char *) buf == delta + buflen);
    return (!munmap (buf, buflen + delta));
}

void CFQnode::Print (int level)
{
    for (int i = 0; i < level; i++)
        cout << "*";
    cout << ":" << root << "\n";
    for (int i = 1; i <= max_dist; i++) {
        if (nodes[i]) {
            for (int j = 0; j < level; j++)
                cout << "  ";
            cout << " dst " << i << "\n";
            nodes[i]->Print (level + 1);
        }
    }
}

void CFQnode::appendAllStrings(string & str)
{
     str.append(root.data(), root.size());
     str.append(&termChar, 1);
     //cout << "Appedning of size " << root.size() << endl;
     for (int i = 1; i <= max_dist; ++i) 
     if (nodes[i]) nodes[i]->appendAllStrings(str);
}

void CFQnode::CreateBuckets(int buckSize, int level)
{
     if (!level) fillTreeCnt();

     buckSize = max(buckSize, 1);

     unsigned int cnt = 0;
     for (int i = 1; i <= max_dist; ++i) {
        cnt += (nodes[i] != NULL);
     }
     if (cnt && treeCnt > 1 && // this is important, only then bulkCnt will be greater than 1
         treeCnt <= buckSize) {
        root.append(&termChar, 1);

        for (int i = 1; i <= max_dist; ++i) 
        if (nodes[i]) {
            nodes[i]->appendAllStrings(root);
            delete nodes[i];
        }

        max_dist = 0;
        nodes.resize(0);
        bulkCnt = count(root.begin(), root.end(), termChar);
     } else {
        for (int i = 1; i <= max_dist; ++i) 
        if (nodes[i]) {
            nodes[i]->CreateBuckets(buckSize, level + 1);
        }
     }
} 

void CFQnode::fillTreeCnt()
{
    treeCnt = bulkCnt;
    for(int i = 1; i <= max_dist; ++i) 
       if (nodes[i]) {
           nodes[i]->fillTreeCnt();
           treeCnt += nodes[i]->treeCnt;
       }
    }
