#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <map>
#include <stdlib.h>

#include "../afuzzy/edist.h"

using namespace std;

typedef pair<int,int> PI;
typedef map<PI, int>  distMap;

distMap               dist;
vector<string>        strs;

int getDistOld(int i1, int i2, bool transp)
{
    return  EditDistanceOld(strs[i1], strs[i2], transp);
}

unsigned NonMatchQty;

void CompareDist(int d, int dOld, int i1, int i2)
{
    if (d == dOld) return;
    cout << "Non-matching distances, Old:" << dOld << " New: " << d << endl;
    cout << strs[i1] << " <> " << strs[i2] << endl;
    ++NonMatchQty;
}

int getDist(int i1, int i2, bool transp)
{
#if 0
    PI p1(i1, i2), p2(i2, i1);

    distMap::iterator it = dist.find(p1);
    if (it != dist.end())
        return it->second;
    it = dist.find(p2);
    if (it != dist.end())
        return it->second;
    return dist[p1] = dist[p2] = EditDistance(strs[i1], strs[i2], transp);
#else
    return  EditDistance(strs[i1], strs[i2], transp);
#endif
}

int main(int argc, char * argv[])
{
    map<int, int> distCnt;

    double avg = 0, avg1 = 0;

    if (argc != 2) {
        cerr << "Usage, to count with transpositions: " << argv[0] << " 1 < [test file]" << endl;
        cerr << "Usage, to count without transpositions: " << argv[0] << " 0 < [test file]" << endl;
    }
    bool transp = atol(argv[1]);
    string s;
    while(cin >> s) {
        strs.push_back(s);
    }
    int N = 10000000;
    int cnt = 0, cnt1 = 0, cnti = 0;

    for (int a = 0; a <= N; ++a) {
        int i = rand() % strs.size();
        int j = rand() % strs.size();
        int k = rand() % strs.size();

        if (j == k || k == i) continue;

        avg1 = (cnti * avg1 + strs[i].size())/(cnti+1);
        ++cnti;

        int dIJ = getDist(i, j, transp);
        avg = (cnt1 * avg + dIJ)/(cnt1 + 1);
        distCnt[dIJ]++;
        ++cnt1;

        int dJK = getDist(j, k, transp);
        int dIK = getDist(i,k, transp);

        int dJKOld = getDistOld(j, k, transp);
        int dIKOld = getDistOld(i,k, transp);

        CompareDist(dJK, dJKOld, j, k);
        CompareDist(dIK, dIKOld, i, k);



        if (!(a % 100000))
            cerr << "Cnt: " << a << endl;
        if (dIJ +  dJK < dIK) {
                cerr << "cnt = " << a << endl;
                cerr << "Triangle inequality failed :"  << endl;
                cerr << strs[i] << "-> " << strs[j] <<  " :" << dIJ << endl;
                cerr << strs[j] << "-> " << strs[k] <<  " :" << dJK << endl;
                cerr << strs[i] << "-> " << strs[k] <<  " :" << dIK << endl;
                return -1;
        }
    }
    cout << "Average dist: " << avg << " average word length: " << avg1 << endl;
    cout << "Dist distributions:\n";
    long long sum = 0;
    for(map<int,int>::iterator it = distCnt.begin(); it != distCnt.end(); ++it) {
        sum += it->second;
    }
    for(map<int,int>::iterator it = distCnt.begin(); it != distCnt.end(); ++it) {
            cout << setw(8) << it->first << "|";
    }
    cout << endl;
    for(map<int,int>::iterator it = distCnt.begin(); it != distCnt.end(); ++it) {
            cout << setw(8) << setfill('-') << "" << "+";
    }
    cout << endl;
    for(map<int,int>::iterator it = distCnt.begin(); it != distCnt.end(); ++it) {
            cout << setw(8) << setfill(' ') << (100LL * it->second / sum ) << "|" ;
    }
    cout << endl;

    cout << NonMatchQty << " cases, when distance was computed incorrectly" << endl;
    return 0;
}
