#ifndef _FQNODE_H
#define _FQNODE_H

#include <vector>
#include <string>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <fstream>
#include <stdexcept>
#include <filesize.h>
#include "myagrep.h"

class CFQNodeMultiPivot;


using namespace std;


typedef vector <string >     StringList;
typedef vector <StringList>  VStringList;

/* 
 * All children are stored in the array of arrays that is indexed by the distance 
 * to the first pivot.
 */
typedef vector<u_int16_t> vus;
struct OneChild {
    CFQNodeMultiPivot  *pChild;
    vus                 dists; // distances for to all pivots starting from the second
    OneChild(CFQNodeMultiPivot *child, vus _dists) : pChild(child), dists(_dists) {}
};

typedef vector<OneChild>     Children;
typedef vector<Children>     FQNodeList;


class CFQNodeMultiPivot {
  protected:
    string        bucket;
    StringList    pivots;
    u_int8_t      pivotCnt;
    int           treeCnt;
    u_int16_t     bulkCnt; // 1 if no bulk

    FQNodeList    nodes;
    void SearchAtLevel (const std::string & s, int nerr, const myagrep & a,
                        int *numComp, StringList * resList, bool hasTransp, 
                        bool silent);
    const static char termChar;

    void fillTreeCnt();
    void appendAllStrings(string & str);
  public:
   CFQNodeMultiPivot (u_int8_t _pivotCnt = PIVOT_CNT): pivotCnt (_pivotCnt), bulkCnt(1), treeCnt(0) {;} 
   ~CFQNodeMultiPivot();
    int DumpToFile (off_t offset, int fd);
    void Insert(const StringList & lst, bool hasTransp, bool silent, int level);
    void Search(const std::string & s, unsigned int nerr, int *numComp, 
                 StringList * resList, bool hasTransp, bool silent) {
        myagrep a(s.c_str(), nerr, hasTransp);

        SearchAtLevel (s, nerr, a, numComp, resList, hasTransp, silent);
    }
    static bool SearchInFile (int fd, const std::string & s,
                              unsigned int nerr, int *numComp,
                              StringList * resList, bool hasTransp, bool silent) {
        cerr << "On disk index searches are not supported" << endl;
        return false;
    }
    void Print (int level = 0);
    void CreateBuckets(int buckSize, int level = 0);
    int  GetPivotCnt() const { return pivotCnt; }
    void SetPivotCnt(int _PivotCnt) {
        if (!nodes.size() && !pivots.size() && !bucket.size()) {
            pivotCnt = _PivotCnt;
            return;
        }
        throw runtime_error("Changing the number of pivots is not allowed on non-empty nodes");
    }
protected:
    void AddTmpList(StringList &tmpList, vus &prev,
                    bool hasTransp, bool silent, int level);
};

#endif
