#include "gensupercond.h"
#include "gensupercond_nfa.h"
#include <iostream>
#include <string>

#define REP_QTY 5

using namespace std;

struct NGPrint {
    void operator()(const char *pStr, unsigned int sz) {
        cout << string(pStr, sz) << endl;
    }
};

struct NGCollect {
    void operator()(const char *pStr, unsigned int sz) {
        mLst.push_back(string(pStr, sz));
    }
    StringList   mLst;
};

void Usage(const char* pProg) {
    cerr << pProg << " <string> <kerr> <print Ng?>" << endl;
}

int main(int argc, char* pArgv[]) {
    if (argc != 4) {
        Usage(pArgv[0]);
        return -1;
    }
    string   sPat       = pArgv[1];
    unsigned kerr       = atol(pArgv[2]);
    bool     bPrintNG   = atol(pArgv[3]);

    cout << "String: '" << sPat << "' kerr = " << kerr << endl;


    string  sAlphaName;

    GetEnv(ENV_ABC, sAlphaName);

    EAlphabet           AlphaType = GetAlphabetType(sAlphaName);

    if (AlphaType == eUnknown) {
        cerr << "Unknown alphabet: '" << sAlphaName << "' check env var " << ENV_ABC << endl;
        return 1;
    }
    const char* pABC = GetAlphabet(AlphaType);

    NGCollect ngc;

    DECLARE_STATIC_COUNTER(GenTime);

    START_COUNTER(GenTime);

    for (unsigned i = 0; i < REP_QTY; ++i) {

        ngc.mLst.clear();
        GenSuperCondNG_NFA(ngc, sPat, pABC, kerr);

    }

    STOP_COUNTER(GenTime);

    cout << "Time spent: " << GET_COUNTER(GenTime)/REP_QTY << " microsecs, RepQty " << REP_QTY << endl;

    StringHashSet   dict;

    bool bHasTransp = false;


    GenCondNG(eSuperCondensed, pABC, sPat, kerr, bHasTransp, dict);


    unsigned nUnfound1 = 0, nUnfound2 = 0;

    cout << "======================================" << endl;
    if (bPrintNG) {
    cout << ngc.mLst.size() << " strings in NG " << endl;
        for (unsigned i = 0; i < ngc.mLst.size(); ++i) {
            cout << ngc.mLst[i] << endl;
        }
    }

    cout << "======================================" << endl;

    for (unsigned i = 0; i < ngc.mLst.size(); ++i) {
        const string& w = ngc.mLst[i];

        if (dict.find(w) == dict.end()) {
            cout << "Unfound (1) string: " << w << endl;
            nUnfound1++;
        }

    }

    StringHashSet ListHash(ngc.mLst.begin(), ngc.mLst.end());

    for (StringHashSet::const_iterator it = dict.begin(); it != dict.end(); ++it) {
        const string& w = *it;

        if (ListHash.find(w) == ListHash.end()) {
         cout << "Unfound (2) string: " << w << endl;
         nUnfound2++;
        }
    }

    cout << ngc.mLst.size() << " -> " << dict.size() << endl;

    if (ngc.mLst.size() == dict.size() && !nUnfound1 && !nUnfound2) {
        cout << "FULL MATCH!" << endl;
    } else {
        cout << "Incomplete MATCH, have unfound1 = " << nUnfound1 << " unfound2 = " << nUnfound2 << endl;
    }
}
