#include "index.h"
#ifndef _USE_BDB
#include "len1indexNoBDB.h"
#define DB_CLASS CLen1IndexNoBDB
#else
#include "len1indexBDB.h"
#define DB_CLASS CLen1IndexBDB
#endif

const int CNgramIndex1::NgramLen;
//const int CNgramIndex1::SeqIndexBound;

CNgramIndex1 *ptr1 = NULL;
void *ptr2 = NULL;

DEFINE_CLASS_STATIC_COUNTER(CNgramIndex1, aseek);

CNgramIndex1::CNgramIndex1 (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false), seqIndex(hasTransp, AlphaType) 
{
    db = new DB_CLASS (NgramLen, 
                        mHasTransp, 
                    HAS_COMPRESS);
#if HAS_COMPRESS
    db->setUseDelta(USE_DELTA);
#endif
    //cout<<size_t(&this->dict)<<" "<<size_t(&this->seqIndex)<<" size="<<sizeof(this->seqIndex)<<endl;
}

CNgramIndex1::~CNgramIndex1 ()
{
    CloseDB();
    for (size_t i = 0; i < dict.size (); i++)
        delete [] dict[i];
    delete db;
}

bool CNgramIndex1::Insert (const char *str)
{
    // The string must be dupplicated and stored into 
    // an array variable dict
    int len = strlen (str);
    if (!len)
        return false;
    char *ptr = NULL;

    ptr = newstrdup (str);
    // Let's insert an entry into a dictionary of strings
    dict.push_back (ptr);

    if (len - MAX_KERR < DB_CLASS::SeqIndexBound(MAX_KERR)) {
        if (!seqIndex.Insert (ptr)) {
            return false;
        }
    }

    return bool (db->Insert (ptr));
}

bool CNgramIndex1::Search (const char *str, int kerr, int *nComp,
                           StringList * resList, bool silent)
{
    int         len = strlen (str);
    myagrep     a(str, kerr ,mHasTransp);

    if (len >= DB_CLASS::SeqIndexBound(kerr)) {
        bool res = db->Search (str, kerr, a, nComp, resList, silent);

        if (!res) {
            return false;
        }
    } else {
        if (!seqIndex.Search (str, kerr, nComp, resList, silent))
            return false;
    }

    return true;
}

bool CNgramIndex1::SearchInDB (const char *str, int kerr, int *nComp,
                               StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is closed.\n";
        return false;
    }
    bool res = true;
    mapDict = (char *) mmap (NULL, mapSize, PROT_READ, MAP_SHARED, fd, 0);
    if ((void *) mapDict == MAP_FAILED) {
        Perror ("Error memory mapping!");
        return false;
    }
    int len = strlen (str);
    myagrep a(str, kerr, mHasTransp);

    if (len >= DB_CLASS::SeqIndexBound(kerr)) {
        if (!db->SearchInDB (mapDict, str, a, kerr, nComp, resList, silent))
            res = false;
    } else {
        START_COUNTER(aseek);
        if (!seqIndex.SearchInDB (str, kerr, nComp, resList, silent))
            res = false;
        STOP_COUNTER(aseek);
    }
exit:
    if (mapDict && mapDict != MAP_FAILED) {
        if (munmap(mapDict, mapSize)) {
             Perror("unmap failed");
             return false;
        }
    }
    return res;
}


bool CNgramIndex1::OpenDB (const char *name, bool create)
{
    //cout<<ptr1<<" "<<__func__<<" "<<ptr1->db<<endl;
    if (opened)
        CloseDB ();
    if (!create && !seqIndex.OpenDB (seqIndexName (name).c_str (), create))
        return false;

    baseName = name;
    grmName = baseName + ".grm";
    mapDict = NULL;
    fd = 0;

    if (!db->OpenDB (grmName.c_str (), create)) {
        return false;
    }
    if (!create) {
        mapSize = FileSize (name);
        if (mapSize < 0) {
            return false;
        }
        if (!mapSize)
            return opened = true;   // Empty dictionary
        fd = open (name, O_RDONLY);
        if (fd < 0) {
            return false;
        }
    } else {
        fd = open (name, O_CREAT | O_TRUNC | O_RDWR, 0664);
        if (fd < 0) {
            Perror ("Cannot open file!");
            return false;
        }
    }
    //cout<<"Ngram this="<<this<<" end "<<__func__<<" "<<ptr1->db<<endl;

    return opened = true;
}

bool CNgramIndex1::CloseDB ()
{
    if (!opened)
        return true;
    if (!seqIndex.CloseDB ())
        return false;
    if (!db->CloseDB ()) {
        cerr << "Cannot close internal database\n";
        return false;
    }
    if (mapDict && mapDict != MAP_FAILED)
        munmap (mapDict, mapSize);
    mapDict = NULL;

    if (fd)
        close (fd);
    opened = false;
    return true;
}



bool CNgramIndex1::DumpDB (const char *name)
{
    system("./clear_db.sh");

    ptr1 = const_cast < typeof (this) > (this);
    //cout<<" Ngram this ="<<this<<" "<<dict.begin()<<" #0## "<<endl;
    //cout<<this<<" 0 "<<ptr1<<" "<<__func__<<" "<<ptr1->db<<endl;
    if (!OpenDB (name, true)) {
        return false;           // create = true
    }
    //cout<<this<<" 1 "<<ptr1<<" "<<__func__<<" "<<ptr1->db<<endl;
    string fname = seqIndexName (name);
    const char *ptr = fname.c_str ();
    ptr2 = &dict;
    //cout<<((PtrListVector *)ptr2)->begin()<<" #1## "<<endl;
    if (!seqIndex.DumpDB (ptr)) {
        return false;
    }
    //cout<<"Ngram this="<<this<<" "<<dict.begin()<<" #2## "<<endl;
    //cout<<((PtrListVector *)ptr2)->begin()<<" #2.check## "<<&dict<<"<->"<<ptr2<<endl;
    //cout<<this<<" 2 "<<ptr1<<" "<<__func__<<" "<<ptr1->db<<endl;
    // First of all we must dump the dictionary itself
    mOffset off_list;
    for (PtrListVector::iterator it = dict.begin (); it != dict.end ();
         it++) {
        //cout<<dict.begin()<<" #4## "<<endl;
        //cout<<" ### "<<*it<<" "<<endl;
        off_t cur = lseek (fd, 0, SEEK_CUR);
        if (cur < 0)
            return false;
        off_list[*it] = cur;
        if (write (fd, *it, strlen (*it)) < (ssize_t) strlen (*it))
            return false;
        write (fd, "\0", 1);
    }
    //cout<<this<<" 3 "<<ptr1<<" "<<__func__<<" "<<ptr1->db<<endl;


    if (!db->DumpDB (grmName.c_str (), off_list))
        return false;
    return CloseDB ();
}

void CNgramIndex1::ResetCounters()
{
     RESET_COUNTER(aseek);
     db->ResetCounters();
}

void CNgramIndex1::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
    if (testNum <= 0) return;
     db->PrintAddStat(testNum, kerr, ostr);
     //ostr << "    Time to seek sequentially (ms): " << GET_COUNTER_MS(aseek);
}


#ifdef _DEBUG
void CNgramIndex1::Print ()
{
    for (int i = 1; i <= MaxLen; i++) {
        cout << "{{{ " << i << "\n";
        db[i]->Print ();
        cout << "}}} " << i << "\n";
    }
}
#endif
