#include "index.h"

#include <set>
#include <stdlib.h>

#include "afuzzy.h"

using namespace std;

void CNgramFiltIndex::printAll()
{
    GramInvList::iterator it;
    
    for (unsigned int i = 0; i < mWordList.size(); ++i) {
         cout << mWordList[i] << ": id = " << i << endl;
    }
    string s;
    for (it = mIndx.begin(); it != mIndx.end(); ++it) {
        if (s != it->first) {
            cout << endl << it->first << " : " << it->second;
        } else { 
            cout << ", " << it->second;
        }
        s = it->first;
    }
    cout << endl;
}

//#define TEST_PRINT
void CNgramFiltIndex::search(const std::string & str, unsigned int kerr, StringList & res, unsigned int & cntCheck, unsigned int condFlags)
{
    unsigned int     maxSize1 = str.size() / (kerr + 1);

    unsigned int     maxSize = maxSize1;

    if (mHasTransp && maxSize) {
        --maxSize;
    }

    afuzzy           a(mHasTransp);
    Sign             hsign(mHmap, GetMaxDigit(), str.c_str());

    a.init(str.c_str(), kerr);

#ifdef TEST_PRINT
    cout << "maxSize = " << maxSize << " maxSize1 = " << maxSize1 << " str sign = " << hsign << endl;
#endif

    res.clear();
    cntCheck = 0;

    if (!maxSize || str.size() < kerr) {
        cntCheck = mWordList.size();
        for (unsigned int i = 0; i < mWordList.size(); ++i) {
            //std::cout << mWordList[i] << std::endl;
            if (a.checkRTR(mWordList[i].c_str()))
                res.push_back(mWordList[i]);
        }
        return;
    }

    set<unsigned int> resWordId;

    unsigned int comSize = min(mGramSize, maxSize);

    for (unsigned int pos = 0; pos <= maxSize1 * kerr; pos += maxSize1) {
        set<unsigned int> tmpWordId;
        set<unsigned int> tmp;

        for(unsigned int i = pos; i <= str.size() - comSize; ++i) {
            if (i + comSize > pos + maxSize || i  > pos + mMaxIntersectCnt) 
                break;


            std::string ngram = str.substr(i, comSize);
#ifdef TEST_PRINT
            cout << "Search ngram '" << ngram << "' pos = " << pos << " i = " << i << endl;
#endif
            GramInvList::iterator it = mIndx.find(ngram);
            if (it == mIndx.end())
                it = mIndx.lower_bound(ngram);
            tmp.clear();
            for (;it != mIndx.end(); ++it) {
                // checking for common prefix
                if (it->first.substr(0, comSize) != ngram) 
                    break;

                // checking for additional conds
                if (condFlags & 1) {
                    if (abs((int)i - (int)it->second.mPos) > kerr)
                        continue;
                }
                if (condFlags & 2) {
                    if (abs((int)str.size() - (int)it->second.mWordLen) > kerr)
                        continue;
                }
                if (condFlags & 4) {
                    int d1 = it->second.mSign.Diff(hsign);
                    int d2 = abs(it->second.mSign.abs() - hsign.abs());
                    int d3 = (d1 - d2) / 2;
                    int minErr = d1 - d3;
#ifdef TEST_PRINT
                    cout << "checking diff from " << it->second.mSign << " d1 = " << d1 << " d2 = " << d2 << " minErr = " << minErr << endl;
#endif
                    if (minErr > kerr)
                        continue;
                }
          
                if (i == pos) {
                    tmpWordId.insert(it->second.mWordId);
                } else {
                    tmp.insert(it->second.mWordId);
                }
            }
            if (i != pos) {
                set<unsigned int> tmp1;
                
#if 0
                if (str == "nked") {
                    cout << "tmpWordId.size() BefInt " << tmpWordId.size() << " tmp.size() = " << tmp.size() << endl;
                }
#endif
                set_intersection(tmp.begin(), tmp.end(),
                                 tmpWordId.begin(), tmpWordId.end(),
                                 inserter(tmp1, tmp1.begin()));

                tmpWordId = tmp1;

#if 0
                if (str == "nked") {
                    cout << "tmpWordId.size() AftInt (gram = " << ngram << ") " << tmpWordId.size() << endl;
               }
#endif
            }
        }

        tmp.clear();
        set_union(resWordId.begin(), resWordId.end(),
                  tmpWordId.begin(), tmpWordId.end(),
                  inserter(tmp, tmp.begin()));
        resWordId = tmp;
#if 0
        if (str == "nked") {
            cout << "pos = " << pos << " tmpWordId.size() = " << tmpWordId.size() << endl;
            cout << "resWordId.size() = " << resWordId.size() << endl;
            
        }
#endif
    }
    for (set<unsigned int>::iterator it = resWordId.begin(); it != resWordId.end(); ++it, cntCheck++) {
        if (a.checkRTR(mWordList[*it].c_str()))
            res.push_back(mWordList[*it].c_str());
    }
        
}

void CNgramFiltIndex::insertWord(const std::string & str) 
{
    if (mWordDict.find(str) != mWordDict.end())
        return;

    std::string str1 = str;
    mWordDict.insert(make_pair(str, mWordList.size()));
    unsigned int wordId = mWordList.size();
    mWordList.push_back(str);
    for (unsigned int k = 0; k < mGramSize - 1; ++k)
        str1 += " ";
    if (str.size()) {
        for (unsigned int i = 0; i < str.size(); ++i) {
#ifdef TEST_PRINT
            cout << "Inserting str ngram of size " << mGramSize << " at pos " << i << " of string " << str << endl;
#endif
            CFullNgramInfo grmInfo(wordId, i, str.size(), str.c_str(), mHmap);

            mIndx.insert(make_pair(str1.substr(i, mGramSize), grmInfo));
        }
    }
}
