#include "storeobj.h"

const char CStringHeap::termChar = '@';

CFreqGramVectorList::~CFreqGramVectorList ()
{
    for (unsigned int i = 0; i < plist.size(); ++i) {
        delete plist[i];
    }
}

void CFreqGramVectorList::Search (const std::string s, int MaxErr,
                              int *numComp, StringList * resList,
                              bool hasTransp, bool silent)
{
    unsigned int gramDist = NgramMaxErrNum (MaxErr, NgramSize, VectorLength, hasTransp);
    myagrep a(s.c_str(), MaxErr, hasTransp);

    CFreqGramVector v (0, VectorLength - 1);
    v.InitFromString (s.c_str ());
    //cout<<s<<":"<<" ";v.Print ();
    Search (v, 0, 0, s, gramDist, a, numComp, resList, hasTransp, silent);
}

int CFreqGramVectorList::SearchInFile (int fd, const std::string s,
                                       int MaxErr, int *numComp,
                                       StringList * resList, bool hasTransp,
                                       bool silent)
{
    unsigned int gramDist = NgramMaxErrNum (MaxErr, NgramSize, VectorLength, hasTransp);
    myagrep a(s.c_str (), MaxErr, hasTransp);
    CFreqGramVector v (0, VectorLength - 1);
    v.InitFromString (s.c_str ());
    return SearchInFile (fd, 0, v, 0, 0, gramDist, a, numComp, resList, hasTransp, silent);
}


int CFreqGramVectorList::SearchInFile (int fd, off_t offset,
                                       const CFreqGramVector & v,
                                       int UpperPosDist, int UpperNegDist,
                                       int MaxErr, myagrep & a,
                                       int *numComp, StringList * resList,
                                       bool hasTransp, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return -1;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now read node vectors
    for (int i = 0; i < num; i++) {
        // Initializing vector from sbuf
        int StartIndex = *((u_int8_t *) (void *) sbuf);
        sbuf++;
        int EndIndex = *((u_int8_t *) (void *) sbuf);
        sbuf++;
        off_t sub_off = *((off_t *) (void *) sbuf);
        sbuf += sizeof (off_t);
        CFreqGramVector vn (StartIndex, EndIndex,
                            (u_int8_t *) (void *) sbuf);
        sbuf += vn.Size ();
        plen = Diff (vn, v, &pd, &nd);
        int _UpperPosDist = UpperPosDist + pd;
        int _UpperNegDist = UpperNegDist + nd;
        if (std::max(_UpperPosDist, _UpperNegDist) <= MaxErr) {
            // If EndIndex equals to v.EndIndex() then subtree is of type CStringHeap
            int res;
            if (v.EndIndex () > vn.EndIndex ()) {
                res =
                    SearchInFile (fd, sub_off,
                                  v.SubVector (v.StartIndex () + vn.Size (), v.EndIndex ()),
                                  _UpperPosDist, _UpperNegDist, MaxErr, a,
                                  numComp, resList, hasTransp, silent);
            } else
                res =
                    CStringHeap::SearchInFile (fd, sub_off, MaxErr, a,
                                               numComp, resList, hasTransp, silent);
            if (res < 0) {
                munmap (buf, buflen + delta);
                return -1;
            }
        }
    }
    if (munmap (buf, buflen + delta) < 0)
        return -1;
    return 1;
}

void CFreqGramVectorList::Search (const CFreqGramVector & v,
                                  int UpperPosDist, int UpperNegDist,
                                  const std::string s, int MaxErr,
                                  myagrep & a, int *numComp,
                                  StringList * resList, bool hasTransp,
                                  bool silent)
{
    int pd = 0, nd = 0, plen = 1;
    for (int i = 0; i < num; i++) {
        const CFreqGramVector & vn = vlist[i];
        plen = Diff (vn, v, &pd, &nd);
        int _UpperPosDist = UpperPosDist + pd;
        int _UpperNegDist = UpperNegDist + nd;
        //if (_UpperPosDist + _UpperNegDist <= MaxErr)
        if (std::max(_UpperPosDist, _UpperNegDist) <= MaxErr)
            plist[i]->Search (v.SubVector (v.StartIndex () + vn.Size (),
                                           v.EndIndex ()), 
                              _UpperPosDist, _UpperNegDist, 
                              s, MaxErr, a, numComp,
                              resList, hasTransp, silent);
    }
}

void CFreqGramVectorList::SearchExact (const std::string & s, int *numComp,
                                       StringList * resList, bool silent)
{
    myagrep a(s.c_str(), 0, 0); // Transpostion do not matter here kerr = 0

    CFreqGramVector v (0, VectorLength - 1);
    v.InitFromString (s.c_str ());
    //cout<<s<<":"<<" ";v.Print ();
    SearchExact (v, s, a, numComp, resList, silent);
}

void CFreqGramVectorList::SearchExact (const CFreqGramVector & v,
                                       const std::string & s, 
                                       myagrep &a, int *numComp,
                                       StringList * resList, bool silent)
{
    int b1 = 0, b2 = num - 1, b;
    int plen = 0;
    for (; b2 >= b1;) {
        b = (b2 + b1) / 2;
        int nd, pd;
        plen = Diff (vlist[b], v, &nd, &pd);
        if (plen == -1 || !(nd + pd)) {
            b1 = b2 = b;
            break;
        } else if (CompFunc (vlist[b], v) > 0) {
            b1 = b + 1;
        } else {
            b2 = b - 1;
        }
    }
    if (b2 == b1) {
        if (plen >= 0)
            plist[b1]->SearchExact (v.SubVector (v.StartIndex () + plen,
                                    v.EndIndex ()), 
                                    s, a,
                                    numComp,
                                    resList, silent);
        // If plen==-1, vectors are equal
        else
            plist[b1]->SearchExact (v, s, a, numComp, resList, silent);
    }

}

void CFreqGramVectorList::Insert (const std::string & s)
{
    CFreqGramVector v (0, VectorLength - 1);
    v.InitFromString (s.c_str ());
    Insert (v, s);
}

void CFreqGramVectorList::Insert (const CFreqGramVector & vi,
                                  const std::string & s)
{
// The case of empty node
    if (!num) {
// if (ptr==NULL) we initialize it 
        CStringHeap *ptr = new CStringHeap;
        ptr->Insert (s);
        vlist.insert (vlist.begin (), vi);
        plist.insert (plist.begin (), ptr);
        num++;
        return;
    }
// If not empty we find a frequency vector
// having a common prefix of at least MinPrefix long.
// If such a prefix does not exist, then v is inserted 
// in the list. The list is kept in a sorted order

// 1. Seeking the most closest vector
    int i, iclosest = -1, plen = 0, md = 256;
    for (i = 0; i < num; i++) {
        int nd, pd, d;
        int l = Diff (vi, vlist[i], &nd, &pd);


        d = (nd > pd) ? pd : nd;
        if (l == -1 || (nd + pd == 0))  // Equal vectors of equal size
        {
            if (nd + pd == 0 && l >= 0) {
                // We found a common prefix which equals to the 
                // node label (node frequency fector)
                // Must insert a string into this node
                CFreqGramVectorList * ptr = dynamic_cast<CFreqGramVectorList *>(plist[i]);
                assert (ptr);
                ptr->Insert(vi.SubVector(vi.StartIndex () + l,
                                         vi.EndIndex ()), s);
                return;
            }
            // Both node label and vector vi are equal
            assert (plist[i]);
            CStringHeap * ptr = dynamic_cast<CStringHeap *>(plist[i]);
            assert (ptr);
            ptr->Insert(s);
            return;
        }
        if ((l > plen) || (l == plen && d < md)) {
            plen = l;
            md = d;
            iclosest = i;
        }

    }
// 2. Prefix is long enough
    if (plen >= MinPrefix) {
        CFreqGramVector & vc = vlist[iclosest];
        // Prefix should be shorter than vector length
        assert (plen <= vi.EndIndex ());
        //assert(plen<=vc.EndIndex());
        CFreqGramVector vcsub(vc.SubVector(vc.StartIndex (),
                                           vc.StartIndex () + plen - 1));
        // Creating subnode 
        CFreqGramVectorList * vlst = new CFreqGramVectorList ();
        // Filling subnode with splitted vectors
        CFreqGramVector vs1 = vc.SubVector (vc.StartIndex () + plen, vc.EndIndex ());
        CFreqGramVector vs2 = vi.SubVector (vi.StartIndex () + plen, vi.EndIndex ());

        CStoreObj *ptr = plist[iclosest];
        CStringHeap *sh = new CStringHeap;
        sh->Insert (s);
        // We must keep the sub-node in a sorted order!
        if (CompFunc (vs1, vs2) > 0) {
            vlst->vlist.insert (vlst->vlist.end (), vs1);
            vlst->plist.insert (vlst->plist.end (), ptr);
            vlst->vlist.insert (vlst->vlist.end (), vs2);
            vlst->plist.insert (vlst->plist.end (), sh);
        } else {
            vlst->vlist.insert (vlst->vlist.end (), vs2);
            vlst->plist.insert (vlst->plist.end (), sh);
            vlst->vlist.insert (vlst->vlist.end (), vs1);
            vlst->plist.insert (vlst->plist.end (), ptr);
        }
        // Substituting the data at the current level
        vlst->num = 2;
        plist[iclosest] = vlst;
        vlist[iclosest] = vcsub;    // vcsub is used in equation operator
        // Here vcsub must be destroyed
        return;
    }
// 3. Prefix is too short -- just filling array in sorted (ascending) order
    TVectList::iterator vlast = vlist.begin ();
    TPtrList::iterator plast = plist.begin ();
    for (i = 0; i < num; i++, vlast++, plast++) {
        int d = CompFunc (vlist[i], vi);
        if (!i && d < 0) {
            break;
        }
        if (d > 0)
            if (i < num - 1)
                if (CompFunc (vlist[i + 1], vi) < 0) {
                    vlast++;
                    plast++;
                    break;
                }
    }
    vlist.insert (vlast, vi);
    CStringHeap * ptr = new CStringHeap;
    ptr->Insert (s);
    plist.insert (plast, ptr);
    num++;

}

#ifdef _DEBUG
void CFreqGramVectorList::Print (int level)
{
    TVectList::iterator v = vlist.begin ();
    TPtrList::iterator p = plist.begin ();
    for (; v != vlist.end (); v++, p++) {
        cout << "L=" << level << "\n";
        for (int i = 0; i < level; i++)
            cout << " ";
        (*v).Print ();
        if (*p)
            (*p)->Print (level + 1);
    }
}
#endif

int CFreqGramVectorList::DumpToFile (int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of sub-trees
// 2. the subvectors themselves with zero offsets
// 3. the sub-trees calculating and storing sub-tree offsets
    int cnt = 0;
    int buflen = 0;
    int sub_sz = 0;
    buflen += sizeof (cnt) + sizeof (buflen);

    TVectList::iterator v = vlist.begin ();
    for (; v != vlist.end (); v++, cnt++) {
        // Vector size + sub-tree offset size + vector indices 
        buflen += sizeof (offset) + (*v).Size () + 2;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf =
        mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    v = vlist.begin ();
    for (int i = 0; v != vlist.end (); v++, i++) {
        assert (i < cnt);
        // Indexes
        *((u_int8_t *) (void *) sbuf) = (*v).StartIndex ();
        sbuf++;
        *((u_int8_t *) (void *) sbuf) = (*v).EndIndex ();
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) (void *) sbuf;
        *ptrs[i] = 0;
        sbuf += sizeof (off_t);
        // Sub-vector
        (*v).CopyToMem ((void *) sbuf);
        sbuf += (*v).Size ();
    }
    // Now storing subtrees starting from offset = offset +sz 
    TPtrList::iterator p = plist.begin ();
    sub_sz = 0;
    for (int j = 0; p != plist.end (); p++, j++) {
        assert (j < cnt);
        *ptrs[j] = offset + buflen + sub_sz;
        // Diving into subtree
        int s_sz = (*p)->DumpToFile (fd, offset + buflen + sub_sz);
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Return the total number of bytes written
    return buflen + sub_sz;
}

int CFreqGramVectorList::GetVectorCount() 
{
    int res = 0;

    for (int i = 0; i < num; ++i)
        res += plist[i]->GetVectorCount();

    return res;
}

#ifdef _DEBUG
void CStringHeap::Print (int level)
{
    TStringList::iterator si = sl.begin ();
    for (; si != sl.end (); si++) {
        for (int i = 0; i < level; i++)
            cout << " ";
        cout << (*si) << "\n";
    }

}
#endif

void CStringHeap::Insert (const std::string & s)
{
#ifdef USE_CHUNK
    dataChunk.append(s);
    dataChunk.append(&termChar, 1);
    ++dataChunkCnt;
#else
    sl.insert (sl.end (), s);
#endif
}

void CStringHeap::SearchExact (const CFreqGramVector & v,
                               const std::string & s, 
                               myagrep &a, int *numComp,
                               StringList * resList, bool silent)
{
#ifdef USE_CHUNK
     a.findPattern(CInterface::filterType, termChar,
                     dataChunk.data(), dataChunk.data() + dataChunk.size(),
                     true, resList);
     if (numComp) *numComp += dataChunkCnt;
#else
    for (TStringList::iterator it = sl.begin (); it != sl.end (); it++) {
        if (numComp)
            (*numComp)++;
        if (s == (*it)) {
            if (resList)
                resList->push_back (*it);
            if (!silent) {
                cout << (*it) << "\n";
                return;
            }
        }
    }
#endif
}

void CStringHeap::Search (const CFreqGramVector & v, int UpperPosDist,
                          int UpperNegDist, const std::string s,
                          int MaxErr, myagrep & a, int *numComp,
                          StringList * resList, bool hasTransp, bool silent)
{
#ifdef USE_CHUNK
     a.findPattern(CInterface::filterType, termChar,
                     dataChunk.data(), dataChunk.data() + dataChunk.size(),
                     true, resList);
     if (numComp) *numComp += dataChunkCnt;
#else
    for (TStringList::iterator it = sl.begin (); it != sl.end (); it++) {
        if (numComp)
            (*numComp)++;
        if (a.checkRTRFilter(it->data(), it->data() + it->size())) {
            if (resList)
                resList->push_back (*it);
            if (!silent) {
                cout << (*it) << "\n";
            }
        }
    }
#endif
}




int CStringHeap::SearchInFile (int fd, off_t offset, int MaxErr,
                               myagrep & a, int *numComp,
                               StringList * resList, bool hasTransp, bool silent)
{
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    // First we must read the record size
    int cnt, buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return -1;
    // Now map to the file
    void *buf =
        (void *) mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
                       offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    buflen = *(int *)(void *)sbuf; 
    sbuf += sizeof(buflen);
    cnt = *(int *) (void *) sbuf;
    sbuf += sizeof (cnt);
#ifdef USE_CHUNK
     a.findPattern(CInterface::filterType, termChar,
                     sbuf, sbuf + buflen - sizeof(cnt) - sizeof(buflen),
                     true, resList);
     if (numComp) *numComp += cnt;
#else
    // Now we have access to the string buffer
    for (int i = 0; i < cnt; i++) {
        int len = strlen(sbuf);
        if (numComp)
            (*numComp)++;
        if (a.checkRTRFilter (sbuf, sbuf + len)) {
            if (resList)
                resList->push_back (sbuf);
            if (!silent) {
                cout << sbuf << "\n";
            }
        }
        sbuf += len + 1;
    }
#endif

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    return 1;
}

CStringHeap::~CStringHeap ()
{
}

int CStringHeap::GetVectorCount() 
{
    return 1;
}

int CStringHeap::DumpToFile (int fd, off_t offset)
{
#ifdef USE_CHUNK
    int sz = dataChunk.size();
    int cnt = dataChunkCnt;
#else
    int sz = 0;
    int cnt = 0;
    for (TStringList::iterator it = sl.begin (); it != sl.end (); it++, cnt++) {
        int len = strlen ((*it).c_str ()) + 1;
        sz += len;
    }
#endif

    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    int buflen = sizeof (cnt) + sizeof (sz) + sz;
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf =
        (void *) mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ,
                       MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    // First of all we write down the number of the strings
    // and record size
    char *sbuf = (char *) buf + delta;
    *(int *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) sbuf = cnt;
    sbuf += sizeof (cnt);
#ifdef USE_CHUNK
    memcpy(sbuf, dataChunk.data(), dataChunk.size());
    sbuf += dataChunk.size();
#else
    for (TStringList::iterator it = sl.begin (); it != sl.end (); it++) {
        int len = strlen ((*it).c_str ());
        memcpy (sbuf, (*it).c_str (), len + 1);
        sbuf += len + 1;
    }
#endif
    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Not return the number of used bytes
    return buflen;
}

