#ifndef _FREQVECTOR_H
#define _FREQVECTOR_H

#include <sys/types.h>
#include <string.h>
#include <iostream>
#include <assert.h>
#include "cgram.h"


extern int VectorLength;
/*
  IMPORTANT COMMENT!!!!
  Maximum vector size == 256
*/

extern int NgramSize;

// The frequency sub-vector class
class CFreqGramVector {
  protected:
    // The sub-vector itself
    VectorType * m_Vector;
    // The interval of indexes of the frequency vector 
    // which sub-vector contains
    u_int8_t m_StartIndex;
    u_int8_t m_EndIndex;

    void InitMem ();
    void FreeMem ();
  public:
    //=== Constructors, destructor, equation operator
     CFreqGramVector (int StartIndex = 0, int EndIndex = -1);   // Creates an empty vector
     CFreqGramVector (const CFreqGramVector &);
     CFreqGramVector (int StartIndex, int EndIndex, const VectorType *);
    ~CFreqGramVector ();
    const CFreqGramVector & operator= (const CFreqGramVector &);
    //==== SubVector
    CFreqGramVector SubVector (int si, int ei) const;
    // Copies only the vector content, returns the number of bytes written
    int CopyToMem (void *mem) const;

    //=== 
    u_int8_t StartIndex () const;
    u_int8_t EndIndex () const;
    u_int8_t Size () const;
    //==== Initialization of the sub-vector according to the contents 
    // of the string 
    void InitFromString (const char *str);
    void Print () const;
    //==== Comparisons
    friend int CompFunc (const CFreqGramVector & v1,
                         const CFreqGramVector & v2);
    /* 
       Diff returns the position of the first unequal digits
       it returns -1 in case of equal vectors
       Positive distance is the sum of digit differences
       where the second argument is less than the first
       Negative distance is the sum of the digit differences
       where the first argument is less than the second
     */
    friend int Diff (const CFreqGramVector & v1,
                     const CFreqGramVector & v2, int *posdist,
                     int *negdist);
};

int CompFunc (const CFreqGramVector & v1, const CFreqGramVector & v2);
int Diff (const CFreqGramVector & v1, const CFreqGramVector & v2,
          int *posdist, int *negdist);

#endif
