#include "index.h"

int NgramSize = _NGRAM_SIZE;

CFreqTrieIndex::CFreqTrieIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false)
{
   GetEnv("NGRAM_LEN", NgramSize); 
   cout << "Ngram size: " << NgramSize << endl;

}

CFreqTrieIndex::~CFreqTrieIndex ()
{
    if (opened)
        CloseDB ();
}

bool CFreqTrieIndex::Insert (const char *str)
{
    ind.Insert (str);
    return true;
}

bool CFreqTrieIndex::Search (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent)
{
    (kerr != 0) ? ind.Search (str, kerr, numComp, resList,
                              mHasTransp, silent) : 
                  ind.SearchExact (str, numComp, resList, silent);
    return true;
}

bool CFreqTrieIndex::SearchInDB (const char *str, int kerr, int *numComp,
                                 StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return ind.SearchInFile (fd, str, kerr, numComp, resList, mHasTransp, silent);
}

bool CFreqTrieIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CFreqTrieIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CFreqTrieIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = ind.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

void CFreqTrieIndex::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
     if (!kerr && type == SearchTypeMem) { // Print once
        ostr << "    The number of vectors: " <<  ind.GetVectorCount() << endl;
     }
}

#ifdef _DEBUG
void CFreqTrieIndex::Print ()
{
    node.Print ();
}
#endif
