#ifndef _STOREOBJ_H
#define _STOREOBJ_H

#include <unistd.h>
#include <sys/mman.h>

#include <vector>
#include <string>
#include "freqgramvector.h"
#include "myagrep.h"
#include "interface.h"



class CStoreObj {
  protected:
  public:
    virtual void SearchExact (const CFreqGramVector & v,
                              const std::string & s, myagrep &a,
                              int *numComp,
                              StringList * resList, bool silent) = 0;
    virtual void Search (const CFreqGramVector & v, int UpperPosDist,
                         int UpperNegDist, const std::string s, int MaxErr,
                         myagrep & a, int *numComp, StringList * resList,
                         bool hasTransp, bool silent) = 0;
    // returns the size of the dumped structure if successful or -1 otherwise
    virtual int DumpToFile (int fd, off_t offset) = 0;
    virtual int GetVectorCount() = 0;
    virtual ~CStoreObj(){}
#ifdef _DEBUG
    virtual void Print (int) = 0;
#endif
};

typedef CStoreObj * PCStoreObj;
typedef std::vector <CFreqGramVector> TVectList;
typedef std::vector <PCStoreObj> TPtrList;

using namespace std;

class CFreqGramVectorList:public CStoreObj {
  protected:
    static const int MinPrefix = 4;
    int num;
    TVectList vlist;
    TPtrList  plist;
// === Searching/inserting private
    void Insert (const CFreqGramVector & v, const std::string & s);
    void SearchExact (const CFreqGramVector & v, const std::string & s,
                      myagrep &a,
                      int *numComp, StringList * resList, bool silent);
    void Search (const CFreqGramVector & v, int UpperPosDist,
                 int UpperNegDist, const std::string s, int MaxErr,
                 myagrep &, int *numComp, StringList * resList, 
                 bool hasTransp, bool silent);
  public:
    // === Searching/inserting public
    ~CFreqGramVectorList();
    void SearchExact (const std::string & s, int *numComp,
                      StringList * resList, bool silent);
    void Search (const std::string s, int MaxErr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent);
    CFreqGramVectorList ():num (0) {;
    } void Insert (const std::string & s);
    // === Working directly with file on disk
    int DumpToFile (int fd, off_t offset);
    static int SearchInFile (int fd, const std::string s, int MaxErr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    static int SearchInFile (int fd, off_t offset,
                             const CFreqGramVector & v, int UpperPosDist,
                             int UpperNegDist, int MaxErr, myagrep & a,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    int GetVectorCount();
#ifdef _DEBUG
    void Print (int level = 0);
#endif
};

typedef std::vector<std::string> TStringList;

class CStringHeap:public CStoreObj {
  protected:
    void SearchExact (const CFreqGramVector & v,
                      const std::string & s, myagrep &a,
                      int *numComp, StringList * resList, bool silent);
    void Search (const CFreqGramVector & v, int UpperPosDist,
                 int UpperNegDist, const std::string s, int MaxErr,
                 myagrep & a, int *numComp, StringList * resList,
                 bool hasTransp, bool silent);
    static const char termChar;
  public:
#ifdef USE_CHUNK
    int    dataChunkCnt;
    string dataChunk;
#else
    TStringList sl;
#endif
    void Insert (const std::string &);
    // === Working directly with file on disk
    int DumpToFile (int fd, off_t offset);
    static int SearchInFile (int fd, off_t offset, int MaxErr, myagrep & a,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
#ifdef USE_CHUNK
    CStringHeap() : dataChunkCnt(0) {}
#endif
     ~CStringHeap();
    int GetVectorCount();
#ifdef _DEBUG
    virtual void Print (int);
#endif
};


#endif
