#include "index.h"
#ifndef _USE_BDB
#include "len1indexNoBDB.h"
#define DB_CLASS CLen1IndexNoBDB
#else
#include "len1indexBDB.h"
#define DB_CLASS CLen1IndexBDB
#endif

DEFINE_CLASS_STATIC_COUNTER(CNgramIndex, aseek);

const int CNgramIndex::MaxLen;

CNgramIndex::CNgramIndex (bool hasTransp, EAlphabet AlphaType):opened (false), CInterface(hasTransp, AlphaType), dbenv(NULL), NgramLen(4), pCompDict(NULL), pDictBuff(NULL), DictBuffPos(0)
{
    GetEnv("NGRAM_LEN", NgramLen);
    cout << "Qgram size: " << NgramLen << endl;
    for (int i = 1; i <= MaxLen; i++) {
        db[i] = new DB_CLASS(NgramLen, 
                        mHasTransp, 
                        HAS_COMPRESS);
#if HAS_COMPRESS
        db[i]->setUseDelta(USE_DELTA);
#endif
    }
    for (int k = 1; k <= MAX_KERR; ++k) {
        pSeqIndex[k - 1] = new CAgrepIndex(hasTransp, AlphaType);
    }
}

CNgramIndex::~CNgramIndex () {
    /*for (unsigned int i = 0; i < dict.size (); i++)
        delete[]dict[i];
    */
    for (int i = 1; i <= MaxLen; i++)
        delete db[i];
    for (int k = 1; k <= MAX_KERR; ++k) {
        delete pSeqIndex[k - 1];
    }
    delete [] pCompDict;
    delete [] pDictBuff;
}

bool CNgramIndex::Insert (const char *str)
{
    // The string must be dupplicated and stored into 
    // an array variable dict
    
    int len = strlen (str);
    
    for (int k = 1; k <= MAX_KERR; ++k) {
        if (len <  DB_CLASS::SeqIndexBound(k) + k &&
            len >= DB_CLASS::SeqIndexBound(k - 1) + k - 1) {
            if (!pSeqIndex[k - 1]->Insert (str)) {
                return false;
            };
        }
    }

    len = (len > MaxLen) ? MaxLen : len;

    if (!len)
        return false;
    char *ptr = NULL;
    if (len < NgramLen) {
        ptr = AllocFromBuff(NgramLen + 1);

        strcpy (ptr, str);
        for (int i = len; i <= NgramLen; i++)
            ptr[i] = 0;
    } else {
        ptr = AllocFromBuff(len + 1);

        strncpy(ptr, str, len);  // We may want to shorten a string
        ptr[len] = 0;
    }
    // Let's insert an entry into a dictionary of strings
    dict.push_back (ptr);
    return bool (db[len]->Insert (ptr));
}

bool CNgramIndex::Search (const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    int slen = strlen (str);
    slen = (slen > MaxLen) ? MaxLen : slen;
    myagrep a(str, kerr, mHasTransp);
    
    /* Cannot use the main n-gram index, try auxilliary indicies */
    if (slen < DB_CLASS::SeqIndexBound(kerr)) {
        for (int k = 1; k <= kerr; ++k) {
            if (!pSeqIndex[k - 1]->Search (str, kerr, numComp, resList, silent)) {
                return false;
            }
        }

        return true;
    }

    for (int i = slen - kerr; i <= slen + kerr; i++) {
        if (i < 1 || i > MaxLen + 1)
            continue;
        int h = min (i, MaxLen);

        bool stat = (HAS_COMPRESS) ? 
         db[h]->SearchComp(pCompDict, str, a, kerr, numComp, resList, silent)
        :db[h]->Search (str, kerr, a, numComp, resList, silent);

        if (!stat) { return false; }
    }

    return true;
}

bool CNgramIndex::SearchInDB (const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is closed.\n";
        return false;
    }
    if (!mapDict) {
    mapDict =
        (char *) mmap (NULL, mapSize, PROT_READ, MAP_SHARED, fd, 0);
    if ((void *) mapDict == MAP_FAILED) {
        Perror ("Error memory mapping!");
        return false;
    }
    }

    bool    res = true;
    int     slen = strlen (str);
    myagrep a(str, kerr, mHasTransp);

    if (slen < DB_CLASS::SeqIndexBound(kerr)) {
        START_COUNTER(aseek);
        for (int k = 1; k <= kerr; ++k) { 
            if (!pSeqIndex[k - 1]->SearchInDB (str, kerr, numComp, resList, silent)) {
                res = false;
            }
        }
        STOP_COUNTER(aseek);
    } else {
       
        int imin = slen - kerr, imax = slen + kerr;
        if (imin < 1)
            imin = 1;
        if (imax > MaxLen)
            imax = MaxLen;
        for (int i = imin; i <= imax; i++) {
            if (!db[i]->SearchInDB (mapDict, str, a, kerr, numComp, resList, silent)) {
                res = false;
                goto exit;
            }
        }
    }
exit:
    return res;
}

bool CNgramIndex::OpenDB (const char *name, bool create)
{
    if (opened) {
        CloseDB ();
    }
    for (int k = 1; k <= MAX_KERR; ++k) {
        if (!pSeqIndex[k - 1]->OpenDB(seqIndexName (name, k).c_str (), create)) {
            return false;
        }
    }

    baseName = name;
    mapDict = NULL;
    fd = 0;

    for (int i = 1; i <= MaxLen; i++) {
        char numb[32];
        snprintf (numb, sizeof (numb), "%d", i);
        std::string index_name = std::string (name) + numb + ".grm";
        if (!db[i]->OpenDB(index_name.c_str (), create)) {
            return false;
        }
    }
    if (!create) {
        mapSize = FileSize (name);;
        if (mapSize < 0) {
            return false;
        }
        if (!mapSize)
            return opened = true;   // Empty dictionary
        fd = open (name, O_RDONLY);
        if (fd < 0) {
            return false;
        }
    } else {
        fd = open (name, O_CREAT | O_TRUNC | O_RDWR, 0664);
        if (fd < 0) {
            Perror ("Cannot open file!");
            return false;
        }
    }

    return opened = true;
}

bool CNgramIndex::CloseDB ()
{
    if (!opened)
        return true;
    for (int k = 1; k <= MAX_KERR; ++k) {
        if (!pSeqIndex[k - 1]->CloseDB ()) {
            return false;
        }
    }
    for (int i = 1; i <= MaxLen; i++)
        if (!db[i]->CloseDB ())
            return false;

    if (mapDict && mapDict != MAP_FAILED) {
        if (munmap (mapDict, mapSize)) {
            Perror("munmap failed");
            return false;
        }
    }

    mapDict = NULL;

    if (fd)
        close (fd);

    opened = false;
    return true;
}


bool CNgramIndex::DumpDB (const char *name)
{
    char numb[32];
    if (!OpenDB (name, true))
        return false;           // create = true

    for (int k = 1; k <= MAX_KERR; ++k) {
        if (!pSeqIndex[k - 1]->DumpDB (seqIndexName (name, k).c_str ())) {
            return false;
        }
    }
    // First of all we must dump the dictionary itself
    mOffset off_list;
    for (PtrListVector::iterator it = dict.begin (); it != dict.end ();
         it++) {
        off_t cur = lseek (fd, 0, SEEK_CUR);
        if (cur < 0)
            return false;
        off_list[*it] = cur;
        if (write (fd, *it, strlen (*it)) < (ssize_t) strlen (*it))
            return false;
        write (fd, "\0", 1);
    }


    for (int i = 1; i <= MaxLen; i++) {
        snprintf (numb, sizeof (numb), "%d", i);
        std::string index_name = std::string (name) + numb + ".grm";
        if (!db[i]->DumpDB (index_name.c_str (), off_list))
            return false;
    }
    return CloseDB ();
}

void CNgramIndex::ResetCounters()
{
    RESET_COUNTER(aseek);
    for (int i = 1; i <= MaxLen; i++) {
        db[i]->ResetCounters();
    }
}

void CNgramIndex::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
    if (!kerr) {
        for (int i = 1; i <= MaxLen; i++) {
            ostr << "Length: " << i << " ";
            db[i]->PrintAddStat(testNum, kerr, ostr);
            ostr << "\n";
        }
    }
    //ostr << "    Time to seek sequentially (ms): " << GET_COUNTER_MS(aseek);
}

bool CNgramIndex::CreateInMemIndex (const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    system("./clear_db.sh");

    DictBuffSize = 0;
    string s;

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }

    vector<string>  vList;


    while (f >> s) {
        toUpper (s);

        DictBuffSize += max(s.size(), (size_t)NgramLen) + 1;

        vList.push_back(s);
    }

    if (dcnt) {
        *dcnt  = vList.size();
    }

    pDictBuff   = new char[DictBuffSize];
    DictBuffPos = 0;
        
    for (size_t i = 0; i < vList.size(); ++i) {
        Insert(vList[i].c_str());

        if (!silent && !((i) % 10000))
            cout << (i) << "\n";
    }



    if (HAS_COMPRESS) {
        size_t  TotalSize = 0;
        // Merge dictionary strings first
        mOffset off_list;

        for (PtrListVector::iterator it = dict.begin (); it != dict.end (); it++) {
          TotalSize += strlen(*it) + 1;
        }

        pCompDict = new char[TotalSize];

        off_t cur = 0;
        for (PtrListVector::iterator it = dict.begin (); it != dict.end (); it++) {
            size_t      len = strlen(*it) + 1;

            memcpy(pCompDict + cur, *it, len);
            off_list[*it] = cur;
            cur += len;
        }

        for (int i = 1; i <= MaxLen; i++) {
            db[i]->CompressInMem(off_list);
        }
    }

    return true;
}

#ifdef _DEBUG
void CNgramIndex::Print ()
{
    for (int i = 1; i <= MaxLen; i++) {
        cout << "{{{ " << i << "\n";
        db[i]->Print ();
        cout << "}}} " << i << "\n";
    }
}
#endif
