#include "index.h"
#include "myagrep.h"
#ifdef USE_CHECKLEVEN
#include "edistCut.h"
#endif

#include <string.h>

#define USE_RTR_FILTER // It is really faster with a filter

// do not change this
const char CAgrepIndex::termChar = 0;

CAgrepIndex::CAgrepIndex (bool hasTransp, EAlphabet AlphaType):CInterface(hasTransp, AlphaType), opened (false), fd (0), recNum(0)
{
}

CAgrepIndex::~CAgrepIndex ()
{
    if (opened) {
        CloseDB ();
    }
}


bool CAgrepIndex::Insert (const char *str)
{
    wordList.push_back(str);
    ++recNum;
    return true;
}

bool CAgrepIndex::Search(const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
#ifndef USE_CHECKLEVEN
    myagrep a(str, kerr, mHasTransp);
#else
    string cmpStr(str);
#endif
    cntFnd = 0;
    for (unsigned int i = 0; i < wordList.size (); i++) {
        if (numComp)
            (*numComp)++;
#ifndef USE_CHECKLEVEN
#ifdef  USE_RTR_FILTER
        if (a.checkRTRFilter(wordList[i])) {
#else
        if (a.checkRTR(wordList[i].c_str(), wordList[i].c_str() + wordList[i].size())) {
#endif
#else
        if (checkDistLeven(mHasTransp, cmpStr, wordList[i], kerr) <= kerr) {
#endif
            if (!silent) {
                cout << wordList[i] << endl;
            }
            if (resList != NULL) {
                resList->push_back(wordList[i]);
            }
            cntFnd++;
        }
    }
    return true;
}

bool CAgrepIndex::SearchInDB(const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    off_t fsz = FileSize (Name.c_str ());
    if (fsz < 0) {
        return false;
    }
    if (fsz == 0) {
        return true;
    }
#ifndef USE_CHECKLEVEN
    myagrep a(str, kerr, mHasTransp);
#else
    string cmpStr(str);
#endif

    if (lseek (fd, 0, SEEK_SET)) {
        Perror ("Cannot lseek");
        return false;
    }
    void *buf = mmap (NULL, fsz, PROT_READ, MAP_SHARED, fd, 0);
    if (buf == MAP_FAILED) {
        Perror ("Cannot mmap!");
        return false;
    }
    int len = 0;
    char *buf1 = (char *) buf + sizeof recNum;
    recNum = *((int *)buf);
#ifndef USE_CHECKLEVEN
    a.findPattern(CInterface::filterType, termChar,
                  buf1, buf1 + fsz - sizeof recNum, 
                  true, resList); 
    if (resList) cntFnd = resList->size();
    if (numComp) *numComp = -1; // undefined
#else
    cntFnd = 0;
    do {
        len = strlen (buf1);
        if (len >= 0) {
            if (numComp)
                (*numComp)++;
            if (checkDistLeven(mHasTransp, cmpStr, buf1, kerr) <= kerr) {
                cntFnd++;
                if (!silent)
                    cout << buf1 << endl;
                if (resList != NULL)
                    resList->push_back (string (buf1));
            }
            buf1 += len + 1;
        };
    } while (buf1 < (char *) buf + fsz);
#endif
    if (munmap (buf, fsz)) {
        Perror ("Cannot unmap");
    }

    return true;
}


bool CAgrepIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
//#define O_DIRECT 040000
        fd = open (file_name, O_RDONLY);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    Name = file_name;
    return (opened = true);
}

bool CAgrepIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (fd) {
        if (close (fd) != 0)
            return false;
    }
    fd = 0;
    opened = false;
    return true;
}


bool CAgrepIndex::DumpDB (const char *file_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    if (write(fd, &recNum, sizeof recNum) != sizeof recNum) {
        return false;
    }
    for (unsigned int i = 0; i < wordList.size (); i++) {
        int len = wordList[i].size();
        if (write (fd, wordList[i].data(), len) != len)
            return false;
        if (write (fd, &termChar, 1) != 1)
            return false;
    }
    bool f = CloseDB ();
    return f;
}
