#include "index.h"

// here you can change this
const char CLevenDFAIndex::termChar = 0;

DEFINE_CLASS_STATIC_COUNTER(CLevenDFAIndex, cr_dfa)

CLevenDFAIndex::CLevenDFAIndex (bool hasTransp, EAlphabet AlphaType):CInterface(hasTransp, AlphaType), opened (false), fd (0), recNum(0)
{
}

CLevenDFAIndex::~CLevenDFAIndex ()
{
    if (opened) {
        CloseDB ();
    }
}


bool CLevenDFAIndex::Insert (const char *str)
{
    wordList.push_back(str);
    ++recNum;
}

bool CLevenDFAIndex::Search(const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{

    const char * ABC = GetAlphabet(mAlphaType);
    START_COUNTER(cr_dfa)
    LevenDFA *pDfa = new LevenDFA(ABC, str, kerr, mHasTransp);
    STOP_COUNTER(cr_dfa)


    int sz = wordList.size();

    for (int i = 0; i < sz; ++i) {
        pDfa->Start();

        bool   final = false;
        string &s = wordList[i];
        int    sz1 = s.size();

        for (int j = 0; j < sz1; ++j) {
            if (!pDfa->Next(s[j], final)) goto nt;
        }
        pDfa->Next(LevenDFA::EOS, final);
nt:
        if (final) {
            if (resList) resList->push_back(s);
        }
    }

    if (resList) {
        cntFnd = resList->size();
    }
    if (numComp) (*numComp) = recNum;

    START_COUNTER(cr_dfa);
    delete pDfa;
    STOP_COUNTER(cr_dfa);
    
    return true;
}

bool CLevenDFAIndex::SearchInDB(const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    off_t fsz = FileSize (Name.c_str ());

    const char * ABC = GetAlphabet(mAlphaType);
    START_COUNTER(cr_dfa)
    LevenDFA a(ABC, str, kerr, mHasTransp);
    STOP_COUNTER(cr_dfa)

    if (fsz < 0) {
        return false;
    }

    if (lseek (fd, 0, SEEK_SET)) {
        Perror ("Cannot lseek");
        return false;
    }
    void *buf = mmap (NULL, fsz, PROT_READ, MAP_SHARED, fd, 0);
    if (buf == MAP_FAILED) {
        Perror ("Cannot mmap!");
        return false;
    }
    int len = 0;
    char *buf1 = (char *) buf + sizeof recNum;
    recNum = *((int *)buf);

    cntFnd = 0;
    do {
        len = strlen (buf1);
        a.Start();
        if (len >= 0) {
            if (numComp)
                (*numComp)++;

            bool final = false;

            for (int j = 0; j < len; ++j) {
                if (!a.Next(buf1[j], final)) goto nt;
            }
            a.Next(LevenDFA::EOS, final);
nt:
            if (final) {
                cntFnd++;
                if (!silent)
                    cout << buf1 << endl;
                if (resList != NULL)
                    resList->push_back (string (buf1));
            }
            buf1 += len + 1;
        };
    } while (buf1 < (char *) buf + fsz);

    if (munmap (buf, fsz)) {
        Perror ("Cannot unmap");
    }

    if (fsz == 0) {
        return true;
    }

    return true;
}


bool CLevenDFAIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
//#define O_DIRECT 040000
        fd = open (file_name, O_RDONLY);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    Name = file_name;
    return (opened = true);
}

bool CLevenDFAIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (fd) {
        if (close (fd) != 0)
            return false;
    }
    fd = 0;
    opened = false;
    return true;
}


bool CLevenDFAIndex::DumpDB (const char *file_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    if (write (fd, &recNum, sizeof recNum) != sizeof recNum) {
        return false;
    }
    for (unsigned int i = 0; i < wordList.size (); i++) {
        int len = wordList[i].size();
        if (write (fd, wordList[i].data(), len) != len)
            return false;
        if (write (fd, &termChar, 1) != 1)
            return false;
    }
    bool f = CloseDB ();
    return f;
}

void CLevenDFAIndex::PrintAddStat(SearchType type, int testNum, int kerr, ostream & str) 
{
    if (type == SearchTypeMem && kerr) {
        double tm = GET_COUNTER_MS(cr_dfa);
        tm /= testNum;
        str << "Kerr = " << kerr << " Average DFA cr time: " << tm ;
    }
}

void CLevenDFAIndex::ResetCounters() {
    RESET_COUNTER(cr_dfa);
}
