#include "block.h"

void FBlock::Clear () 
{
    memset (mem, 0, BlockSize);
    BlockAddr = 0;
    rptr = NULL;
    FBlockHeader r;            // default constructor clears everything, look h-file
    memcpy (mem, (void huge *) &r, sizeof (r));
};

FBlock::FBlock (const FBlock & b):BlockAddr (0), rptr (NULL) 
{
    InitMem ();
    memcpy (mem, b.mem, BlockSize);
    BlockAddr = b.BlockAddr;
};

FBlock::FBlock (const char *ptr, int32 b_addr):BlockAddr (b_addr), rptr (NULL) 
{
    mem = (void *) ptr;
    alloc_mem = false;
};

FBlock & FBlock::operator= (const FBlock & b) 
{
    if (this != &b) {
        memcpy (mem, b.mem, BlockSize);
        rptr = NULL;
        BlockAddr = b.BlockAddr;
    }
    return (*this);
};

void FBlock::InitMem () 
{
    mem = new char[BlockSize];
    if (!mem) {
        fprintf (stderr, "Not enough memory!");
        abort ();
    }                           //throw "Not enough virtual memory!";
    alloc_mem = true;
}

FBlock::FBlock ():BlockAddr (0), rptr (NULL) 
{
    InitMem ();
    Clear ();
};

FBlock::~FBlock () 
{
    if (!alloc_mem)
        return;                 //Memory is taken without allocating
    delete[](char *) mem;
};

bool FBlock::Write (int fd, int32 addr) 
{
    if (lseek (fd, addr, SEEK_SET) < 0)
        return false;
    if (write (fd, mem, BlockSize) != BlockSize)
        return false;
    BlockAddr = addr;
    return true;
};

bool FBlock::Read (int fd, int32 addr) 
{
    if (lseek (fd, addr, SEEK_SET) < 0)
        return false;
    int32 size = read (fd, mem, BlockSize);
    if (size < BlockSize) {
        fprintf (stderr, "Only %ld read of the block at addr %ld\n", size, addr);
        return false;
    }
    if (((FBlockHeader huge *) mem)->CheckMark != BMark) {
        fprintf (stderr, "Internal error - marks differ %d\n", ((FBlockHeader huge *) mem)->CheckMark);
        return false;
    }
    BlockAddr = addr;
    return true;
};

bool FBlock::AppendRecord (const char huge * str, const Sign huge * sgn)
{
    FBlockHeader huge * h = (FBlockHeader huge *) mem;
    int32 len = strlen (str) + 1;
    if (len > FreeSpace ())
        return false;
    memcpy (((char huge *) mem) + h->free_pos, str, len);
    h->free_pos += len;
    return true;
};

char huge * FBlock::ToFirst (Sign huge * s) 
{
    FBlockHeader huge * h = ((FBlockHeader huge *) mem);
    if (h->free_pos > sizeof (FBlockHeader)) {
        rptr = ((char huge *) mem) + sizeof (FBlockHeader);
        // s Can be NULL!!!
        //if (s) *s=Sign(* ((int16 *)rptr) );
        return rptr;
    } else return NULL;
};

char huge * FBlock::ToNext (Sign huge * s) 
{
    FBlockHeader huge * h = (FBlockHeader huge *) mem;
    int32 len = strlen (rptr) + 1;
    if (rptr - ((char huge *) mem) < h->free_pos - len) {
        rptr += len;
        
        // s Can be NULL!!!
        //if (s) *s=Sign(* ((int16 *)rptr) );
        return rptr;        //+sizeof(int16);
    } else return NULL;
};

