#include "calchash.h"
#include <iostream>
#include <algorithm>
#include "langcase.h"

freqcmp varfreqcmp;

using namespace std;

bool CalcHMap (int m, const vecfreqs & infdata, HMap & hmp, bool print)
{
    // Copying data
    vecfreqs fdata (infdata.begin (), infdata.end ());
    // Nothing to do
    if (!fdata.size ()) {
        if (print) {
            std::cerr << "Empty statistics, exiting!" << std::endl;
        }
        return false;
    }
    // First of all sort fdata by the field freq in ascending order
    sort(fdata.begin (), fdata.end (), varfreqcmp);
    hmp.clear_mem ();
    double esum = 1;
    for (int i = 0; i < m - 1; i++) {
        double reps = esum / (m - i);
        int melem = fdata.size () - 1;
        if (melem < 0)
            return false;       // if occasionally array is empty than return false;
        // First of all check whether the largest frequency is greater than esum/(m-i);
        if (fdata[melem].freq > reps) {
            hmp.setValue (uchar (tolower_eng_rus (fdata[melem].c)), i + 1);
            hmp.setValue (uchar (toupper_eng_rus (fdata[melem].c)), i + 1);
            esum -= fdata[melem].freq;
            fdata.erase (fdata.begin () + melem);
            continue;
        } else
            // Now we will put in the bucket i elements while 
        {
            double ts = 0;
            for (int j = melem; j >= 0; j--)
                if (ts + fdata[j].freq <= reps) {
                    ts += fdata[j].freq;
                    hmp.setValue (uchar (tolower_eng_rus (fdata[j].c)), i + 1);
                    hmp.setValue (uchar (toupper_eng_rus (fdata[j].c)), i + 1);
                    fdata.erase (fdata.begin () + j);
                    // After erasing we do not need to decrease j;
                }
            esum -= ts;
        }
    }
    /* Special case i==m: We just put everything left in the bucket number m */
    for (int j = 0; j < fdata.size (); j++) {
        hmp.setValue (uchar (tolower_eng_rus (fdata[j].c)), m);
        hmp.setValue (uchar (toupper_eng_rus (fdata[j].c)), m);
    }

    /*
       Now printing (for debug purposes only)
     */
    if (print) {
        for (int i = 1; i <= m; i++) {
            std::cout << "Bucket " << i << ", ";
            double bs = 0;
            for (int j = 0; j < infdata.size (); j++)
                if (hmp[uchar (infdata[j].c)] == i)
                    bs += infdata[j].freq;
            std::cout << "sum=" << bs << " :";
            for (int j = 0; j < infdata.size (); j++) {
                if (hmp[uchar (infdata[j].c)] == i)
                    std::cout << infdata[j].c << " " << infdata[j].freq << "  ";
            }
            std::cout << std::endl;
        }
    }
    return true;
};
