#include "afuzzycontext.h"

context::context ():k (0), R (NULL), RP (NULL), cnt (0), startpos (0)
{
};

void context::init (int _k)
{
    k = _k;
    R = new Uint[k + 1];
    RP = new Uint[k + 1];
    cnt = startpos = 0;
}

context::~context ()
{
    if (RP)
        delete[]RP;
    if (R)
        delete[]R;
    cnt = startpos = k = 0;
}

context::context (const context & c):k (c.k), cnt (c.cnt),
startpos (c.startpos)
{
    R = new Uint[k + 1];
    RP = new Uint[k + 1];
    memcpy ((void *) R, (void *) c.R, (k + 1) * sizeof (Uint));
    memcpy ((void *) RP, (void *) c.RP, (k + 1) * sizeof (Uint));
}

const context & context::operator= (const context & c)
{
    if (this == &c)
        return (*this);
    if (R)
        delete[]R;
    if (RP)
        delete[]RP;
    R = new Uint[c.k + 1];
    RP = new Uint[c.k + 1];
    memcpy ((void *) R, (void *) c.R, (c.k + 1) * sizeof (Uint));
    memcpy ((void *) RP, (void *) c.RP, (c.k + 1) * sizeof (Uint));
    k = c.k;
    cnt = c.cnt;
    startpos = c.startpos;
    return (*this);
}


afuzzycontext::afuzzycontext ():S (NULL), R1 (NULL), RP (NULL), RI (NULL)
{;
}

void afuzzycontext::init (const std::string & p, int kerr, context * cntx)
{
    k = kerr;
    m = p.length ();
    mask_ok = 1 << (m - 1);
    memset (Map, 0, sizeof (Map));
    if (S)
        delete[]S;
    if (R1)
        delete[]R1;
    if (RP)
        delete[]RP;
    if (RI)
        delete[]RI;
    S = new Uint[m + 1];
    R1 = new Uint[k + 1];
    RP = new Uint[k + 1];
    RI = new Uint[k + 1];
    cntx->init (k);
    S[0] = 0;
    cntx->R[0] = 0;
    cntx->RP[0] = 0;
    RP[0] = R1[0] = RI[0] = 0;
    for (int i = 0, cnt = 0; i < m; i++) {
        int l = Map[(unsigned char) p[i]];
        if (!l) {
            l = Map[(unsigned char) p[i]] = ++cnt;
            S[l] = 0;
        }
        S[l] |= 1 << i;
    }


    for (int d = 0; d <= k; d++)
        RI[d] = (1 << d) - 1;

    r_size = sizeof (Uint) * (k + 1);
}

afuzzycontext::~afuzzycontext ()
{
    if (S)
        delete[]S;
    if (R1)
        delete[]R1;
    if (RP)
        delete[]RP;
    if (RI)
        delete[]RI;
}

/**  */
bool afuzzycontext::checkRTR (const std::string & t, context * cntx,
                              bool final)
{
    register int cnt1;
    if (!cntx->startpos) {
        memcpy (cntx->R, RI, r_size);   // R = RI
        cntx->cnt = 0;
    }

    register int j = cntx->startpos;
    Uint SMapP = (j != 0) ? (S[Map[(unsigned char) t[j - 1]]]) : 0;
    for (; j < t.length (); j++) {
        cnt1 = cntx->cnt;
        register Uint SMap = S[Map[(unsigned char) t[j]]];
        for (register int d = cntx->cnt; d <= k; d++) {
            R1[d] = (((cntx->R[d] << 1) | (j <= d)) & SMap);
            if (d > cntx->cnt) {
                R1[d] |=
                    ((cntx->R[d - 1] | R1[d - 1]) << 1) | cntx->R[d -
                                                                  1] | (j
                                                                        <=
                                                                        d -
                                                                        1);
                /* Now transpositions */
#if HAS_TRANSP
                if (j > 0 && d)
                    R1[d] |=
                        (cntx->
                         RP[d -
                            1] << 2 | ((j <=
                                        d) << 1)) & SMapP & (SMap << 1);
#endif
            }
            if (j > d && !R1[d] && !(cntx->R[d]))
                cnt1 = cntx->cnt + 1;
        }                       // end for(register int d=cnt ...



        for (register int i = cntx->cnt; i <= k; i++) {
            // R->RP, R1->R
            cntx->RP[i] = cntx->R[i];
            cntx->R[i] = R1[i];
        }

        cntx->cnt = cnt1;
        SMapP = SMap;

    }                           // end for(register int j=0 ...
    cntx->startpos = t.length ();   // Moving start position
    //Uint mask1 = 1<<(t.length()-1);

    //return cntx->R[k] & ( (final) ? mask_ok: mask1 );
    if (final)
        return (cntx->R[k] & mask_ok) != 0;
    else
        return cntx->cnt <= k;
}
