#ifndef  _LEAF
#define  _LEAF

#include "trie.h"
#include "node.h"

class TrieNode;
/* 
  The implementaion of the leaf class
  Every node could contain several leaves, 
  and each leaf, in turn, could contain subtree (node)
*/
class TrieLeaf {
  protected:
    TrieNode * SubTree;
    /* All leafs in the node are chained       */
    /* The address of the nextleaf in the list */
    TrieLeaf *NextLeaf,
        /* An upper node address */
    *UpperLink;
    /* The part of the string stored in the path: from node root to the leaf subtree */
     std::string KeyPart;
  public:
/* Constructor */
     TrieLeaf (std::string = EOS, TrieLeaf * UpperLink =
               NULL, TrieNode * SubTree = NULL);
/* Destructor  */
    ~TrieLeaf ();
    TrieLeaf *Next () {
        return NextLeaf;
    };
/* Function appends a new leaf to list of node leavs */
    TrieLeaf *Append (std::string, TrieLeaf * UpperLink =
                      NULL, TrieNode * _SubTree = NULL);
/* Function inserts string into the leaf subtree */
    TrieLeaf *Insert (std::string, TrieNode * SubTree = NULL);
/* Function checks whether the KeyPart and Key have a common  prefix */
    int CheckCommon (const std::string &Key);
    inline int min (int a, int b) {
        return (a <= b) ? a : b;
    }
    const std::string & Key () const {
        return KeyPart;
    } TrieNode *Sub () {
        return SubTree;
    }
    void UpdateUpperLink (TrieLeaf * uplnk) {
        UpperLink = uplnk;
    }
    // The function reconstructs the term string
    std::string ThisTerm () {
	std::string s = "";
        for (TrieLeaf * t = this; t; t = t->UpperLink)
            s = t->KeyPart + s;
        return s;
    }
#ifdef _DEBUG
    void Print (std::string s = "", int level = 0);
#endif
};
#endif
