#ifndef  _NODE
#define  _NODE

#include <vector>
#include <string>
#include "trie.h"
#include "leaf.h"
#include "edistk.h"
#include "interface.h"

class TrieLeaf;
/*
 This is an implementation of the tree node. Each node
 could contain several leavs
*/

class TrieNode {
  protected:
    // The link to the first link of the node
    // all leavs are linked in a list
    TrieLeaf * FirstLeaf;
    /* 
       The following functions perform a search startings from the current node
       The function SearchAtLevelExact uses the results of prefix comparisons
       at upper levels. 
    */
    bool SearchAtLevelExact(const std::string & s, const std::string & st,
                             bool * pfound, int *numComp,
                             StringList * resList, bool silent) const;
    bool SearchAtLevel(const std::string & s, const std::string & st,
                        const checkDistContext & c, int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFileAtLevel(int fd, off_t offset,
                                    const std::string & s, const std::string & st,
                                    const checkDistContext & c,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent);
    static bool SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, const std::string & st,
                                    int *numComp, StringList * resList, bool silent);
  public:

     TrieNode();
    ~TrieNode();
    //  Function inserts a key into a given node
    TrieLeaf * Insert(std::string Key, TrieLeaf * UpperLeaf =
                      NULL, TrieNode * SubTree = NULL);
// The following to functions perform search from the root of the trie
    bool Search(const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFile(int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    bool Search(int fd, const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const;
    bool SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const;
// This function updates the UpperLink variable. It is useful when the node
// is splitted and migrates to another memory chunk. In this case the 
// child records (namely leavs) must be aware of changes
    void UpdateUpperLink(TrieLeaf * uplnk);
// Storing node content to the file
    int DumpToFile(int fd, off_t offset);
#ifdef _DEBUG
    void Print(std::string s = "", int level = 0);
#endif
};

#endif
