#include "triedfa.h"
#include <algorithm>
#include "../Interface/triedeb.h"

DEFINE_CLASS_STATIC_COUNTER(CTrieDFA, cr_dfa)

void CTrieDFA::SearchAtLevelExact(const std::string & s, int stLen,
                              int *numComp,
                             StringList * resList, bool silent) const
{
   if (stLen >= s.size()) {
       if (!silent)
           cout << s << '\n';
       if (resList != NULL)
           resList->push_back (s.substr (0, s.size () - 1));

       return;
   }
   int chSize = children.size();
   
   if (!chSize) 
       return;

   unsigned int minPos =  0, sz = chSize;

   const char *pBeg = s.c_str() + stLen;
   int cmp;
   unsigned int half; 
   unsigned int middle;

   while (sz > 0) {
       half = sz >> 1;
       middle = minPos + half;

       cmp = strcmp(children[middle].label.c_str(), pBeg);
       if (cmp < 0) {
           minPos = middle + 1;
           sz = sz - half - 1;
       } else {
           sz = half;
       }
   }

   if (minPos == chSize)  {
        --minPos;
   }
   
   if (1) {
        const Child *pChld = &children[minPos];
        int len = pChld->checkCommon(s, stLen);
        // We stand after if we have an exact prefix
        if (len < pChld->label.size() && minPos) {
            pChld = &children[minPos - 1];
            len = pChld->checkCommon(s, stLen);
        }
        CALC_NUM_COMP(numComp, pChld->ptr);
        if (len >= pChld->label.size()) {
            int stLen1 = stLen + pChld->label.size();
            if (stLen1 == s.size()) {
                if (!silent)
                    cout << s << '\n';
                if (resList != NULL)
                    resList->push_back (s.substr (0, s.size () - 1));
            } else {
                if (pChld->ptr) {
                    pChld->ptr->SearchAtLevelExact(s, stLen1, numComp, resList,silent);
                }
            }
        }
   }

}

void CTrieDFA::SearchAtLevel(const char *st, int stLen,
                LevenDFA &dfa,
                int *numComp, 
                StringList * resList, bool silent) const 
{
    LevenDFA::LevenNodeBase *pState = dfa.GetState();
    char  *sk = NULL;
    int   allocSize = 0; 

    int  sz = children.size();

    for (int i = 0; i < sz; ++i) {
        const string     &label  = children[i].label;
        int              childSz = label.size();
        int              newSz = stLen + childSz;

        const CTrieDFA *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree);

        bool final = false;
        bool ok = true;
        for (int k = 0; k < childSz; ++k) {
           if (!dfa.Next(label[k], final)) {
               ok = false;
               break;
           }
        }

        if (ok) {
            if (newSz > allocSize) {
                allocSize = newSz;
                sk = (char *)alloca(allocSize + 1);
                strcpy(sk, st);
            }
            strcpy(sk + stLen, label.c_str());
            if (SubTree) {
                SubTree->SearchAtLevel(sk, newSz,
                                        dfa,
                                        numComp, resList, silent);
            } else if (final) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL) {
                    resList->push_back (string(sk, newSz - 1));
                }
            }
        }
        dfa.SetState(pState);
    }
}

bool CTrieDFA::SearchInFile(const char *pABC, int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent)
{
    LevenDFA   dfa(pABC, s.c_str(), kerr, hasTransp);
    dfa.Start();

    string ss = s + LevenDFA::EOS;

    return (kerr) ? SearchInFileAtLevel(fd, 0, "", 
                    dfa,
                    numComp, resList, silent):
                    SearchInFileAtLevelExact (fd, 0, ss, "", numComp, resList, silent);
}

bool CTrieDFA::SearchInFileAtLevel(int fd, off_t offset,
                                    std::string st,
                                    LevenDFA &dfa,
                                    int *numComp, StringList * resList,
                                    bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    
    LevenDFA::LevenNodeBase *pState = dfa.GetState();

    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        string label(sbuf);
        string sk = st + label;
        CALC_NUM_COMP(numComp, sub_off)

        bool ok = true;
        bool final = false;
        for (int k = 0; k < label.size(); ++k) {
           if (!dfa.Next(label[k], final)) {
               ok = false;
               break;
           }
        }

        if (ok) {
            if (sub_off) {
                if (!SearchInFileAtLevel
                    (fd, sub_off, sk, 
                     dfa,
                     numComp, resList,  silent))
                    return false;
            } else if (final) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
        dfa.SetState(pState);
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

bool CTrieDFA::SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        if (numComp && !sub_off) {
            (*numComp)++;
        }
        if (sub_off) {
            if (sk.size() <= s.size() && !strncmp(s.c_str(), sk.c_str(), sk.size())) {
                if (!SearchInFileAtLevelExact(fd, sub_off, s, sk, numComp, resList, silent))
                    return false;
            }
        } else {
            if (sk == s) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

CTrieDFA::~CTrieDFA()
{
    for (int i = 0; i < children.size(); ++i) {
        if (children[i].ptr) delete children[i].ptr;
    }
}

void CTrieDFA::Search(const char *pABC, const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const
{
    if (!kerr) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    START_COUNTER(cr_dfa);
    LevenDFA   *pDfa = new LevenDFA(pABC, s.c_str(), kerr, hasTransp);
    pDfa->Start();
    STOP_COUNTER(cr_dfa);

    SearchAtLevel ("",  0,
                    *pDfa,
                    numComp, resList, silent);
    START_COUNTER(cr_dfa);
    delete pDfa;
    STOP_COUNTER(cr_dfa);
}

void CTrieDFA::SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const
{
    string ss = s + LevenDFA::EOS;

    SearchAtLevelExact(ss, 0, numComp, resList, silent);
}

void CTrieDFA::InsertInternal(const std::string &s)
{
    int indx = -1;
    int maxLen = 0;
    int len;
    for (int i = 0; i < children.size(); ++i) 
    if ((len = children[i].checkCommon(s, 0)) > maxLen) {
        indx = i;
        maxLen = len;
    }
    if (!maxLen) {
        Child chld(s);

        vector<Child>::iterator it = lower_bound(children.begin(), children.end(), chld);
        children.insert(it, chld);
        return;
    }
    string rest(s.begin() + maxLen, s.end());
    if (rest.empty()) {
        throw string("String '") + s + string("' length: ") + Format("%d", s.size()) + string(" is a duplicate");
    }
    Child &maxChld = children[indx];
    // If the string to be inserted has label as its prefix
    // insert it into the corresponding subtree
    if (maxLen == maxChld.label.size()) {
        if (maxChld.ptr) {
            maxChld.ptr->InsertInternal(rest);
        } else {
            throw string("String '") + s + string("' has a proper prefix in the dictionary!");
        }
        return;
    }
    //... otherwise the node must be splitted and subtree be created
    string restChld(maxChld.label.begin() + maxLen, maxChld.label.end());
    // Shorten the label
    maxChld.label.erase(maxChld.label.begin() + maxLen, maxChld.label.end());
    CTrieDFA *oldPtr = maxChld.ptr;
    CTrieDFA *pNewNode= new CTrieDFA;
    maxChld.ptr = pNewNode;
    pNewNode->children.push_back(Child(restChld, oldPtr));
    pNewNode->children.push_back(Child(rest));
    if (pNewNode->children[0] > pNewNode->children[1]) {
        swap(pNewNode->children[0], pNewNode->children[1]);
    }
}

void CTrieDFA::Print(int level)
{
    string indent(level, ' ');
    cout << indent << "Node address =" << this << " level=" << level << "\n";
    for (int i = 0; i < children.size(); ++i) {
        cout << indent << " Key " << children[i].label << " : " << children[i].ptr << endl;
        if (children[i].ptr) {
            children[i].ptr->Print(level + 1);
        }
    }
}

int CTrieDFA::DumpToFile(int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of leaves
// 2. the leaf keys themselvs, with zero offsets
// 3. leaf subtrees - updating offsets

    int buflen = 0;
    int sub_sz = 0;
    int cnt = children.size();
    buflen += sizeof (cnt) + sizeof (buflen);
    for (int i = 0; i < cnt; ++i) {
        buflen += sizeof (off_t) /* <- pointer size */ + 
                 + 1 /* string record size (must be<255) */  
                 + children[i].label.size() + 1;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    for (int i = 0; i < cnt; ++i) {
        // String length
        int len = children[i].label.size();
        *((u_int8_t *) (void *) sbuf) = len + 1;
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) ((void *) sbuf);
        *ptrs[i] = 0;
        // Leaf key
        sbuf += sizeof (off_t);
        memcpy (sbuf, children[i].label.c_str (), len + 1);
        sbuf += len + 1;
    }
    // Step 3.
    for (int i = 0; i < cnt; ++i) {
        // Diving into subtree
        int s_sz = (children[i].ptr)? children[i].ptr->DumpToFile (fd, offset + buflen + sub_sz) : 0;
        if (s_sz > 0)
            *ptrs[i] = offset + buflen + sub_sz;
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Returning the total number of bytes written
    return buflen + sub_sz;
}

void CTrieDFA::countBranch(vector<int> &cnt, vector<int> &cntBranch, int level)
{
    if (level + 1 > cnt.size()) {
        cnt.resize(level + 1);
        cntBranch.resize(level + 1);
    }
    cnt[level]++;
    int sz = children.size();
    cntBranch[level] += sz;
    for (int i = 0; i < sz; ++i) 
    if (children[i].ptr) {
        children[i].ptr->countBranch(cnt, cntBranch, level + 1);
    }
}

void CTrieDFA::PrintAddStat(int testNum, int kerr, ostream & str) 
{
    if (testNum >= 0 && kerr) {
        double tm = GET_COUNTER_MS(cr_dfa);
        tm /= testNum;
        str << "Kerr = " << kerr << " Average DFA cr time: " << tm ;
    }
}

void CTrieDFA::ResetCounters() 
{
    RESET_COUNTER(cr_dfa);
}

