#ifndef  TRIEDFA_H
#define  TRIEDFA_H

#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/timeb.h>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include "levendfa.h"

#include <vector>
#include <string>
#include "interface.h"

class CTrieDFA {
    DECLARE_STATIC_COUNTER(cr_dfa)
    class Child {
        CTrieDFA  *ptr;
        string label;
        friend class CTrieDFA;
    public:

        Child(const string & label = "", CTrieDFA *ptr = NULL) : label(label), ptr(ptr) {}
        ~Child(){} // do not delete here it will be done on the upper level
        CTrieDFA *getPtr(){return ptr;}
        bool operator<(const Child &o) const {if (label != o.label) return label < o.label;return ptr < o.ptr;}
        bool operator>(const Child &o) const {if (label != o.label) return label > o.label;return ptr > o.ptr;}
        inline int checkCommon(const string &s, int start) const;
    };
    vector<Child>  children;
  protected:
    /* 
       The following functions perform a search startings from the current node
       The function SearchAtLevelExact uses the results of prefix comparisons
       at upper levels. 
    */
    void SearchAtLevelExact(const std::string & s, int stLen, int *numComp,
                             StringList * resList, bool silent) const;
    void SearchAtLevel(const char * st, int stLen,
                        LevenDFA &dfa,
                        int *numComp,
                        StringList * resList, bool silent) const;
    bool SearchInFileAtLevel(int fd, off_t offset,
                        std::string st,
                        LevenDFA &dfa,
                        int *numComp, StringList * resList,
                        bool silent);
    bool SearchInFileAtLevelExact(int fd, off_t offset,
                        const std::string & s, std::string st,
                        int *numComp, StringList * resList, bool silent);
    void InsertInternal(const string &str);
  public:

    ~CTrieDFA();

    void Search(const char *pABC, const std::string & s, int kerr, int *numComp,
                 StringList * resList, 
                 bool hasTransp, bool silent) const;
    bool SearchInFile(const char *pABC, int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    void SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const;
    void Insert(const char * str) {
        string s = str;
        s += LevenDFA::EOS;
        InsertInternal(s);
    }
    void PrintAddStat(int testNum, int kerr, ostream & str);
    void ResetCounters();

// Storing node content to the file
    int DumpToFile(int fd, off_t offset);
    void Print(int level = 0);
    void countBranch(vector<int> &cnt, vector<int> &cntBranch, int level = 0);
};

int CTrieDFA::Child::checkCommon(const std::string &Key, int start) const
{
    int i = 0, k = start;

    int kSize = Key.size() - start;
    int iMax = min(int(label.size()), kSize);
    for (; i < iMax;)
        if (Key[k] == label[i]) {
            i++; k++;
        }
        else
            break;
    return i;
}

#endif
