#include <algorithm>

#include "triedfa.h"
#include "../Interface/triedeb.h"

#define HINT_PART_LEN 32 // this is a hint 

static int numNodes;

DEFINE_CLASS_STATIC_COUNTER(CTrieDFA, cr_dfa_static)

void CTrieDFA::SearchAtLevelExact(const std::string & s, int stLen,
                              int *numComp,
                             StringList * resList, bool silent) const
{
   numNodes++;
   if (stLen >= s.size()) {
       if (!silent)
           cout << s << '\n';
       if (resList != NULL)
           resList->push_back (s.substr (0, s.size () - 1));

       return;
   }
   int chSize = children.size();

   
   if (!chSize) 
       return;

   unsigned int minPos =  0, sz = chSize;

   const char *pBeg = s.c_str() + stLen;
   int cmp;
   unsigned int half; 
   unsigned int middle;

   while (sz > 0) {
       half = sz >> 1;
       middle = minPos + half;

       cmp = strcmp(children[middle].label.c_str(), pBeg);
       if (cmp < 0) {
           minPos = middle + 1;
           sz = sz - half - 1;
       } else {
           sz = half;
       }
   }

   if (minPos == chSize)  {
        --minPos;
   }
   
   if (1) {
        const Child *pChld = &children[minPos];
        int len = pChld->checkCommon(s, stLen);
        // We stand after if we have an exact prefix
        CALC_NUM_COMP(numComp, pChld->ptr);
        if (len < pChld->label.size() && minPos) {
            pChld = &children[minPos - 1];
            len = pChld->checkCommon(s, stLen);
        }
        if (len >= pChld->label.size()) {
            int stLen1 = stLen + pChld->label.size();
            if (stLen1 == s.size()) {
                if (!silent)
                    cout << s << '\n';
                if (resList != NULL)
                    resList->push_back (s.substr (0, s.size () - 1));
            } else {
                if (pChld->ptr) {
                    pChld->ptr->SearchAtLevelExact(s, stLen1, numComp, resList,silent);
                }
            }
        }
   }
}

void CTrieDFA::SearchAtLevelPartExact(
                char *st, int stLen,
                const string &str,
                int switchLen,
                LevenDFA &dfa2,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    LevenDFA::LevenNodeBase *pState2 = dfa2.GetState();
    numNodes++;
    
    int sz = children.size();
    const char *pStrData = (const char *)str.c_str() + stLen;

    for (int i = 0; i < sz; ++i) {
        const string     &label = children[i].label;
        const CTrieDFA *SubTree = children[i].ptr;
        const char     *pLabelData = label.c_str();
        int              childSz = label.size();
        int              newSz = stLen + childSz;

        PrintStr(stLen, pLabelData);

        CALC_NUM_COMP(numComp, SubTree);
            
        if (newSz < switchLen) {
            if (SubTree && !strncmp(pLabelData, pStrData, childSz)) {

                strncpy(st + stLen, pLabelData, childSz + 1);

                SubTree->SearchAtLevelPartExact(st, newSz,
                                        str,
                                        switchLen,
                                        dfa2,
                                        numComp, resList, hasTransp, silent);
            }
        } else {
            int dist;
            if (!strncmp(pLabelData, pStrData, switchLen - stLen)) {
                int newLen = newSz - switchLen; // The length of the pattern, which is going to be matched using a new context

                strncpy(st + stLen, pLabelData, childSz + 1);


                    bool okDFA = true; 
                    bool final = false;

                    for (int k = 0, pos = switchLen - stLen; 
                         k < newLen; 
                         ++k, ++pos) {
                       if (!dfa2.Next(pLabelData[pos], final)) {
                           okDFA = false; 
                           break;
                       }
                    }

                    if (okDFA) {
                        if (SubTree) {
                            SubTree->SearchAtLevel(
                                        st, newSz,
                                        dfa2,
                                        numComp, resList, hasTransp, silent);
                        } else if (final) {
                            if (!silent)
                            cout << st << '\n';
                            if (resList != NULL)
                            resList->push_back (string(st, newSz - 1));
                        }
                    }
            }
        }
        dfa2.SetState(pState2);
    }
}

void CTrieDFA::SearchAtLevel(
                char *st, int stLen,
                LevenDFA &dfa1,
                int switchLen,
                LevenDFA &dfa2,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    LevenDFA::LevenNodeBase  *pState1 = dfa1.GetState();
    LevenDFA::LevenNodeBase  *pState2 = dfa2.GetState();
    numNodes++;

    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
        const string     &label = children[i].label;
        const char       *pLabelData = label.c_str();
        int              childSz = label.size();
        int              newSz = stLen + childSz;

        PrintStr(stLen, pLabelData);

        const CTrieDFA *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree);

        int startPos = stLen;

        // If the difference between two word lengths
        // is k, then the edit distance between them is
        // also at least k
        int gap = dfa1.getKerr();

        for (int labelPos = 0; startPos < newSz; ++labelPos, ++startPos) {

            int startPos1 = startPos + 1;
            // startPos1 is the length of the prefix, not startPos!!!
            if (startPos1 > switchLen + gap) {
                break; // no matches follow
            }

            bool finalState = false;
            if (!dfa1.Next(pLabelData[labelPos], finalState)) {
                break;
            }
            // startPos1 is the length of the prefix, not startPos!!!
            if (startPos1 < switchLen - gap) {
                continue;
            }

            LevenDFA::LevenNodeBase *pBeforeState = dfa1.GetState();
            dfa1.Next(LevenDFA::EOS, finalState);
            dfa1.SetState(pBeforeState);
            //bool final = (startPos1 == newSz && !SubTree);

            if (finalState) {
                strncpy(st + stLen, pLabelData, childSz + 1);

                bool okDFA2 = true;
                bool dfa2final = false;
                for (int k = startPos1; k < newSz; ++k) {
                    if (!dfa2.Next(st[k], dfa2final)) {
                        okDFA2 = false;
                        break;
                    }
                }

                if (startPos1 == newSz || okDFA2) {
                    if (SubTree) {
                        SubTree->SearchAtLevel(
                                    st, newSz,
                                    dfa2,
                                    numComp, resList, hasTransp, silent);
                    } else if (dfa2final) { // it is final state already
                        if (!silent)
                            cout << st << '\n';
                            if (resList != NULL) {
                                resList->push_back (string(st, newSz - 1));
                        }
                    }
                }
                dfa2.SetState(pState2);
            } 
        }
        if (startPos == newSz && SubTree) { // it might be also
            // Additional case to check
            strncpy(st + stLen, pLabelData, childSz + 1);
            // ok to continue with matching for the first part in the subtree
            SubTree->SearchAtLevel(
                st, newSz,
                dfa1,
                switchLen,
                dfa2,
                numComp, 
                resList, hasTransp, silent);
        }
        dfa1.SetState(pState1);
        dfa2.SetState(pState2);
    }
}

void CTrieDFA::SearchAtLevel(
                char *st, int stLen,
                LevenDFA & dfa,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    LevenDFA::LevenNodeBase  *pState = dfa.GetState();
    numNodes++;

    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
        int dist;
        const string     &label = children[i].label;
        const char       *pLabelData = label.c_str();
        int              childSz = label.size();
        int              newSz = stLen + childSz;

        PrintStr(stLen, pLabelData);

        const CTrieDFA *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree);

        bool okDFA = true;
        bool final = false;

        for (int k = 0; k < childSz; ++k)
        if (!dfa.Next(pLabelData[k], final)) {
            okDFA = false;
            break;
        }
          
        if (okDFA) {
            strncpy(st + stLen, pLabelData, childSz + 1);

            if (SubTree) {
                SubTree->SearchAtLevel(st, newSz,
                                        dfa,
                                        numComp, resList, hasTransp, silent);
            } else if (final) {
                if (!silent)
                    cout << st << '\n';
                if (resList != NULL) {
                    resList->push_back (string(st, newSz - 1));
                }
            }
        }
        dfa.SetState(pState);
    }
}

bool CTrieDFA::SearchInFile(const char *pABC, int fd, 
                         const std::string & s, int kerr,
                         int *numComp, StringList * resList,
                         bool hasTransp, bool silent)
{
    string    ss = s + LevenDFA::EOS;
    START_COUNTER(cr_dfa_static);
    LevenDFA  dfa(pABC, s, kerr, hasTransp);
    dfa.Start();
    STOP_COUNTER(cr_dfa_static);


    return (kerr) ? SearchInFileAtLevel(fd, 0, 
                        0,  // startCmpLen
                        "", 0,
                        dfa,
                        numComp, resList, hasTransp, silent):
                    SearchInFileAtLevelExact (fd, 0, ss, "", numComp, resList, silent);
}

bool CTrieDFA::SearchInFileSplit(const char *pABC,
                 int fd,
                 const std::string & s, 
                 int kerr1, bool exDist1,
                 int kerr2, bool exDist2,
                 int *numComp,
                 int switchLen,
                 StringList * resList, 
                 bool hasTransp, bool silent) 
{
    if (!kerr1 && !kerr2) {
        return SearchInFileAtLevelExact (fd, 0, s.c_str(), "", numComp, resList, silent);
    }
    if (switchLen <= 0) {
        return SearchInFile(pABC, fd, s, kerr1 + kerr2, numComp, resList, hasTransp, silent);
    }
    string ss1 = s.substr(0, switchLen);
    string ss2 = s.substr(switchLen);

    START_COUNTER(cr_dfa_static);
    LevenDFA  dfa1(pABC, ss1, kerr1, hasTransp, exDist1);
    LevenDFA  dfa2(pABC, ss2, kerr2, hasTransp, exDist2);
    dfa1.Start();
    dfa2.Start();
    STOP_COUNTER(cr_dfa_static);

    if (kerr1) {
        return SearchInFileAtLevel (fd, 0,
                    "",  0,
                    dfa1,
                    switchLen,
                    dfa2,
                    numComp, resList, hasTransp, silent);
    } else {
        return SearchInFileAtLevelPartExact (fd, 0,
                    "",  0,
                    s,
                    switchLen,
                    dfa2,
                    numComp, resList, hasTransp, silent);
    
    }
    return true;
}

bool CTrieDFA::SearchInFileAtLevelPartExact(
                int fd, off_t offset,
                const char *st, int stLen,
                const string &str,
                int switchLen,
                LevenDFA &dfa,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) 
{
    int allocSize = stLen + HINT_PART_LEN;
    char *sk = (char *)alloca(allocSize + 1);
    strcpy(sk, st);

    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int sz = *((int *) (void *) sbuf);
    sbuf += sizeof (sz);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    
    LevenDFA::LevenNodeBase *pState = dfa.GetState();
    
    for (int i = 0; i < sz; ++i) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string label(sbuf);

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        if (newSz > allocSize) {
            allocSize = newSz;
            sk = (char *)alloca(allocSize + 1);
            strcpy(sk, st);
        }

        strcpy(sk + stLen, sbuf);

        CALC_NUM_COMP(numComp, sub_off)
        if (newSz < switchLen) {
            if (sub_off && !strncmp(sk + stLen, str.c_str() + stLen, label.size())) {
                SearchInFileAtLevelPartExact(
                                        fd, sub_off,
                                        sk, newSz,
                                        str,
                                        switchLen,
                                        dfa,
                                        numComp, resList, hasTransp, silent);
            }
        } else {
            int dist;
            if (!strncmp(sk + stLen, str.c_str() + stLen, switchLen - stLen)) {
                int newLen = newSz - switchLen; // The length of the pattern, which is going to be matched using a new context

                    bool okDFA = true; 
                    bool final = false;

                    for (int k = 0, pos = switchLen - stLen; 
                         k < newLen; 
                         ++k, ++pos) {
                       if (!dfa.Next(label[pos], final)) {
                           okDFA = false; 
                           break;
                       }
                    }

                    if (okDFA) {
                        if (sub_off) {
                            SearchInFileAtLevel(
                                        fd, sub_off,
                                        switchLen,
                                        sk, newSz,
                                        dfa,
                                        numComp, resList, hasTransp, silent);
                        } else if (final) {
                            if (!silent)
                            cout << sk << '\n';
                            if (resList != NULL) {
                                resList->push_back (string(sk, newSz - 1));
                            }
                        }
                    }
            }
        }

        sbuf += slen;
        dfa.SetState(pState);
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    
    return true;
}

bool CTrieDFA::SearchInFileAtLevel(
                int fd, off_t offset,
                const char *st, int stLen,
                LevenDFA &dfa1,
                int switchLen,
                LevenDFA &dfa2,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) 
{
    int allocSize = stLen + HINT_PART_LEN;
    char *sk = (char *)alloca(allocSize + 1);
    strcpy(sk, st);

    LevenDFA::LevenNodeBase *pState1 = dfa1.GetState();
    LevenDFA::LevenNodeBase *pState2 = dfa2.GetState();

    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int sz = *((int *) (void *) sbuf);
    sbuf += sizeof (sz);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    
    for (int i = 0; i < sz; ++i) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string label(sbuf);

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        if (newSz > allocSize) {
            allocSize = newSz;
            sk = (char *)alloca(allocSize + 1);
            strcpy(sk, st);
        }

        strcpy(sk + stLen, sbuf);

        CALC_NUM_COMP(numComp, sub_off)

        int startPos = stLen;

        for (;startPos < newSz; ) {
            int startPos1 = startPos + 1;
            if (startPos1 > switchLen + dfa1.getKerr()) { 
                break; // no matches follow
            }

            bool finalState = false;
            if (!dfa1.Next(sk[startPos], finalState)) {
                break;
            }

            LevenDFA::LevenNodeBase *pBeforeState = dfa1.GetState();
            dfa1.Next(LevenDFA::EOS, finalState);
            dfa1.SetState(pBeforeState);
            //bool final = (startPos1 == newSz && !SubTree);

            if (finalState) {
                bool okDFA2 = true;
                bool dfa2final = false;
                for (int k = startPos1; k < newSz; ++k) {
                    if (!dfa2.Next(sk[k], dfa2final)) {
                        okDFA2 = false;
                        break;
                    }
                }

                if (startPos1 == newSz || okDFA2) {

                    if (sub_off) {
                        if (!SearchInFileAtLevel(
                                        fd, sub_off,
                                        startPos1,
                                        sk, newSz,
                                        dfa2,
                                        numComp, resList, 
                                        hasTransp, silent)) {
                            return false;
                        }
       
                    } else if (dfa2final) {
                        if (!silent) {
                            cout << sk << endl;
                        }
                        if (resList) {
                            resList->push_back(string(sk, newSz - 1));
                        }
                    }
                }
                dfa2.SetState(pState2);
            }
            startPos = startPos1;
        } /* for (;startPos < newSz;) */
        // Additional case to check
        if (startPos == newSz && sub_off) { // it might be also
            // ok to continue matching the first part in the subtree
                if (!SearchInFileAtLevel(
                                        fd, sub_off,
                                        sk, newSz,
                                        dfa1,
                                        switchLen,
                                        dfa2,
                                        numComp, resList, hasTransp, silent)) {
                   return false;
                   // Restore DFA states after searching
                }
        }

        sbuf += slen;
        dfa1.SetState(pState1);
        dfa2.SetState(pState2);
    }
    if (munmap (buf, buflen + delta) < 0) {
        return false;
    }
    return true;
}


bool CTrieDFA::SearchInFileAtLevel(int fd, off_t offset,
                                    int startCmpLen,
                                    const char * st, int stLen,
                                    LevenDFA &dfa,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent)
{
    int allocSize = stLen + HINT_PART_LEN;
    char *sk = (char *)alloca(allocSize + 1);
    strcpy(sk, st);

    LevenDFA::LevenNodeBase *pState = dfa.GetState();

    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        if (newSz > allocSize) {
            allocSize = newSz;
            sk = (char *)alloca(allocSize + 1);
            strcpy(sk, st);
        }
        strcpy(sk + stLen, sbuf);
        
        CALC_NUM_COMP(numComp, sub_off)

        bool okDFA = true;
        bool final = false;
        for (int k = startCmpLen; k < newSz; ++k)
        if (!dfa.Next(sk[k], final)) {
            okDFA = false;
            break;
        }
         
        if (okDFA) {
            if (sub_off) {
                if (!SearchInFileAtLevel
                    (fd, sub_off, 
                     startCmpLen,
                     sk, newSz,
                     dfa,
                     numComp, resList, hasTransp, silent)) {
                    return false;
                }
            } else if (final) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL) {
                    resList->push_back (string(sk, newSz - 1));
                }
            }
        }
        sbuf += slen;
        dfa.SetState(pState);
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

bool CTrieDFA::SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        CALC_NUM_COMP(numComp, sub_off)
        
        if (sub_off) {
            if (sk.size() <= s.size() && !strncmp(s.c_str(), sk.c_str(), sk.size())) {
                if (!SearchInFileAtLevelExact(fd, sub_off, s, sk, numComp, resList, silent))
                    return false;
            }
        } else {
            if (sk == s) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

CTrieDFA::CTrieDFA()
{
}

CTrieDFA::~CTrieDFA()
{
    for (int i = 0; i < children.size(); ++i) {
        if (children[i].ptr) delete children[i].ptr;
    }
}

static    char                stComm[8192];

void CTrieDFA::Search(const char *pABC, const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const
{
    PrintPat(s.c_str(), kerr);
    START_COUNTER(stime);
    if (!kerr) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    STOP_COUNTER(stime);
    START_COUNTER(cr_dfa);
    LevenDFA  *pDfa1 = new LevenDFA(pABC, s, kerr, hasTransp);
    pDfa1->Start();
    STOP_COUNTER(cr_dfa);

    stComm[0] = 0;

    START_COUNTER(stime);
    SearchAtLevel (stComm,  0,
                    *pDfa1,
                    numComp, resList, hasTransp, silent);
    STOP_COUNTER(stime);
    START_COUNTER(cr_dfa);
    delete pDfa1;
    STOP_COUNTER(cr_dfa);
}

void CTrieDFA::SearchSplit(const char *pABC,
                 const std::string & s, 
                 int kerr1, bool exDist1,
                 int kerr2, bool exDist2,
                 int *numComp,
                 int switchLen,
                 StringList * resList, bool hasTransp, bool silent) const
{
    PrintPat(s.c_str(), kerr1, kerr2);
    START_COUNTER(stime);
    if (!kerr1 && !kerr2) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    if (switchLen <= 0) {
        Search(pABC, s, kerr1 + kerr2, numComp, resList, hasTransp, silent);
        return;
    }
    string ss1 = s.substr(0, switchLen);
    string ss2 = s.substr(switchLen);
    STOP_COUNTER(stime);

    START_COUNTER(cr_dfa);
    LevenDFA  *pDfa1 = new LevenDFA(pABC, ss1, kerr1, hasTransp, exDist1);
    LevenDFA  *pDfa2 = new LevenDFA(pABC, ss2, kerr2, hasTransp, exDist2);
    pDfa1->Start();
    pDfa2->Start();
    STOP_COUNTER(cr_dfa);

    stComm[0] = 0;

    START_COUNTER(stime);
    if (kerr1) {
        SearchAtLevel (stComm,  0,
                    *pDfa1,
                    switchLen,
                    *pDfa2,
                    numComp, resList, hasTransp, silent);
    } else {
        SearchAtLevelPartExact (stComm,  0,
                    s,
                    switchLen,
                    *pDfa2,
                    numComp, resList, hasTransp, silent);
    
    }
    STOP_COUNTER(stime);
    START_COUNTER(cr_dfa);
    delete pDfa1;
    delete pDfa2;
    STOP_COUNTER(cr_dfa);
}

void CTrieDFA::SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const
{
    string ss = s + LevenDFA::EOS;

    START_COUNTER(stime);
    SearchAtLevelExact(ss, 0, numComp, resList, silent);
    STOP_COUNTER(stime);
}

void CTrieDFA::InsertInternal(const std::string &s)
{
    int indx = -1;
    int maxLen = 0;
    int len;
    for (int i = 0; i < children.size(); ++i) 
    if ((len = children[i].checkCommon(s, 0)) > maxLen) {
        indx = i;
        maxLen = len;
    }
    if (!maxLen) {
        Child chld(s);

        vector<Child>::iterator it = lower_bound(children.begin(), children.end(), chld);
        children.insert(it, chld);
        return;
    }
    string rest(s.begin() + maxLen, s.end());
    if (rest.empty()) {
        throw string("String '") + s + string("' length: ") + Format("%d", s.size()) + string(" is a duplicate");
    }
    Child &maxChld = children[indx];
    // If the string to be inserted has label as its prefix
    // insert it into the corresponding subtree
    if (maxLen == maxChld.label.size()) {
        if (maxChld.ptr) {
            maxChld.ptr->InsertInternal(rest);
        } else {
            throw string("String '") + s + string("' has a proper prefix in the dictionary!");
        }
        return;
    }
    //... otherwise the node must be splitted and subtree be created
    string restChld(maxChld.label.begin() + maxLen, maxChld.label.end());
    // Shorten the label
    maxChld.label.erase(maxChld.label.begin() + maxLen, maxChld.label.end());
    CTrieDFA *oldPtr = maxChld.ptr;
    CTrieDFA *pNewNode= new CTrieDFA;
    maxChld.ptr = pNewNode;
    pNewNode->children.push_back(Child(restChld, oldPtr));
    pNewNode->children.push_back(Child(rest));
    if (pNewNode->children[0] > pNewNode->children[1]) {
        swap(pNewNode->children[0], pNewNode->children[1]);
    }
}

void CTrieDFA::Print(int level)
{
    string indent(level, ' ');
    cout << indent << "Node address =" << this << " level=" << level << "\n";
    for (int i = 0; i < children.size(); ++i) {
        cout << indent << " Key " << children[i].label << " : " << children[i].ptr << endl;
        if (children[i].ptr) {
            children[i].ptr->Print(level + 1);
        }
    }
}

int CTrieDFA::DumpToFile(int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of leaves
// 2. the leaf keys themselvs, with zero offsets
// 3. leaf subtrees - updating offsets

    int buflen = 0;
    int sub_sz = 0;
    int cnt = children.size();
    buflen += sizeof (cnt) + sizeof (buflen);
    for (int i = 0; i < cnt; ++i) {
        buflen += sizeof (off_t) /* <- pointer size */ + 
                 + 1 /* string record size (must be<255) */  
                 + children[i].label.size() + 1;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    for (int i = 0; i < cnt; ++i) {
        // String length
        int len = children[i].label.size();
        *((u_int8_t *) (void *) sbuf) = len + 1;
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) ((void *) sbuf);
        *ptrs[i] = 0;
        // Leaf key
        sbuf += sizeof (off_t);
        memcpy (sbuf, children[i].label.c_str (), len + 1);
        sbuf += len + 1;
    }
    // Step 3.
    for (int i = 0; i < cnt; ++i) {
        // Diving into subtree
        int s_sz = (children[i].ptr)? children[i].ptr->DumpToFile (fd, offset + buflen + sub_sz) : 0;
        if (s_sz > 0)
            *ptrs[i] = offset + buflen + sub_sz;
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Returning the total number of bytes written
    return buflen + sub_sz;
}

void CTrieDFA::PrintAddStat(int testNum, int kerr, ostream & str) 
{
    if (testNum >= 0 && kerr) {
        double tm_cr = GET_COUNTER_MS(cr_dfa);
        double tm1 = GET_COUNTER_MS(stime);
        tm_cr /= testNum;
        tm1 /= testNum;
        str << "Kerr = " << kerr << " Average DFA cr/del time: " << tm_cr << " Num of nodes visited: " << numNodes /testNum << " average search time: " << tm1 << endl;
    }
}

void CTrieDFA::ResetCounters() 
{
    numNodes = 0;
    RESET_COUNTER(cr_dfa);
    RESET_COUNTER(stime);
}
