#ifndef  TRIEDFADOUBLE_H
#define  TRIEDFADOUBLE_H

#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/timeb.h>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>

#include "../LevenDFA/levendfa.h"

#include <vector>
#include <string>

class CTrieDFA {
    DECLARE_STATIC_COUNTER(cr_dfa_static)
    DECLARE_COUNTER(cr_dfa)
    DECLARE_COUNTER(stime)
    class Child {
        CTrieDFA  *ptr;
        string label;
        friend class CTrieDFA;
    public:

        Child(const string & label = "", CTrieDFA *ptr = NULL) : label(label), ptr(ptr) {}
        ~Child(){} // do not delete here it will be done on the upper level
        CTrieDFA *getPtr(){return ptr;}
        bool operator<(const Child &o) const {if (label != o.label) return label < o.label;return ptr < o.ptr;}
        bool operator>(const Child &o) const {if (label != o.label) return label > o.label;return ptr > o.ptr;}
        inline int checkCommon(const string &s, int start) const;
    };
    vector<Child>  children;
  protected:
    /* 
       The following functions perform a search startings from the current node
       The function SearchAtLevelExact uses the results of prefix comparisons
       at upper levels. 
    */
    void SearchAtLevelExact(const std::string & s, int stLen, int *numComp,
                             StringList * resList, bool silent) const;
    void SearchAtLevel(char * st, int stLen,
                        LevenDFA & dfa,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    void SearchAtLevel(char * st, int stLen,
                        LevenDFA &dfa1,
                        int switchLen,
                        LevenDFA &dfa2,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    void SearchAtLevelPartExact(
                        char * st, int stLen,
                        const string & str,
                        int switchLen,
                        LevenDFA &dfa,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;

    static bool SearchInFileAtLevel(int fd, off_t offset,
                                    int startCmpLen,
                                    const char * st, int stLen,
                                    LevenDFA & dfa,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent);
    static bool SearchInFileAtLevel(int fd, off_t offset,
                        const char * st, int stLen,
                        LevenDFA &dfa1,
                        int switchLen,
                        LevenDFA &dfa2,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent);
    static bool SearchInFileAtLevelPartExact(
                        int fd, off_t offset,
                        const char * st, int stLen,
                        const string & str,
                        int switchLen,
                        LevenDFA &dfa,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent);
    static bool SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent);
    void InsertInternal(const string &s);
  public:

     CTrieDFA();
    ~CTrieDFA();

    void Search(const char *pABC, const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const;
    void SearchSplit(const char *pABC, const std::string & s, 
                     int kerr1, bool exDist1, 
                     int kerr2, bool exDist2,
                     int *numComp,
                     int switchLen,
                     StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFile(const char *pABC, int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    static bool SearchInFileExact(int fd, const std::string & s,
                             int *numComp, StringList * resList,
                             bool silent) {
                        return SearchInFile("", fd, s, 0, numComp, resList, false, silent);
                     }
    static bool SearchInFileSplit(const char *pABC, int fd,
                     const std::string & s, 
                     int kerr1, bool exDist1,
                     int kerr2, bool exDist2,
                     int *numComp,
                     int switchLen,
                     StringList * resList, 
                     bool hasTransp, bool silent);
    void SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const;
    void Insert(string s) {
        s += LevenDFA::EOS;
        InsertInternal(s);
    }

// Storing node content to the file
    int DumpToFile(int fd, off_t offset);
    void Print(int level = 0);
    void PrintAddStat(int testNum, int kerr, ostream & str);
    void ResetCounters();
};

int CTrieDFA::Child::checkCommon(const std::string &Key, int start) const
{
    int i = 0, k = start;

    int kSize = Key.size() - start;
    int iMax = min(int(label.size()), kSize);
    for (; i < iMax;)
        if (Key[k] == label[i]) {
            i++; k++;
        }
        else
            break;
    return i;
}

#endif
