#ifndef  TRIE_H
#define  TRIE_H

#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/timeb.h>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>

const std::string EOS = "\1";

#include <vector>
#include <string>
#include "trie.h"
#include "afuzzycontext.h"
#include "interface.h"

class CTrie {
    class Child {
        CTrie  *ptr;
        string label;
        friend class CTrie;
    public:

        Child(const string & label = "", CTrie *ptr = NULL) : label(label), ptr(ptr) {}
        ~Child(){} // do not delete here it will be done on the upper level
        CTrie *getPtr(){return ptr;}
        bool operator<(const Child &o) const {if (label != o.label) return label < o.label;return ptr < o.ptr;}
        bool operator>(const Child &o) const {if (label != o.label) return label > o.label;return ptr > o.ptr;}
        inline int checkCommon(const string &s, int start) const;
    };
    vector<Child>  children;
  protected:
    /* 
       The following functions perform a search startings from the current node
       The function SearchAtLevelExact uses the results of prefix comparisons
       at upper levels. 
    */
    void SearchAtLevelExact(const std::string & s, int stLen, int *numComp,
                             StringList * resList, bool silent) const;
    void SearchAtLevel(int startCmpLen,
                        char * st, int stLen,
                        const afuzzycontext & a, 
                        const context & tmpCon,
                        bool exDist,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    void SearchAtLevel(char * st, int stLen,
                        const afuzzycontext & a1, 
                        const context & tmpCon1, 
                        bool exDist1,
                        int switchLen,
                        const afuzzycontext & a2, const context & tmpCon2,
                        bool exDist2, int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    void SearchAtLevelPartExact(
                        char * st, int stLen,
                        const string & str,
                        int switchLen,
                        const afuzzycontext & a2, 
                        const context & tmpCon2, 
                        bool exDist2,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;

    static bool SearchInFileAtLevel(int fd, off_t offset,
                                    int startCmpLen,
                                    char * st, int stLen,
                                    const afuzzycontext & a, 
                                    const context & tmpCon,
                                    bool exDist,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent);
    static bool SearchInFileAtLevel(int fd, off_t offset,
                        char * st, int stLen,
                        const afuzzycontext & a1, 
                        const context & tmpCon1, 
                        bool exDist1,
                        int switchLen,
                        const afuzzycontext & a2, const context & tmpCon2,
                        bool exDist2, int *numComp,
                        StringList * resList, bool hasTransp, bool silent);
    static bool SearchInFileAtLevelPartExact(
                        int fd, off_t offset,
                        char * st, int stLen,
                        const string & str,
                        int switchLen,
                        const afuzzycontext & a2, 
                        const context & tmpCon2, 
                        bool exDist2,
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent);
    static bool SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent);
  public:

     CTrie();
    ~CTrie();

    void Search(const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const;
    void SearchSplit(const std::string & s, int kerr1, bool exDist1, 
                     int kerr2, bool exDist2, 
                     int *numComp,
                     int switchLen,
                     StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFile(int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    static bool SearchInFileExact(int fd, const std::string & s,
                             int *numComp, StringList * resList,
                             bool silent) {
                        return SearchInFile(fd, s, 0, numComp, resList, false, silent);
                     }
    static bool SearchInFileSplit(int fd,
                     const std::string & s, int kerr1, bool exDist1, 
                     int kerr2, bool exDist2, 
                     int *numComp,
                     int switchLen,
                     StringList * resList, 
                     bool hasTransp, bool silent);
    void SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const;
    void Insert(const std::string &s);

// Storing node content to the file
    int DumpToFile(int fd, off_t offset);
    void Print(int level = 0);
};

int CTrie::Child::checkCommon(const std::string &Key, int start) const
{
    int i = 0, k = start;

    int kSize = Key.size() - start;
    int iMax = min(int(label.size()), kSize);
    for (; i < iMax;)
        if (Key[k] == label[i]) {
            i++; k++;
        }
        else
            break;
    return i;
}

#endif
