#include "index.h"
#include <numeric>

CTrieIndex::CTrieIndex (bool hasTransp, EAlphabet AlphaType)
            : CInterface(hasTransp, AlphaType), opened (false)
{
}

CTrieIndex::~CTrieIndex ()
{
    if (opened)
        CloseDB ();
}

bool CTrieIndex::Insert (const char *str)
{
    try {
        node.Insert (string (str) + EOS);
    } catch (const string & e) {
        cerr << "Error inserting of '" << (string(str) +EOS) << "': " << e << endl;
        throw;
    }
    return true;
}

bool CTrieIndex::Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string s (str);

    if (kerr != 0) {
        node.Search (s, kerr, numComp, resList, mHasTransp, silent);
    } else {
        node.SearchExact (s, numComp, resList, silent);
    }
    return true;
}

bool CTrieIndex::SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, 
                             bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return node.SearchInFile (fd, string (str), kerr, numComp, resList,
                              mHasTransp, silent);
}

bool CTrieIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CTrieIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CTrieIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    //node.Print();
    stat = node.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

void CTrieIndex::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
     if (!kerr && type == SearchTypeMem) {
          vector<int>   cnt;
          vector<int>   cntBranch; 
          vector<int>   cntLeaves; 

          node.countBranch(cnt, cntBranch, cntLeaves);
          int sz = cnt.size();
          ostr << "Average number of children per level:\n";
          for (int i = 0; i < sz; ++i) {
              ostr << i << "\t";
          }
          ostr << endl;
          for (int i = 0; i < sz; ++i) {
              if (cnt[i]) {
                  ostr << double(cntBranch[i])/cnt[i] << "\t";
              } else {
                  ostr << "N/A\t" ;
              }
          }
          ostr << endl;


          ostr << "Average number of leaves per level:\n";
          for (int i = 0; i < sz; ++i) {
              ostr << i << "\t";
          }
          ostr << endl;

          for (int i = 0; i < sz; ++i) {
              if (cnt[i]) {
                  ostr << double(cntLeaves[i])/cnt[i] << "\t";
              } else {
                  ostr << "N/A\t" ;
              }
          }
     }
}

#ifdef _DEBUG
void CTrieIndex::Print ()
{
    node.Print ();
}
#endif
