#include "index.h"
#include "mcheck.h"
#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <interface.h>

using namespace std;

typedef vector<string> VS;

void usage(const char *prog)
{
    cerr << "Usage: " << prog << " transp? dict test \n";
}

int main (int argc, char *argv[])
{
    if (argc < 2) {
        usage(argv[0]);
        return -1;
    }
    bool                hasTransp = atol(argv[1]);
    string              dict = argv[2];
    string              test = argv[3];
    string              s;
    GetEnv(ENV_ABC, s);
    EAlphabet           AlphaType = GetAlphabetType(s);
    if (AlphaType == eUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return -1;
    }

    CTrieIndex ind(hasTransp, AlphaType);

    ifstream inpTest(test.c_str());

    if (!inpTest) {
        Perror("Cannot open file" + test);
        return -1;
    }

    ifstream inpDict(dict.c_str());

    if (!inpDict) {
        Perror("Cannot open file" + dict);
        return -1;
    }

    string      word;
    VS          testList;

    while (inpTest >> word) {
        testList.push_back(word);
    }

    while (inpDict >> word) {
        ind.Insert(word.c_str());
    }
    int testListSz = testList.size();

    DECLARE_STATIC_COUNTER(tm);
    START_COUNTER(tm);
    for (int i = 0; i < testListSz; ++i) {
        ind.Search(testList[i].c_str(), 0, NULL, NULL);
    }
    STOP_COUNTER(tm);
    cout << "Avg: " << GET_COUNTER_MS(tm) / double(testListSz) << endl;
};
