#include "trie.h"
#include "../Interface/triedeb.h"
#include <algorithm>

bool CTrie::SearchAtLevelExact(const char *pStr, int StrLen, int StartLen,
                              int *numComp,
                             StringList * resList, bool silent) const
{
   if (StartLen >= StrLen) {
       if (!silent) {
           cout << pStr << '\n';
       }
       if (resList != NULL) {
           assert(StrLen);
           resList->push_back (string(pStr, StrLen - 1));
       }

       return true;
   }
   int chSize = children.size();
   
   if (!chSize) {
       return false;
   }

   unsigned int minPos =  0, sz = chSize;

   const char *pBeg = pStr + StartLen;
   int cmp;
   unsigned int half; 
   unsigned int middle;

   while (sz > 0) {
       half = sz >> 1;
       middle = minPos + half;

       const Child *pChild = &children[middle];
       CALC_NUM_COMP(numComp, pChild);

       cmp = strcmp(pChild->label.c_str(), pBeg);
       if (cmp < 0) {
           minPos = middle + 1;
           sz = sz - half - 1;
       } else {
           sz = half;
       }
   }

   if (minPos == chSize)  {
        --minPos;
   }
   
   if (1) {
        const Child *pChld = &children[minPos];
        int len = pChld->checkCommon(pStr, StartLen);
        // We stand after if we have an exact prefix
        CALC_NUM_COMP(numComp, pChld);
        if (len < pChld->label.size() && minPos) {
            pChld = &children[minPos - 1];
            len = pChld->checkCommon(pStr, StartLen);
        }
        if (len >= pChld->label.size()) {
            int StartLen1 = StartLen + pChld->label.size();
            if (StartLen1 == StrLen) {
                if (!silent) {
                    cout << pStr << '\n';
                }
                if (resList != NULL) {
                    assert(StrLen);
                    resList->push_back (string(pStr, StrLen - 1));
                }
                return true;
            } else {
                if (pChld->ptr) {
                    return pChld->ptr->SearchAtLevelExact(pStr, StrLen, StartLen1, numComp, resList,silent);
                }
            }
        }
   }
   return false;
}

void CTrie::SearchAtLevel(const std::string & s, 
                const char *st, int stLen,
#ifdef USE_AFUZZYCONTEXT
                const afuzzycontext & a, const context & tmpCon,
#else
                const checkDistContext & c, 
#endif
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    int allocSize = stLen + 32;
    char *sk = (char *)alloca(allocSize + 1);
    strcpy(sk, st);
#ifdef USE_AFUZZYCONTEXT
    context c1;
#endif
    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
#ifdef USE_AFUZZYCONTEXT
        c1 = tmpCon; // copy is faster than every time construction 
#else
        checkDistContext c1 = c;
#endif
        int              childSz = children[i].label.size();
        int              newSz = stLen + childSz;
        if (newSz > allocSize) {
            allocSize = newSz;
            sk = (char *)alloca(allocSize + 1);
            strcpy(sk, st);
        }
        strcpy(sk + stLen, children[i].label.c_str());

        const CTrie *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree);
#ifndef USE_AFUZZYCONTEXT
        if (c1.checkDist (sk, SubTree == NULL, hasTransp)) 
#else
        if (a.checkRTR (sk, newSz, &c1, SubTree == NULL, NULL))    
#endif
        {
            if (SubTree) {
                SubTree->SearchAtLevel(s, 
                                        sk, newSz,
#ifdef USE_AFUZZYCONTEXT
                                        a,
#endif
                                        c1, numComp, resList, hasTransp, silent);
            } else {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    //resList->push_back (sk.substr (0, sk.size () - 1));
                    resList->push_back (string(sk, newSz - 1));
            }
        }
    }
}

bool CTrie::SearchInFile(int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent)
{
    string ss = s + EOS;
#ifdef USE_AFUZZYCONTEXT
    context tmpCon;
    afuzzycontext a;

    a.init(ss, kerr, hasTransp, &tmpCon);
#else
    checkDistContext c (kerr, ss.c_str ());
#endif

    return (kerr) ? SearchInFileAtLevel(fd, 0, ss, "", 
#ifdef USE_AFUZZYCONTEXT
                        a, tmpCon,
#else
                        c, 
#endif
                        numComp, resList, hasTransp, silent):
                    SearchInFileAtLevelExact (fd, 0, ss, "", numComp, resList, silent);
}

bool CTrie::SearchInFileAtLevel(int fd, off_t offset,
                                    const std::string & s, std::string st,
#ifdef USE_AFUZZYCONTEXT
                                    const afuzzycontext & a, const context & tmpCon,
#else
                                    const checkDistContext & c,
#endif
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
#ifdef USE_AFUZZYCONTEXT
    context c1;
#endif
    for (int i = 0; i < num; i++) {
#ifdef USE_AFUZZYCONTEXT
        context c1 = tmpCon;// faster than constructor
#else
        checkDistContext c1 = c; 
#endif

        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        CALC_NUM_COMP(numComp, sub_off);
#ifndef USE_AFUZZYCONTEXT
        if (c1.checkDist (sk.c_str (), sub_off == 0, hasTransp)) {
#else
        if (a.checkRTR (sk.c_str(), sk.size(), &c1, sub_off == 0, NULL)) {
#endif
            if (sub_off) {
                if (!SearchInFileAtLevel
                    (fd, sub_off, s, sk, 
#ifdef USE_AFUZZYCONTEXT
                                        a,
#endif
                                        c1, 
                                        numComp, resList, hasTransp, silent))
                    return false;
            } else {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

bool CTrie::SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        CALC_NUM_COMP(numComp, sub_off);
        if (sub_off) {
            if (sk.size() <= s.size() && !strncmp(s.c_str(), sk.c_str(), sk.size())) {
                if (!SearchInFileAtLevelExact(fd, sub_off, s, sk, numComp, resList, silent))
                    return false;
            }
        } else {
            if (sk == s) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

CTrie::CTrie()
{
}

CTrie::~CTrie()
{
    for (int i = 0; i < children.size(); ++i) {
        if (children[i].ptr) delete children[i].ptr;
    }
}

void CTrie::Search(const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const
{
    if (!kerr) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    string ss = s + EOS;
#ifdef USE_AFUZZYCONTEXT
    context tmpCon;
    afuzzycontext a;

    a.init(ss, kerr, hasTransp, &tmpCon);
#else
    checkDistContext c (kerr, ss.c_str ());
#endif

    SearchAtLevel (ss, "",  0,
#ifdef USE_AFUZZYCONTEXT
                    a, tmpCon,
#else
                    c, 
#endif
                    numComp, resList, hasTransp, silent);
}

bool CTrie::SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const
{
    int sz = s.size();
    char aStr[sz + 2];
    memcpy(aStr, s.c_str(), sz);
    aStr[sz] = EOS[0];
    aStr[sz + 1] = 0;

    return SearchAtLevelExact(aStr, sz + 1, 0, numComp, resList, silent);
}

void CTrie::Insert(const std::string &s)
{
    int indx = -1;
    int maxLen = 0;
    int len;
    for (int i = 0; i < children.size(); ++i) 
    if ((len = children[i].checkCommon(s, 0)) > maxLen) {
        indx = i;
        maxLen = len;
    }
    if (!maxLen) {
        Child chld(s);

        vector<Child>::iterator it = lower_bound(children.begin(), children.end(), chld);
        children.insert(it, chld);
        return;
    }
    string rest(s.begin() + maxLen, s.end());
    if (rest.empty()) {
        throw string("String '") + s + string("' length: ") + Format("%d", s.size()) + string(" is a duplicate");
    }
    Child &maxChld = children[indx];
    // If the string to be inserted has label as its prefix
    // insert it into the corresponding subtree
    if (maxLen == maxChld.label.size()) {
        if (maxChld.ptr) {
            maxChld.ptr->Insert(rest);
        } else {
            throw string("String '") + s + string("' has a proper prefix in the dictionary!");
        }
        return;
    }
    //... otherwise the node must be splitted and subtree be created
    string restChld(maxChld.label.begin() + maxLen, maxChld.label.end());
    // Shorten the label
    maxChld.label.erase(maxChld.label.begin() + maxLen, maxChld.label.end());
    CTrie *oldPtr = maxChld.ptr;
    CTrie *pNewNode= new CTrie;
    maxChld.ptr = pNewNode;
    pNewNode->children.push_back(Child(restChld, oldPtr));
    pNewNode->children.push_back(Child(rest));
    if (pNewNode->children[0] > pNewNode->children[1]) {
        swap(pNewNode->children[0], pNewNode->children[1]);
    }
}

void CTrie::Print(int level)
{
    string indent(level, ' ');
    cout << indent << "Node address =" << this << " level=" << level << "\n";
    for (int i = 0; i < children.size(); ++i) {
        cout << indent << " Key " << children[i].label << " : " << children[i].ptr << endl;
        if (children[i].ptr) {
            children[i].ptr->Print(level + 1);
        }
    }
}

int CTrie::DumpToFile(int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of leaves
// 2. the leaf keys themselvs, with zero offsets
// 3. leaf subtrees - updating offsets

    int buflen = 0;
    int sub_sz = 0;
    int cnt = children.size();
    buflen += sizeof (cnt) + sizeof (buflen);
    for (int i = 0; i < cnt; ++i) {
        buflen += sizeof (off_t) /* <- pointer size */ + 
                 + 1 /* string record size (must be<255) */  
                 + children[i].label.size() + 1;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    for (int i = 0; i < cnt; ++i) {
        // String length
        int len = children[i].label.size();
        *((u_int8_t *) (void *) sbuf) = len + 1;
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) ((void *) sbuf);
        *ptrs[i] = 0;
        // Leaf key
        sbuf += sizeof (off_t);
        memcpy (sbuf, children[i].label.c_str (), len + 1);
        sbuf += len + 1;
    }
    // Step 3.
    for (int i = 0; i < cnt; ++i) {
        // Diving into subtree
        int s_sz = (children[i].ptr)? children[i].ptr->DumpToFile (fd, offset + buflen + sub_sz) : 0;
        if (s_sz > 0)
            *ptrs[i] = offset + buflen + sub_sz;
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Returning the total number of bytes written
    return buflen + sub_sz;
}

int CTrie::countLeaves() {
    int res = 0;

    int sz = children.size();

    if (!sz) { ++res; }
    else { 
       for (int i = 0; i < sz; ++i) {
           if (children[i].ptr) {
               res += children[i].ptr->countLeaves();
           } else {
               ++res;
           }
       }
    }

    return res;
}

void CTrie::countBranch(vector<int> &cnt, vector<int> &cntBranch, vector<int> &cntLeaves, int level)
{
    if (level + 1 > cnt.size()) {
        cnt.resize(level + 1);
        cntBranch.resize(level + 1);
        cntLeaves.resize(level + 1);
    }
    int sz = children.size();

#if 0
    if (!level) {
        cout << "Have: " << sz << " children\n\n";
    }
#endif

    cnt[level]++;
    int leafNum = countLeaves();
    cntLeaves[level] += leafNum;
    cntBranch[level] += sz;

    for (int i = 0; i < sz; ++i) {
        const string &label = children[i].label;
        int   newLevel = level + label.size();
#if 0
        if (!level) {
            cout << "Child # " << i << " Label size: " << label.size() << " label: " << label << endl;
        }
#endif

        if (newLevel + 1 > cnt.size()) {
            cnt.resize(newLevel + 1);
            cntBranch.resize(newLevel + 1);
            cntLeaves.resize(newLevel + 1);
        }

        for (int j = 1; j < label.size(); ++j) {
            cnt[level + j]++; 
            cntLeaves[level + j] += leafNum;
            cntBranch[level + j] += 1;
        }
        if (children[i].ptr) {
            children[i].ptr->countBranch(cnt, cntBranch, cntLeaves, newLevel);
        }
    }
}
