#include "afuzreg.h"
extern int *ptr;
/*
 Author: Leonid Boitsov
 Date  : Summer 2000
 This file could be freely distributed under the terms of GPL
*/

char afuzreg::CurrChar (bool * Spec, const char *p)
{
    *Spec = false;
    if (p[patcnt]) {
        if (p[patcnt] == SpecChar) {
            if (p[patcnt + 1]) {
                *Spec = true;
                return (IgnoreCase) ? tolower (p[patcnt + 1]) : p[patcnt +
                                                                  1];
            } else
                return 0;
        } else
            return (IgnoreCase) ? tolower (p[patcnt]) : p[patcnt];
    }
    return 0;
};

char afuzreg::NextChar (bool * Spec, const char *p)
{
    int cnt = patcnt;
    ToNextChar (p);
    char c = CurrChar (Spec, p);
    patcnt = cnt;
    return c;
};

void afuzreg::ToNextChar (const char *p)
{
    if (p[patcnt]) {
        if (p[patcnt] == SpecChar) {
            if (p[patcnt + 1])
                patcnt += 2;
            else
                patcnt++;
        } else
            patcnt++;
    }
};

bool afuzreg::ParseReg (const char *p, char *ErrStr, int ErrLen)
{
    char estr[1024];
    memset (Map, 0, sizeof (Map));
    m = 0;
    cnt = 0;
    int cntSet = 0;
    char FirstRangeChar;
    Uint b = 1;
    //***************************
    SWild = 0;
    SLeftMostWild = 0;
    SAny = 0;
    //****** State **************
    State st = Normal;
    bool Spec, Specn;
    char c;
    S = (Uint *) malloc (sizeof (Uint));
    S[0] = 0;
    for (patcnt = 0; CurrChar (&Spec, p); ToNextChar (p)) {
        bool Spec;
        c = CurrChar (&Spec, p);
        switch (c) {
        case WildCard:
            if (!(st & AfterLeftBracket) && !Spec) {
                bool s;
                // The sequence of wildcards are ignored
                if (b > 1 && NextChar (&s, p) != WildCard) {
                    SWild |= (b >> 1);
                } else if (b == 1 && NextChar (&s, p) != WildCard) {
                    SLeftMostWild = 1;
                }
                break;
            }
        case AnyChar:
            if (!(st & AfterLeftBracket) && !Spec) {
                SAny |= b;
                b <<= 1;
                m++;
                break;
            }
        case LeftBracket:
            if (!Spec) {
                if ((st & AfterLeftBracket)) {
                    sprintf (estr,
                             "Syntax error: left bracket in pos %d\n",
                             patcnt);
                    strncpy (ErrStr, estr, ErrLen);
                    return false;
                }
                FirstRangeChar = 0;
                st = State (int (AfterLeftBracket) | int (Normal));
                cntSet = 0;
                break;
            }
        case RightBracket:
            if (!Spec) {
                if (!(st & AfterLeftBracket)) {
                    sprintf (estr,
                             "Syntax error: right bracket in pos %d\n",
                             patcnt);
                    strncpy (ErrStr, estr, ErrLen);
                    return false;
                }
                if ((st & AfterRangeChar)) {
                    sprintf (estr,
                             "Syntax error: right bracket in pos %d just after the char %c\n",
                             patcnt, RangeChar);
                    strncpy (ErrStr, estr, ErrLen);
                    return false;
                }
                // Now everything is done: we increment the number of letters
                m++;
                b <<= 1;
                st = Normal;
                cntSet = 0;
                break;
            }
        case RangeChar:
            if (!Spec && (st & AfterLeftBracket)) {
                if (!IsLatinChar (FirstRangeChar)
                    && !isdigit (FirstRangeChar)) {
                    sprintf (estr,
                             "Syntax error: ranges are allowed only with digits or with lattin letters of the same case, pos %d\n",
                             patcnt);
                    strncpy (ErrStr, estr, ErrLen);
                    return false;
                }
                st = State (int (st) | int (AfterRangeChar));
                break;
            }
        default:
            if (!(st & AfterLeftBracket)) {
                int l = Map[c];
                if (!l) {
                    l = Map[c] = ++cnt;
                    S = (Uint *) realloc ((void *) S,
                                          (cnt + 1) * sizeof (Uint));
                    if (!S) {
                        strncpy (ErrStr, "Not enough memory", ErrLen);
                        return false;
                    }
                    S[0] = 0;
                    // Do not forget to let all chars that not in p have 0 mask
                    S[l] = 0;
                }               // if (!l)
                S[l] |= b;
                m++;
                b <<= 1;
            }                   // if (!st&AfterLeftBracket ...
            else {
                bool s;
                if (NextChar (&s, p) == RangeChar)
                    FirstRangeChar = c;
                if (!(st & AfterRangeChar)) {
                    int l = Map[c];
                    if (!l) {
                        l = Map[c] = ++cnt;
                        S = (Uint *) realloc ((void *) S,
                                              (cnt + 1) * sizeof (Uint));
                        if (!S) {
                            strncpy (ErrStr, "Not enough memory", ErrLen);
                            return false;
                        }
                        S[l] = 0;
                    }           // If (!l)
                    S[l] |= b;
                }               // if(!(st &AfterRangeChar
                else {
                    if ((!IsLatinChar (c) && !isdigit (c))
                        ||
                        (IsLatinUpper (c) != IsLatinUpper (FirstRangeChar))
                        ||
                        (IsLatinLower (c) != IsLatinLower (FirstRangeChar))
                        || (isdigit (c) != isdigit (FirstRangeChar))
                        )
                        // If c and FirstRangeChar are both digits then both
                        // IsLatinUpper and IsLatinLower equals to 0
                    {
                        sprintf (estr,
                                 "Syntax error: ranges are allowed only with digits (1-4) or with lattin letters of the same case pos %d\n",
                                 patcnt);
                        strncpy (ErrStr, estr, ErrLen);
                        return false;
                    }
                    for (int ic = (int)(unsigned char)FirstRangeChar;
                         ic <= (int) (unsigned char) c; ic++) {
                        int l = Map[ic];
                        if (!l) {
                            l = Map[ic] = ++cnt;
                            S = (Uint *) realloc ((void *) S,
                                                  (cnt +
                                                   1) * sizeof (Uint));
                            if (!S) {
                                strncpy (ErrStr, "Not enough memory",
                                         ErrLen);
                                return false;
                            }
                            S[l] = 0;
                        }
                        S[l] |= b;
                    }           // end for
                    st = State (int (st) ^ AfterRangeChar);
                }               // else
            }
            // end of the last else 
            break;
        }
    }
    if (st & AfterLeftBracket) {
        sprintf (estr, "Syntax error: right bracket not found!\n", patcnt);
        strncpy (ErrStr, estr, ErrLen);
        return false;
    }
    //

    return true;
};


afuzreg::afuzreg ()
:  S (NULL), R (NULL), RI (NULL), RP (NULL), R1 (NULL), m (0)
{
}


bool afuzreg::init (const char *p, int kerr, char *ErrStr, int ErrLen,
                    bool _IgnoreCase)
{
    // First - free memory that is left from the previous init
    FreeMem ();
    IgnoreCase = _IgnoreCase;
    //
    // To determine the size of the S array we need to parse the string p
    // Parsing procedure allocates memory for S and
    // modify S, SWild, SAny and Map
    // 
    k = kerr;
    if (!ParseReg (p, ErrStr, ErrLen))
        return false;
    if (m > MaxPatSize ()) {
        char str[128];
        sprintf (str, "The size of the pattern exceeds %d!",
                 MaxPatSize ());
        strncpy (ErrStr, str, ErrLen);
        return false;
    }

    R = new Uint[k + 1];
    R1 = new Uint[k + 1];
    RI = new Uint[k + 1];
    RP = new Uint[k + 1];

    for (int d = 0; d <= k; d++)
        RI[d] = (1 << d) - 1;

    mask_ok = (1 << (m - 1));

    r_size = sizeof (Uint) * (k + 1);
    int p_len = m;
    if (p_len > sizeof (Uint) * 8)
        p_len = sizeof (Uint) * 8;
    return true;
};

void afuzreg::FreeMem ()
{
    if (S)
        free (S);
    if (R)
        delete[]R;
    if (R1)
        delete[]R1;
    if (RI)
        delete[]RI;
    if (RP)
        delete[]RP;
}

afuzreg::~afuzreg ()
{
    FreeMem ();
};

bool afuzreg::checkREG (const char *t)
{
    if (!m)
        return true;
    memcpy (R, RI, r_size);     // R = RI
    Uint SMapP = 0;
    register Uint SMap;


    register char c;
    for (register int j = 0; t[j]; j++) {
        for (register int d = 0; d <= k; d++) {
            SMap = S[Map[t[j]]];
            R1[d] =
                (((R[d] << 1) | (j <= d) | SLeftMostWild) & (SMap | SAny));
            if (d > 0) {
                R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | 1 | R[d - 1];
            }
            R1[d] = R1[d] | (R[d] & SWild);
            /* Now transpositions */
            if (j > 0 && d)
                R1[d] |=
                    (RP[d - 1] << 2 | (((j <= d) | SLeftMostWild) << 1)) &
                    SMapP & (SMap << 1);
        }
        memcpy (RP, R, r_size);
        memcpy (R, R1, r_size);
        SMapP = SMap;

    }                           // end for(register int j=0 ...

    return (R1[k] & mask_ok);
}
