#include "afuzzy.h"

afuzzy::afuzzy(bool hasTransp):R (NULL), R1 (NULL), RP (NULL), S (NULL), RI (NULL),
FilterS (NULL), mHasTransp(hasTransp)
{;
}

void afuzzy::init (const char *p, unsigned int kerr, bool UseFilter)
{
    k = kerr;
    m = strlen (p);
    FilterSet = false;
    memset (Map, 0, sizeof (Map));
    if (S)
        delete[]S;
    if (R)
        delete[]R;
    if (R1)
        delete[]R1;
    if (RP)
        delete[]RP;
    if (RI)
        delete[]RI;
    if (FilterS)
        delete[]FilterS;
    FilterS = NULL;
    S = new Uint[m + 1];
    R = new Uint[k + 1];
    R1 = new Uint[k + 1];
    RI = new Uint[k + 1];
    RP = new Uint[k + 1];
    S[0] = R[0] = R1[0] = RI[0] = 0;
    for (unsigned int i = 0, cnt = 0; i < m; i++) {
        int l = Map[(unsigned char) p[i]];
        if (!l) {
            l = Map[(unsigned char) p[i]] = ++cnt;
            S[l] = 0;
        }
        S[l] |= Uint(1) << i;
    }


    for (unsigned int d = 0; d <= k; d++)
        RI[d] = (Uint(1) << d) - 1;

    mask_ok = (Uint(1) << (m - 1));
    r_size = sizeof (Uint) * (k + 1);
    unsigned int p_len = m;
    if (p_len > sizeof (Uint) * 8)
        p_len = sizeof (Uint) * 8;
    /* If k is zero then no filter is needed! */
    if (k && (p_len >= 2 * (k + 1)))
        if (UseFilter) {
            FilterSet = true;
            memset (FilterMap, 0, sizeof (FilterMap));
            FilterS = new Uint[m + 1];
            FilterS[0] = 0;
            // Not let's fill the interleaved pattern
            unsigned int dd = p_len / (k + 1);
            p_len = dd * (k + 1);
            for (unsigned int i = 0, cnt = 0; i < dd; i++)
                for (unsigned int j = 0; j < (k + 1); j++, cnt++)
                    PatFilter[cnt] = (unsigned char) p[j * dd + i];
            PatFilter[p_len] = 0;

            for (unsigned int i = 0, cnt = 0; i < p_len; i++) {
                int l = FilterMap[PatFilter[i]];
                if (!l) {
                    l = FilterMap[PatFilter[i]] = ++cnt;
                    FilterS[l] = 0;
                }
                FilterS[l] |= Uint(1) << i;
            }
            filter_ok = 0;
            for (unsigned int i = p_len - k - 1; i <= p_len - 1; i++)   // k+1 times
                filter_ok = filter_ok | (Uint(1) << i);

            filter_shift = (Uint(1) << (k + 2)) - 1;  // k+1 first bits set to 1
        }
}

afuzzy::~afuzzy ()
{
    delete[]S;
    delete[]R;
    delete[]R1;
    delete[]RI;
    delete[]RP;
    if (FilterSet) {
        delete[]FilterS;
    };
}

/**  */
bool afuzzy::checkRTR (const char *t, unsigned int n) const
{
    if (m == 0)
        return n <= k;
    
    register unsigned int cnt = 0, cnt1;
    memcpy (R, RI, r_size);     // R = RI
    Uint SMapP = 0;

    for (register unsigned int j = 0; j < n; j++) {
        cnt1 = cnt;
        register Uint SMap = S[Map[(unsigned char) t[j]]];
        for (register unsigned int d = cnt; d <= k; d++) {
            R1[d] = (((R[d] << 1) | (j <= d)) & SMap);
            if (d > cnt) {
                R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | R[d - 1] | (j <= d - 1);
                /* Now transpositions */
                if (mHasTransp) {
                    if (j > 0 && d) {
                        R1[d] |= (RP[d - 1] << 2 | (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
                    }
                }
            }
            // Here R[d] == RP[d] for the next step
#           if 1
            if (d == cnt1 && j > d && !R1[d] && !R[d]) {
                ++cnt1;
            }
#           endif
        }

        for (register unsigned int i = cnt; i <= k; i++) {// R->RP, R1->R 
            RP[i] = R[i];
            R[i] = R1[i];
        }
        SMapP = SMap;

        cnt = cnt1;
        if (cnt > k) {
            return false;
        }
    }

    return (R[k] & mask_ok);
}


bool afuzzy::checkRTR (const char *t, unsigned int *dist, char term_char) const
{
    if (m == 0) {
        unsigned int n = 0;
        for (; t[n] != term_char; n++); /* empty body */

        return n <= k;
    }
    
    register unsigned int cnt = 0, cnt1;

    memcpy (R, RI, r_size);     // R = RI
    Uint SMapP = 0;

    for (register unsigned int j = 0; t[j] != term_char; j++) {
        cnt1 = cnt;
        register Uint SMap = S[Map[(unsigned char) t[j]]];
        for (register unsigned int d = cnt; d <= k; d++) {
            R1[d] = (((R[d] << 1) | Uint(j <= d)) & SMap);
            if (d > cnt) {
                R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | R[d - 1] | (j <= d - 1);
                /* Now transpositions */
                if (mHasTransp) {
                    if (j > 0 && d) {
                        R1[d] |= (RP[d - 1] << 2 | (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
                    }
                }
            }
            // Here R[d] == RP[d] for the next step
#           if 1
            if (d == cnt1 && j > d && !R1[d] && !R[d]) {
                ++cnt1;
            }
#           endif
        }

        for (register unsigned int i = cnt; i <= k; i++) { // R->RP, R1->R
            RP[i] = R[i];
            R[i] = R1[i];
        }
        SMapP = SMap;

        cnt = cnt1;
        if (cnt > k) {
            return false;
        }
    }                   

    if (dist) {
        register int l;
        for (l = k; l >= 0 && (R[l] & mask_ok); l--);

        *dist = l + 1;
        return l < int(k);
    } else
        return (R[k] & mask_ok);
}


bool afuzzy::checkSUB (const char *t, char term_char) const
{
    // For eficciency this case should be little bit optimized
    if (!k) {
        Uint R = 0, R1;

        register char c;
        for (register int j = 0; (c = t[j]) != term_char; j++) {
            R1 = (((R << 1) | 1) & S[Map[(unsigned char) c]]);
            R = R1;

            if (R1 & mask_ok)
                return true;
        }                       // end for(register int j=0 ...
        return false;
    }
    if (FilterSet && !checkFLTterm (t, term_char))
        return false;
    memcpy (R, RI, r_size);     // R = RI

    register char c;
    for (register int j = 0; (c = t[j]) != term_char; j++) {
        for (unsigned register int d = 0; d <= k; d++) {
            R1[d] = (((R[d] << 1) | 1) & S[Map[(unsigned char) c]]);
            if (d > 0) {
                R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | 1 | R[d - 1];
            }
        }
        if (R1[k] & mask_ok)
            return true;
        memcpy (R, R1, r_size);

    }                           // end for(register int j=0 ...

    return false;
}


bool afuzzy::checkSUB (const char *t, unsigned int n) const
{
    // For eficciency this case should be little bit optimized
    if (!k) {
        Uint R = 0, R1;

        register char c;
        for (register int j = 0; j < n; j++) {
            R1 = (((R << 1) | 1) & S[Map[(unsigned char) c]]);
            R = R1;

            if (R1 & mask_ok)
                return true;
        }                       // end for(register int j=0 ...
        return false;
    }
    if (FilterSet && !checkFLT (t, n))
        return false;
    memcpy (R, RI, r_size);     // R = RI

    for (unsigned register int j = 0; j < n; j++) {
        for (unsigned register int d = 0; d <= k; d++) {
            R1[d] = (((R[d] << 1) | 1) & S[Map[(unsigned char) t[j]]]);
            if (d > 0) {
                R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | 1 | R[d - 1];
            }
        }
        if (R1[k] & mask_ok)
            return true;
        memcpy (R, R1, r_size);

    }                           // end for(register int j=0 ...

    return false;
}


bool afuzzy::checkFLT (const char *t, unsigned int n) const
{
    register Uint FilterR = 0;
    register Uint FilterR1;

    for (register unsigned int j = 0; j < n; j++) {
        FilterR1 =
            (((FilterR << (k + 1)) | filter_shift) &
             FilterS[FilterMap[(unsigned char) t[j]]]);
        if (FilterR1 & filter_ok)
            return true;
        FilterR = FilterR1;
    }                           // end for(register int j=0 ...

    return false;
}

bool afuzzy::checkFLTterm (const char *t, char term_char) const
{
    register Uint FilterR = 0;
    register Uint FilterR1;

    for (register int j = 0; t[j] != term_char; j++) {
        FilterR1 =
            (((FilterR << (k + 1)) | filter_shift) &
             FilterS[FilterMap[(unsigned char) t[j]]]);
        if (FilterR1 & filter_ok)
            return true;
        FilterR = FilterR1;
    }                           // end for(register int j=0 ...

    return false;
}
