#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <iterator>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

#define TEST_MYAGREPREAD

using namespace std;

#define LEVEN_DFA       "levendfa"
#define LEVEN_DFA_STAT  "levendfa_stat"

void usage(const char *progName)
{
    cerr << "Usage: " << progName << " -ns<minError> -ne<maxError> <patternfile> <dictfile> <optional transp filter 0|1>\n";
}

string escape_shell(const string & src)
{
    string res = src;

    int cnt = res.size();
    for (int i = 0; i < cnt; ++i) 
    if (res[i] == '\'' || res[i] == '"') {
       res.insert(res.begin() + i, '\\');
       ++cnt;
       ++i;
    }
    return res;
}

string replace_all(const string & src, const string & from, const string & to)
{
    string res = src;

    int n;
    while ((n = res.find(from))  != string::npos) {
        res.replace(n, from.size(), to);
    }
    return res;
}

int main(int argc, char *argv[])
{
    const char *runCmd[] = {
        // Pipe is to simulate complete-word match
        "../agrep-2.04/agrep -<num> <pattern> <file> | ../agrep-2.04/agrep -<num> -x <pattern>",
        //"../agrep-2.04/agrep -<num> -x <pattern> <file>",
        "../nrgrep-1.1.2/nrgrep -k <num>ids -x <pattern> <file>",
        "../nrgrep-1.1.2/nrgrep -k <num>idst -x <pattern> <file>",
#ifdef TEST_MYAGREPREAD
        "./myagrepread -fexact -n<num> -t0 <pattern> <file>",
        "./myagrepread -fexact -n<num> -t1 <pattern> <file>",
        "./myagrepread -fagrep -n<num> -t0 <pattern> <file>",
        "./myagrepread -fagrep -n<num> -t1 <pattern> <file>",
        "./myagrepread -fcombined -n<num> -t0 <pattern> <file>",
        "./myagrepread -fcombined -n<num> -t1 <pattern> <file>",
#endif
        "../NavarroDFA/search_complw <pattern> <num> <file>",
        "../LevenDFA/grepdfa <num> 0 <pattern> <file> " LEVEN_DFA_STAT,
        "../LevenDFA/grepdfa <num> 1 <pattern> <file> " LEVEN_DFA_STAT,
        "./grepmyers <num> 0 <pattern> <file> ",
        "./grepmyers <num> 1 <pattern> <file> ",
    };
    const char *cmdName[] = {
        "agrep pipe -x",
        //"agrep -x", // This one IS SLOW
        "nrgrep -x -k..[ids]",
        "nrgrep -x -k..[idst]",
#ifdef TEST_MYAGREPREAD
        "myagrepread filter=exact_delayed transp=0",
        "myagrepread filter=exact_delayed transp=1",
        "myagrepread filter=agrep transp=0",
        "myagrepread filter=agrep transp=1",
        "myagrepread filter=combined transp=0",
        "myagrepread filter=combined transp=1",
#endif
        "lazydfa transp=0",
        LEVEN_DFA " transp=0",
        LEVEN_DFA " transp=1",
        "grepmyers transp=0",
        "grepmyers transp=1",
    };
    const bool aHasTransp[] = {
        false,  // agrep pipe -x
        // false, // for agrep -x
        false, // nrgrep
        true,  // nrgrep
#ifdef TEST_MYAGREPREAD
        false,
        true,
        false,
        true,
        false,
        true,
#endif
        false,  // lazydfa
        false,  // LEVEN_DFA
        true,   // LEVEN_DFA
        false,  // grepmyers
        true,   // grepmyers
    };

    assert(sizeof runCmd == sizeof cmdName);
    int progNum = sizeof runCmd / sizeof *runCmd;
    if (argc < 5) {
         usage(argv[0]);
         return -1;
    } if (strncmp(argv[1], "-ns", 3)) {
         cerr << "Wrong minError argument\n";
         usage(argv[0]);
         return -1;
    }
    if (strncmp(argv[2], "-ne", 3)) {
         cerr << "Wrong maxError argument\n";
         usage(argv[0]);
         return -1;
    }
    bool  bUseTranspFilter = false;
    bool  bHasTransp;  
    if (argc >= 6) {
        bUseTranspFilter = true;
        bHasTransp = atol(argv[5]);
    }
    const char *pMinErr = argv[1] + 3;
    const char *pMaxErr = argv[2] + 3;
    int maxErr = atol(pMaxErr);
    int minErr = atol(pMinErr);
    vector<string> testStrings;
    ifstream       testFile(argv[3]);

    if (!testFile) {
         cerr << "Cannot open '" << argv[3] << "' : " << strerror(errno) << endl;
         return -1;
    }
    string         s;
    while (testFile >> s) {
         // This is essential for agrep
         if (s.size() > maxErr 
             && s[0] != '-') {
             testStrings.push_back(s);
         }
    }
#if 0
    testStrings.erase(testStrings.begin() + 7, testStrings.end());
#endif
    long long           runTm[maxErr + 1][progNum];
    timeval             t1, t2;

    memset(runTm, 0, sizeof runTm);

    for (int k = minErr; k <= maxErr; ++k) {
        char buf_k[32];
        char buf[2048];
        snprintf(buf_k, sizeof buf_k, "%d", k);
        for (int i = 0; i < progNum; ++i) {
            if (bUseTranspFilter && bHasTransp != aHasTransp[i]) {
                continue;
            }
            bool IsLevenDFA = strncmp(cmdName[i], LEVEN_DFA, strlen(LEVEN_DFA)) == 0; 

            if (IsLevenDFA && k > 2) {
                    continue;
            }
            // Skip a test for levendfa if k > 2
            {
                ofstream EraseDFAStat(LEVEN_DFA_STAT); // Just create an empty file
            }
            for (int n = 0; n < testStrings.size(); ++n) {
                string cmd = replace_all(runCmd[i], "<num>", buf_k);
                cmd = replace_all(cmd, "<pattern>", escape_shell(testStrings[n]) );
                cmd = replace_all(cmd, "<file>", argv[4]);
                
                if (getenv("PRINT_CMD") && atol(getenv("PRINT_CMD"))) {
                    cout << cmd << endl;
                    continue;
                }
                gettimeofday(&t1, NULL);
                FILE *f = popen(cmd.c_str(), "r");
                if (!f) {
                    cerr << "Cannot popen '"<< cmd << "' : " << strerror(errno) << endl;
                    return -1;
                }
                while(fgets(buf, sizeof buf, f)) {
                }
                pclose(f);
                gettimeofday(&t2, NULL);
                long long dt = ((t2.tv_sec - t1.tv_sec) * 1000000LL + t2.tv_usec - t1.tv_usec);
                runTm[k][i] += dt; 
                cout << dt / 1000.0 << "\t:" << cmd << endl;
            }
            if (IsLevenDFA) { // Adjust runtimes by subtracting automaton creation times
                ifstream DFAStat(LEVEN_DFA_STAT);
                int      tm;
                int      n = 0;
                for (; (DFAStat >> tm) && n < testStrings.size(); ++n) {
                    //cout << "Adjusting by " << tm/1000.0 << " Ms\n";
                    runTm[k][i] -= tm;
                }
                if (n < testStrings.size()) {
                    cerr << "Cannot get full stat, read " << n << " timings for " << testStrings.size() << " strings" << endl;
                    return 1;
                }
            }
        }
    }
    for (int k = minErr; k <= maxErr; ++k) {
        for (int i = 0; i < progNum; ++i) {
            cout << "K = " << k << " " <<  ": \t" << double(runTm[k][i]) / (1000.0 * testStrings.size()) << "\t" << cmdName[i] << endl;
        }
    }
    return 0;
}
