#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "myers.hpp"

using namespace std;

void Usage(const char* pProg)
{
    cerr << "Usage: " << pProg << " <kerr> <HasTransp> <pattern> <InputFileName> " << endl;
}

void ScanSegment(int kerr, char* pBeg, char* pEnd, CMyersEdist& md, char TermChar)
{
    const char* pCurr   = pBeg;

    for (char *p = pBeg; pBeg < pEnd; pBeg = p + 1) {
        p = strchr(pBeg, TermChar);

        if (!p) {
            cerr << "Internal error: expect TermChar" << endl;
            abort();
        }
        *p = 0;


        int dist = md.CalcDist(pBeg);
        
        if (dist <= kerr) {
            write(1, pBeg, p - pBeg);
            char c = '\n';
            write(1, &c, 1);
        }
        *p = TermChar;
    }
}

int FindPattern(const char *pPattern, const char *pInputFileName, unsigned int kerr, bool bHasTransp, char TermChar)
{
    CMyersEdist md(pPattern, bHasTransp);

    int fd = open(pInputFileName, O_RDONLY);

    if (fd < 0) {
        cerr << "Cannot open '" << pInputFileName << "' for reading" << endl;
        return 1;
    }

    int  nr = 0;
    int  maxReadLen = 1024 * 8;
    char buf[maxReadLen * 2 + 2];

    nr = read(fd, buf, maxReadLen);
    int start = 0;

    while (nr > 0)
    {

        char*   pBeg = buf; 
        char*   pEnd = pBeg + start + nr;

        *pEnd = TermChar; // There should be always a following TermChar

        ScanSegment(kerr, pBeg, pEnd, md, TermChar);

        int pos;
        for (pos = nr - 1; pos > 0 && pBeg[pos] != TermChar; --pos);
        start = nr - 1 - pos;
        if (start) {
            start = nr - 1 - pos;
            memmove(buf, pBeg + pos + 1, start);
        }
        nr = read(fd, &buf[start], maxReadLen);
    }

    return 0;
}

int main(int argc, char* pArgv[])
{


    if (argc != 5) {
        Usage(pArgv[0]);
        return 1;
    }

    unsigned int        kerr        = atol(pArgv[1]);
    bool                bHasTransp  = atol(pArgv[2]);

    return FindPattern(pArgv[3], pArgv[4], kerr, bHasTransp, '\n');
}
