#ifndef MY_AFUZZY_H
#define MY_AFUZZY_H

#include <string.h>
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include "interface.h"


class myagrep {
public:
private:
    int cnt;
    Uint *RI;
    unsigned int SMapArr[256];
    unsigned char mMember[8192];
    int shift_1;
    unsigned char mShift[256];
    unsigned int FilterSFilterMapArr[256];
    unsigned int kerr;
    unsigned int m;
    Uint mask_ok;
    Uint filter_ok;
    Uint filter_shift;
    Uint filterInitValue;
    unsigned int r_size;
    bool mHasTransp;
    bool FilterSet;
    /*
     * Boyer-Moore variables. 
     */
    vector<int>     suff;
    vector<int>     boyerMooreGoodSuffix;
    int             boyerMooreBadChar[256]; 
    string          mPattern;

    void findPatternBoyerMoore(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const;
    void findPatternNone(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const;
    void findPatternAgrep(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const;
    void findPatternExact(const char *pBegin, const char *pEnd, char termChar, bool silent, StringList *res) const;
    void findPatternExactImmediate(const char *pBegin, const char *pEnd, char termChar, bool silent, StringList *res) const;
    void suffixes(const char *pat);
    void initBoyerMooreGoodSuffix(const char *x);
    void initBoyerMooreBadChar(const char *pat);
    void initAgrepFilt(void);
    void agrepIntern(const char *text, const char *textend, char termChar, bool silent, StringList * res) const;

    inline void checkCand(const char *pBeg, const char *pEnd, const char *apCand[], int candCnt, char termChar, bool silent, StringList *res) const;

    void init(const char *p, unsigned int k, bool hasTransp);
    void initBoyerMoore(const char *p);
  public:
    inline bool checkRTR(const char *pBeg, const char *pEnd) const {return ((mHasTransp) ? checkRTRTransp(pBeg, pEnd) : checkRTRNoTransp(pBeg, pEnd));}
    inline bool checkRTRFilter(const char *pBeg, const char *pEnd) const ;
    inline bool checkRTRFilter(const std::string & str)  const { return checkRTRFilter(str.data(), str.data() + str.size()); }

    myagrep(const char *p, unsigned int k, bool hasTransp);
    ~myagrep ();
    // pEnd is at zero char, or one char after the last symbol
    bool inline checkRTRNoTransp(const char * pBeg, const char *pEnd) const;
    bool inline checkRTRTransp(const char * pBeg, const char *pEnd) const;
    void findPattern(AlgoType checkMeth, char termChar, 
                    const char *pBeg, const char *pEnd,
                    bool silent, StringList *res) const;
    void findPatternFile(int fd, AlgoType checkMeth, 
                    char termChar, bool silent = false, StringList *res = NULL) const;
    static unsigned int MaxPatSize () { return sizeof (Uint) * 8; }
};

    
bool inline myagrep::checkRTRTransp(const char *pBeg, const char *pEnd) const
{
    if (m == 0) {
        return (pEnd - pBeg) <= kerr;
    }
    unsigned int cnt = 0, cnt1;
    Uint     R[kerr+1], R1[kerr+1], RP[kerr+1];

    memcpy (R, RI, r_size);     // R = RI
    Uint        SMapP = 0, SMap;

    for (unsigned int j = 0; pBeg < pEnd; ++pBeg, ++j) {
        cnt1 = cnt;
        SMap = SMapArr[(unsigned char) *pBeg];
        // Here R[cnt] == RP[cnt] for the next step
        R1[cnt] = (((R[cnt] << 1) | Uint(j <= cnt)) & SMap);
        if (j > cnt && !R1[cnt] && !R[cnt]) {
            ++cnt1;
        }
        for (unsigned int d = cnt + 1; d <= kerr; d++) {
            R1[d] = (((R[d] << 1) | Uint(j <= d)) & SMap);
            R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | R[d - 1] | (j <= d - 1);
            /* Now transpositions */
            if (j > 0 && d) {
                R1[d] |= (RP[d - 1] << 2 | (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
            }
            // Here R[d] == RP[d] for the next step and new j = current j + 1
            if (d == cnt1 && j >= d && !R1[d] && !R[d]) {
                ++cnt1;
            }
        }
        for (unsigned int i = cnt; i <= kerr; i++) { // R->RP, R1->R
            RP[i] = R[i];
        }
        SMapP = SMap;

        cnt = cnt1;
        if (cnt > kerr) {
            return false;
        }

        ++pBeg, ++j;
        if (pBeg == pEnd) return (R1[kerr] & mask_ok);

        cnt1 = cnt;
        Uint SMap = SMapArr[(unsigned char) *pBeg];
        // Here R1[cnt] == RP[cnt] for the next step
        R[cnt] = (((R1[cnt] << 1) | Uint(j <= cnt)) & SMap);
        if (j > cnt && !R[cnt] && !R1[cnt]) {
            ++cnt1;
        }
        for (unsigned int d = cnt + 1; d <= kerr; d++) {
            R[d] = (((R1[d] << 1) | Uint(j <= d)) & SMap);
            R[d] |= ((R1[d - 1] | R[d - 1]) << 1) | R1[d - 1] | (j <= d - 1);
            /* Now transpositions */
            if (j > 0 && d) {
                R[d] |= (RP[d - 1] << 2 | (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
            }
            // Here R[d] == RP[d] for the next step
            if (d == cnt1 && j > d && !R[d] && !R1[d]) {
                ++cnt1;
            }
        }
        for (unsigned int i = cnt; i <= kerr; i++) { // R1->RP
            RP[i] = R1[i];
        }
        SMapP = SMap;

        cnt = cnt1;
        if (cnt > kerr) {
            return false;
        }
    }                   

    return (R[kerr] & mask_ok);
}

bool inline myagrep::checkRTRNoTransp (const char *pBeg, const char *pEnd) const
{
    if (m == 0) {
        return (pEnd - pBeg) <= kerr;
    }
    
    unsigned int cnt = 0, cnt1;
    Uint     R[kerr+1], R1[kerr+1];

    memcpy (R, RI, r_size);     // R = RI
    Uint SMapP = 0;

    for (unsigned int j = 0; pBeg < pEnd; ++pBeg, ++j) {
        cnt1 = cnt;
        Uint SMap = SMapArr[(unsigned char) *pBeg];
        R1[cnt] = (((R[cnt] << 1) | Uint(j <= cnt)) & SMap);
        if (j > cnt && !R1[cnt]) {
                ++cnt1;
        }
        for (unsigned int d = cnt + 1; d <= kerr; d++) {
            R1[d] = (((R[d] << 1) | Uint(j <= d)) & SMap);
            R1[d] |= ((R[d - 1] | R1[d - 1]) << 1) | R[d - 1] | (j <= d - 1);
            // next j is greater by 1
            if (d == cnt1 && j >= d && !R1[d]) {
                ++cnt1;
            }
        }

        cnt = cnt1;
        if (cnt > kerr) {
            return false;
        }
        
        ++pBeg, ++j;
        if (pBeg == pEnd) return (R1[kerr] & mask_ok);

        cnt1 = cnt;
        SMap = SMapArr[(unsigned char) *pBeg];
        R[cnt] = (((R1[cnt] << 1) | Uint(j <= cnt)) & SMap);
        if (j > cnt && !R[cnt]) {
            ++cnt1;
        }
        for (unsigned int d = cnt + 1; d <= kerr; d++) {
            R[d] = (((R1[d] << 1) | Uint(j <= d)) & SMap);
            R[d] |= ((R1[d - 1] | R[d - 1]) << 1) | R1[d - 1] | (j <= d - 1);
            
            if (d == cnt1 && j > d && !R[d]) {
                ++cnt1;
            }
        }

        cnt = cnt1;
        if (cnt > kerr) {
            return false;
        }
    }                   

    return (R[kerr] & mask_ok);
}

inline void myagrep::checkCand(const char *pBeg, const char *pEnd,
                        const char *apCand[], int candCnt, char termChar,
                        bool silent, StringList *res) const
{
    const char *pLast = NULL;
#if 0
    cout << "CheckCand cnt: " << candCnt << endl;
    return;
#endif

    for (int i = 0; i < candCnt; ++i) {
        if (pLast && apCand[i] <= pLast) continue;
        pLast = const_cast<const char *>(strchr(apCand[i], termChar));
        if (!pLast) pLast = pEnd;
            
        const char *p1 = apCand[i];
        for (;p1 > pBeg; --p1) {
            if (*p1 == termChar) {++p1;break;}
        }
        if (checkRTR(p1, pLast)) {
            if (!silent) {(void)write(1, p1, pLast - p1); putchar('\n');}
            if (res) {res->push_back(string(p1, pLast - p1));}
        }
    }
}

inline bool myagrep::checkRTRFilter(const char *pBeg, const char *pEnd) const
{
    if (!FilterSet) {
        return checkRTR(pBeg, pEnd);
    }
    Uint FilterR = filterInitValue;

    if (pEnd <= pBeg) {
       return m <= kerr;
    }
    const char *pCur = pBeg;
    switch ((pEnd - pBeg + 1) % 8) {
        case 0: do {
#undef  CHECK_CODE
#define CHECK_CODE \
            FilterR <<= (kerr + 1);  \
            FilterR |= FilterSFilterMapArr[(unsigned char) *pCur];\
            if ((FilterR & filter_ok) != filter_ok) {\
                return checkRTR(pBeg, pEnd); \
            }\
            ++pCur;
            CHECK_CODE
        case 7:
            CHECK_CODE
        case 6:
            CHECK_CODE
        case 5:
            CHECK_CODE
        case 4:
            CHECK_CODE
        case 3:
            CHECK_CODE
        case 2:
            CHECK_CODE
        case 1:
            CHECK_CODE
        } while (pCur < pEnd);
    }
    return false;
}

#endif
