#ifndef MYERS_HPP
#define MYERS_HPP

#include <string.h>

#include <string>
#include <stdexcept>

typedef unsigned int MaskType;
//typedef unsigned long long MaskType;


unsigned int CalcDistGlob(const unsigned char* p, bool HasTransp, unsigned int PatLen, MaskType Checker, MaskType VP0, MaskType *PMText);

class CMyersEdist {
public:
    CMyersEdist(const char *pPat, bool bHasTransp) :mPatLen(0), mHasTransp(bHasTransp) {
        MaskType  Flag = 1;

        memset(mPMText, 0, sizeof mPMText);

        mVP0 = 0;
        for (const unsigned char *p = reinterpret_cast<const unsigned char*>(pPat);*p; ++mPatLen, Flag <<= 1, ++p) {
            if (!Flag) {
                throw std::runtime_error("Pattern '" + std::string(pPat) + " is too long, should be no more  than the computer word size");
            }
            mPMText[*p] |= Flag;
            mVP0        |= Flag;
            mChecker     = Flag;
        }
    }
    unsigned int CalcDist(const char* pStr) {
        return CalcDistGlob(reinterpret_cast<const unsigned char*>(pStr), mHasTransp, mPatLen, mChecker, mVP0, mPMText);
    }
    const MaskType* GetPMText() const { return mPMText; }
    MaskType GetVP0() const { return mVP0; }
    MaskType GetChecker() const { return mChecker; }
private:
    MaskType        mPMText[256];
    MaskType        mChecker;
    MaskType        mVP0;
    unsigned int    mPatLen; 
    bool            mHasTransp;
};

class CMyersEdistFastPair {
public:
    CMyersEdistFastPair()  { memset(mPMText, 0, sizeof mPMText); }
    unsigned int CalcDist(const std::string& s1, const std::string& s2, bool bHasTransp) {
        const char* pStr1 = s1.c_str();
        const char* pStr2 = s2.c_str();
        MaskType        Checker = 0;
        MaskType        VP0     = 0;
        MaskType        Flag    = 1;
        unsigned int    PatLen  = 0; 

        /* Init pattern mask */
        for (const unsigned char *p = reinterpret_cast<const unsigned char*>(pStr1);*p; ++PatLen, Flag <<= 1, ++p) {
            if (!Flag) {
                throw std::runtime_error("Pattern '" + std::string(pStr1) + " is too long, should be no more  than the computer word size");
            }
            mPMText[*p] |= Flag;
            VP0         |= Flag;
            Checker      = Flag;
        }
    
        unsigned int dist = CalcDistGlob(reinterpret_cast<const unsigned char*>(pStr2), bHasTransp, PatLen, Checker, VP0, mPMText);

        /* Clear pattern mask */
        for (const unsigned char *p = reinterpret_cast<const unsigned char*>(pStr1);*p; ++p) {
            mPMText[*p] = 0;
        }

        return dist;
    }
private:
    MaskType        mPMText[256];
};

#endif
