/* cldev.c
 * Copyright (C) 2001 by Leonid Boitsov
 *
 * Creating of char device to clear the file cash
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/version.h>
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c)((a)*65536+(b)*256+c)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#else
#endif

#include <linux/fs.h>           /* Here there are definitions for char devices */
#include <linux/wrapper.h>      /* For future compatability */





#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>        /* for (get|put)_user */
#endif
#define SUCCESS 0
/*
 Device declaration
*/
#define DEVICE_NAME "cldev"

static const char *_strchr (const char *str, int c)
{
    for (; *str; str++)
        if (*str == c)
            return str;
    return NULL;
}

static int Device_Open = 0;
static int Major = 1;

#define BUF_LEN 1024
static char buf[BUF_LEN + 1];

#include <linux/pagemap.h>

/*
 * Truncate the page cache at a set offset, removing the pages
 * that are beyond that offset (and zeroing out partial pages).
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
void check_inode_pages (struct inode *inode)
{
    struct page **p;
    struct page *page;

    p = &inode->i_pages;
    while ((page = *p) != NULL) {
        if (PageLocked (page)) {
            printk ("Page is locked %x\n", (int) page);
        } else {
            printk ("Page is unlocked %x\n", (int) page);
        }
        p = &page->next;
    }
    truncate_inode_pages (inode, 0);
}
#endif

static int device_open (struct inode *inode, struct file *file)
{
    if (Device_Open)
        return -EBUSY;
    Device_Open++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
static int device_release (struct inode *inode, struct file *file)
#else
static void device_release (struct inode *inode, struct file *file)
#endif
{
    Device_Open--;
    MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
    return 0;
#endif
}

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
static ssize_t device_read (struct file *file,
                            char *buffer, size_t length, loff_t * offset)
#else
static int device_read (struct inode *inode,
                        struct file *file, char *buffer, int length)
#endif
{
    return -EINVAL;
}


#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
static ssize_t device_write (struct file *file,
                             const char *buffer,
                             size_t length, loff_t * offset)
#else
static int device_write (struct inode *inode,
                         struct file *file, char *buffer, int length)
#endif
{
    int err = 0;
    char *p1, *p2;
    int wlen = -1;
    struct dentry *d;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    struct nameidata nd;
#endif
    do {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
        get_user (buf[++wlen], buffer++);
#else
        buf[++wlen] = get_user (buffer++);
#endif
    }
    while (wlen < BUF_LEN && buf[wlen]);
    buf[wlen] = 0;
    for (p1 = buf; *p1; p1 = p2) {
        // Skip any white spaces
        for (; _strchr ("\r\n\t ", *p1); p1++);
        // Find the first white space

        for (p2 = p1; (*p2) && !_strchr ("\r\n\t ", *p2); p2++);
        if (*p2) {
            *p2++ = 0;
        }
        if (!*p1)
            continue;           //< Let's skip an empty file name
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        if (path_init (p1, 0, &nd))
            err = path_walk (p1, &nd);
        d = nd.dentry;
#else
        d = lookup_dentry (p1, NULL, 0);
#endif
        //printk ("Invalidated file name %s (err=%d)\n", p1, err);
        if (!IS_ERR (d)) {
            if (d->d_inode) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
                //truncate_inode_pages (d->d_inode->i_mapping, 0);
                invalidate_inode_pages(d->d_inode);
#else
                vmtruncate (d->d_inode, 0);
                //check_inode_pages(d->d_inode);
#endif
                //printk ("truncate_inode_pages - Ok\n");
            } else
                printk ("lookup_dentry - failed \n");
        } else
            printk ("lookup_dentry - failed \n");
    }
    return wlen;
}

/* Module declaration */
struct file_operations Fops = {
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,0)
    NULL,                       /* owner */
#endif
    NULL,                       /* seek */
    device_read,
    device_write,
    NULL,                       /* readdir */
    NULL,                       /* select */
    NULL,                       /* ioctl  */
    NULL,                       /* mmap */
    device_open,
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
    NULL,                       /* unclear function */
#endif
    device_release
};

int init_module ()
{
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,0)
    Major = register_chrdev (0, DEVICE_NAME, &Fops);
#else
    Major = module_register_chrdev (0, DEVICE_NAME, &Fops);
#endif

    if (Major < 0) {
        printk ("%s device failed with %d\n",
                "Sorry, registering the character", Major);
        return Major;
    }
    printk ("%s The major device number is %d\n",
            "Registration is a success.", Major);
    printk (" To use the device driver create a device file.\n");
    printk (" mknod /dev/%s c %d 0\n", DEVICE_NAME, Major);
    return 0;
}


void cleanup_module ()
{
    int ret;
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,0)
    ret = unregister_chrdev (Major, DEVICE_NAME);
#else
    ret = module_unregister_chrdev (Major, DEVICE_NAME);
#endif
    if (ret < 0)
        printk ("Error in unregister_chrdev: %d\n", ret);
}

void __get_user_X (void)
{;
}
