/* cldev2.6.c
 * Copyright (C) 2005 by Leonid Boitsov
 *
 * Creating of char device to clear the file cash
*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/version.h>
#include <linux/init.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c)((a)*65536+(b)*256+c)
#endif

#include <linux/fs.h>           /* Here there are definitions for char devices */
#include <linux/namei.h>

#include <asm/uaccess.h>        /* for (get|put)_user */
#define SUCCESS 0
/*
 Device declaration
*/
#define DEVICE_NAME "cldev"

MODULE_LICENSE("GPL");

static const char *_strchr (const char *str, int c)
{
    for (; *str; str++)
        if (*str == c)
            return str;
    return NULL;
}

static int Major = 1;

#define BUF_LEN 1024*128
static char buf[BUF_LEN + 1];

#include <linux/pagemap.h>


static ssize_t device_write (struct file *file,
                             const char *buffer,
                             size_t length, loff_t * offset)
{
    int err = 0;
    char *p1, *p2;
    int wlen = -1;
    struct dentry * d = NULL;
    struct nameidata nd;

    do {
        get_user(buf[++wlen], buffer++);
    } while (wlen < BUF_LEN && buf[wlen]);

    buf[wlen] = 0;
    for (p1 = buf; *p1; p1 = p2) {
        // Skip any white spaces
        for (; _strchr ("\r\n\t ", *p1); p1++);
        // Find the first white space

        for (p2 = p1; (*p2) && !_strchr ("\r\n\t ", *p2); p2++);
        if (*p2) {
            *p2++ = 0;
        }
        //printk ("Invalidated file name %s (err=%d)\n", p1, err);
        if (!*p1) {
            continue;           //< Let's skip an empty file name
	}
	err = path_lookup(p1, 0, &nd);
    d = nd.dentry;

        //printk ("Path walk result for file %s err=%d\n", p1, err);

	if (!err) {
            invalidate_remote_inode(d->d_inode);
#if 0
            printk ("Invalidated file name %s - Ok (err=%d)\n", p1, err);
#endif

	    path_release(&nd);
	} else {
            printk ("Invalidated file name %s - Failed (err=%d)\n", p1, err);
	}
    }
    return wlen;
}

/* Module declaration */
struct file_operations Fops = {
    .owner = THIS_MODULE,
    .write = device_write,
};

static int __init my_init_module(void)
{
    Major = register_chrdev (0, DEVICE_NAME, &Fops);

    if (Major < 0) {
        printk ("%s device failed with %d\n",
                "Sorry, registering the character", Major);
        return Major;
    }
    printk ("%s The major device number is %d\n",
            "Registration is a success.", Major);
    printk (" To use the device driver create a device file.\n");
    printk (" mknod /dev/%s c %d 0\n", DEVICE_NAME, Major);
    return 0;
}


static void __exit my_cleanup_module(void)
{
/*    
    int ret;
    ret = unregister_chrdev (Major, DEVICE_NAME);
    if (ret < 0)
        printk ("Error in unregister_chrdev: %d\n", ret);
*/
    unregister_chrdev (Major, DEVICE_NAME);
}

module_init(my_init_module)
module_exit(my_cleanup_module)

void __get_user_X (void)
{;
}
