#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <sys/types.h>

extern u_int32_t CRC32 (char *s, int size);


const int MaxLen = 64;

static int CharStat[256];
static int CharStatNZ[256];
static unsigned char CharNZ[256];
static int LenStat[MaxLen];
static int LenStatNZ[MaxLen];
static int LenNZ[MaxLen];
static int iLenNZ;
static int iCharNZ;
static int CharStatSum;
static int LenStatSum;

struct st {
    int Stat;
    char Char;
};
st CharStatStruct[256];

void Error (const char *p = NULL)
{
    perror (p);
}

int CompStat (const void *a, const void *b)
{
    int i = ((st *) a)->Stat;
    int j = ((st *) b)->Stat;
    return j - i;
}

bool ReadStat (const char *fname)
{
    char str[MaxLen + 3];
    FILE *f0 = fopen (fname, "r");
    if (!f0) {
        Error ("Failed to open the file");
        return false;
    }
    errno = 0;
    while (fgets (str, MaxLen, f0)) {
        str[MaxLen + 2] = 0;
        int l = strlen (str) - 1;
        if (l < 0)
            continue;
        for (int i = 1; i <= 2; i++)
            if (str[l] == '\n' || str[l] == '\r')
                str[l--] = 0;
        if (l < 0)
            continue;


        LenStatSum++;
        LenStat[l]++;
        for (int i = 0; i < l; i++) {
            CharStatSum++;
            CharStat[toupper ((unsigned char) str[i])]++;
            //CharStat[tolower ((unsigned char) str[i])]++;
        }
    }
    if (errno) {
        Error ("Error reading the file");
        return false;
    }
    if (fclose (f0)) {
        Error ("Error closing the file");
        return false;
    }
    //printf("%f \n",CharStat[(unsigned char)'E']/double(CharStatSum));
    iCharNZ = 0;
    int CharCheckSum = 0;
    for (int i = 0; i < 256; i++)
        if (CharStat[i]) {
            CharStatNZ[iCharNZ] = CharStat[i];
            CharNZ[iCharNZ] = i;
            iCharNZ++;
            CharCheckSum += CharStat[i];
            CharStatStruct[iCharNZ].Stat = CharStat[i];
            CharStatStruct[iCharNZ].Char = i;
        }
    if (CharCheckSum != CharStatSum) {
        fprintf (stderr, "Charstat checksum does not match!\n");
        return false;
    }
    // Sorting CharStat[]
    qsort ((void *) CharStatStruct, iCharNZ, sizeof (st), CompStat);
    //
    int LenCheckSum = 0;
    iLenNZ = 0;
    for (int i = 0; i < MaxLen; i++)
        if (LenStat[i]) {
            LenStatNZ[iLenNZ] = LenStat[i];
            LenNZ[iLenNZ] = i;
            iLenNZ++;
            LenCheckSum += LenStat[i];
        }
    if (LenCheckSum != LenStatSum) {
        fprintf (stderr, "Lenstat checksum does not match!\n");
        return false;
    }

    return true;
};

typedef std::map < u_int32_t, bool > CheckMap;

int main (int argc, char *argv[])
{
    CheckMap UniqFlag;
    if (argc < 2)
        return 3;
    if (!ReadStat (argv[1]))
        return 1;

    if (!(LenStatSum && CharStatSum)) {
        fprintf (stderr, "Zero statistics!\n");
        return 2;
    }
    time_t t;
    srand (time (&t));
    if (argc == 3 && LenStatSum && CharStatSum && strcmp (argv[2], "-s")) {
        for (int i = 0; i < atoi(argv[2]); i++) {
            while (true)        //Cycle to get an unique word
            {
                // First of all get the length

                int len = 0;
                int rl = rand () % LenStatSum;
                bool flag = false;
                int sum = 0;
                for (int i = 0; i <= iLenNZ; sum += LenStatNZ[i++])
                    if (rl >= sum && rl < sum + LenStatNZ[i]) {
                        len = LenNZ[i] + 1;
                        flag = true;
                        break;
                    }
                if (!flag) {
                    fprintf (stderr,
                             "Error evaluating random length StatSum =%d Random Number =%d\n",
                             sum, rl);
                    // Dump pairs sum, sum + LenStatNZ[i]     
                    sum = 0;
                    for (int i = 0; i <= iLenNZ; sum += LenStatNZ[i++])
                        fprintf (stderr, "(%d,%d)\n", sum,
                                 sum + LenStatNZ[i]);
                    return -1;
                }
                //len+=5;

                char str[len + 1];
                str[len] = 0;

                for (int i = 0; i < len; i++) {
                    int rl = rand () % CharStatSum;
                    bool flag = false;
                    for (int sum = 0, j = 0; j <= iCharNZ;
                         sum += CharStatNZ[j++])
                        if (rl >= sum && rl < sum + CharStatNZ[j]) {
                            str[i] = CharNZ[j];
                            flag = true;
                            break;
                        }
                    if (!flag) {
                        fprintf (stderr, "Error evaluating random char\n");
                        return -1;
                    }
                }
                u_int32_t c = CRC32 (str, len);
                if (!UniqFlag.count (c)) {
                    UniqFlag[c] = 1;
                    printf ("%s\n", str);
                    break;
                }
            }
        }
    }
    return 0;
}
