#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <time.h>
#include <map>
#include "../Common/utils.h"

#include "interface.h"

using namespace std;

void usage(const char *pProg) 
{
    cerr << "Usage : " << pProg << " <file> <hasTransp> <kerr> <testcnt> " << endl;
}

int main(int argc, char *pArgv[])
{
    if (argc != 5) {
        usage(pArgv[0]);
        return -1;
    }


    ifstream f(pArgv[1]);

    if (!f) {
        perror(("Cannot open file: '" + string(pArgv[1]) +  "'").c_str());
        usage(pArgv[0]);
        return -1;
    }

    bool hasTransp = atol(pArgv[2]);
    int kerr = atol(pArgv[3]);
    int cnt = atol(pArgv[4]);

    srand(time(NULL));

    vector<string> vs;
    char           buf[1024];

    while(f.getline(buf, sizeof buf - 1)) {
        vs.push_back(buf); 
        buf[sizeof buf - 1] = 0;
    }


    int cntLeft = cnt;

    int pos = 0; 

    map<string, int> used;

    if (!vs.size()) {
        cerr << "Empty dictionary\n";
        return 1;
    }

    string              s;
    GetEnv(ENV_ABC, s);

    EAlphabet           AlphaType = GetAlphabetType(s);
    if (AlphaType == eUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return 1;
    }


    while(cntLeft > 0) {
        StringList lst;

        int pos = rand() % vs.size();
        string str = vs[pos];

        if (kerr) {
             FillExpansionList (str,
                        hasTransp,
                        kerr,
                        &lst,
                        GetAlphabet(AlphaType),
                        (kerr < 3) ? 0: kerr - 1, true, 
                        0, false);

             int sz = lst.size();
             assert(sz);
             int indx = rand() % lst.size();

             //cout << str << " -> " << lst[indx] << endl;
             if (used[lst[indx]]) {
                 continue;
             }
             used[lst[indx]] = 1;
             cout << lst[indx] << endl;
        } else {
             cout << str << endl;
        }
        --cntLeft;
    }

    return 0;
}
